@@ -27,10 +27,11 @@ from scripts.test.shared import (
     WASM_DIS, WASM_REDUCE, binary_format_check, delete_from_orbit, fail, fail_with_error,
     fail_if_not_identical, fail_if_not_contained, has_vanilla_emcc,
     has_vanilla_llvm, minify_check, num_failures, options, tests,
-    requested, warnings, has_shell_timeout
+    requested, warnings, has_shell_timeout, fail_if_not_identical_to_file
 )
 
 import scripts.test.asm2wasm as asm2wasm
+import scripts.test.lld as lld
 import scripts.test.s2wasm as s2wasm
 import scripts.test.wasm2asm as wasm2asm
 
@@ -38,6 +39,19 @@ if options.interpreter:
   print '[ using wasm interpreter at "%s" ]' % options.interpreter
   assert os.path.exists(options.interpreter), 'interpreter not found'
 
+# run a check with BINARYEN_PASS_DEBUG set, to do full validation
+def with_pass_debug(check):
+  old_pass_debug = os.environ.get('BINARYEN_PASS_DEBUG')
+  try:
+    os.environ['BINARYEN_PASS_DEBUG'] = '1'
+    check()
+  finally:
+    if old_pass_debug is not None:
+      os.environ['BINARYEN_PASS_DEBUG'] = old_pass_debug
+    else:
+      if 'BINARYEN_PASS_DEBUG' in os.environ:
+        del os.environ['BINARYEN_PASS_DEBUG']
+
 # tests
 
 def run_help_tests():
@@ -59,11 +73,12 @@ def run_help_tests():
 def run_wasm_opt_tests():
   print '\n[ checking wasm-opt -o notation... ]\n'
 
-  wast = os.path.join(options.binaryen_test, 'hello_world.wast')
-  delete_from_orbit('a.wast')
-  cmd = WASM_OPT + [wast, '-o', 'a.wast', '-S']
-  run_command(cmd)
-  fail_if_not_identical(open('a.wast').read(), open(wast).read())
+  for extra_args in [[], ['--no-validation']]:
+    wast = os.path.join(options.binaryen_test, 'hello_world.wast')
+    delete_from_orbit('a.wast')
+    cmd = WASM_OPT + [wast, '-o', 'a.wast', '-S'] + extra_args
+    run_command(cmd)
+    fail_if_not_identical_to_file(open('a.wast').read(), wast)
 
   print '\n[ checking wasm-opt binary reading/writing... ]\n'
 
@@ -81,7 +96,10 @@ def run_wasm_opt_tests():
     if t.endswith(('.wast', '.wasm')):
       print '..', t
       binary = '.wasm' in t
-      passname = os.path.basename(t).replace('.wast', '').replace('.wasm', '')
+      base = os.path.basename(t).replace('.wast', '').replace('.wasm', '')
+      passname = base
+      if passname.isdigit():
+        passname = open(os.path.join(options.binaryen_test, 'passes', passname + '.passes')).read().strip()
       opts = [('--' + p if not p.startswith('O') else '-' + p) for p in passname.split('_')]
       t = os.path.join(options.binaryen_test, 'passes', t)
       actual = ''
@@ -95,28 +113,21 @@ def run_wasm_opt_tests():
         debugged = run_command(cmd + ['--debug'], stderr=subprocess.PIPE)
         fail_if_not_contained(actual, debugged)
         # also check pass-debug mode
-        old_pass_debug = os.environ.get('BINARYEN_PASS_DEBUG')
-        try:
-          os.environ['BINARYEN_PASS_DEBUG'] = '1'
+        def check():
           pass_debug = run_command(cmd)
           fail_if_not_identical(curr, pass_debug)
-        finally:
-          if old_pass_debug is not None:
-            os.environ['BINARYEN_PASS_DEBUG'] = old_pass_debug
-          else:
-            if 'BINARYEN_PASS_DEBUG' in os.environ:
-              del os.environ['BINARYEN_PASS_DEBUG']
+        with_pass_debug(check)
 
-      fail_if_not_identical(actual, open(os.path.join('test', 'passes', passname + ('.bin' if binary else '') + '.txt'), 'rb').read())
+      expected_file = os.path.join(options.binaryen_test, 'passes',
+                                   base + ('.bin' if binary else '') + '.txt')
+      fail_if_not_identical_to_file(actual, expected_file)
 
       if 'emit-js-wrapper' in t:
         with open('a.js') as actual:
-          with open(t + '.js') as expected:
-            fail_if_not_identical(actual.read(), expected.read())
+          fail_if_not_identical_to_file(actual.read(), t + '.js')
       if 'emit-spec-wrapper' in t:
         with open('a.wat') as actual:
-          with open(t + '.wat') as expected:
-            fail_if_not_identical(actual.read(), expected.read())
+          fail_if_not_identical_to_file(actual.read(), t + '.wat')
 
   print '\n[ checking wasm-opt parsing & printing... ]\n'
 
@@ -127,7 +138,8 @@ def run_wasm_opt_tests():
       cmd = WASM_OPT + [os.path.join(options.binaryen_test, 'print', t), '--print']
       print '    ', ' '.join(cmd)
       actual, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
-      fail_if_not_identical(actual, open(os.path.join(options.binaryen_test, 'print', wasm + '.txt')).read())
+      expected_file = os.path.join(options.binaryen_test, 'print', wasm + '.txt')
+      fail_if_not_identical_to_file(actual, expected_file)
       cmd = WASM_OPT + [os.path.join(options.binaryen_test, 'print', t), '--print-minified']
       print '    ', ' '.join(cmd)
       actual, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
@@ -144,10 +156,7 @@ def run_wasm_opt_tests():
       actual = run_command(cmd)
       actual = actual.replace('printing before:\n', '')
 
-      expected = open(f, 'rb').read()
-
-      if actual != expected:
-        fail(actual, expected)
+      fail_if_not_identical_to_file(actual, f)
 
       binary_format_check(t, wasm_as_args=['-g']) # test with debuginfo
       binary_format_check(t, wasm_as_args=[], binary_suffix='.fromBinary.noDebugInfo') # test without debuginfo
@@ -165,19 +174,22 @@ def run_wasm_dis_tests():
       if os.path.isfile(t + '.map'): cmd += ['--source-map', t + '.map']
 
       actual = run_command(cmd)
+      fail_if_not_identical_to_file(actual, t + '.fromBinary')
 
-      with open(t + '.fromBinary') as f:
-        expected = f.read()
-        if actual != expected:
-          fail(actual, expected)
+      # also verify there are no validation errors
+      def check():
+        cmd = WASM_OPT + [t]
+        actual = run_command(cmd)
+      with_pass_debug(check)
 
 def run_wasm_merge_tests():
   print '\n[ checking wasm-merge... ]\n'
 
-  for t in os.listdir(os.path.join('test', 'merge')):
+  test_dir = os.path.join(options.binaryen_test, 'merge')
+  for t in os.listdir(test_dir):
     if t.endswith(('.wast', '.wasm')):
       print '..', t
-      t = os.path.join('test', 'merge', t)
+      t = os.path.join(test_dir, t)
       u = t + '.toMerge'
       for finalize in [0, 1]:
         for opt in [0, 1]:
@@ -189,18 +201,17 @@ def run_wasm_merge_tests():
           out = t + '.combined'
           if finalize: out += '.finalized'
           if opt: out += '.opt'
-          with open(out) as f:
-            fail_if_not_identical(f.read(), actual)
-          with open(out + '.stdout') as f:
-            fail_if_not_identical(f.read(), stdout)
+          fail_if_not_identical_to_file(actual, out)
+          fail_if_not_identical_to_file(stdout, out + '.stdout')
 
 def run_crash_tests():
   print "\n[ checking we don't crash on tricky inputs... ]\n"
 
-  for t in os.listdir(os.path.join('test', 'crash')):
+  test_dir = os.path.join(options.binaryen_test, 'crash')
+  for t in os.listdir(test_dir):
     if t.endswith(('.wast', '.wasm')):
       print '..', t
-      t = os.path.join('test', 'crash', t)
+      t = os.path.join(test_dir, t)
       cmd = WASM_OPT + [t]
       # expect a parse error to be reported
       run_command(cmd, expected_err='parse exception:', err_contains=True, expected_status=1)
@@ -208,50 +219,61 @@ def run_crash_tests():
 def run_ctor_eval_tests():
   print '\n[ checking wasm-ctor-eval... ]\n'
 
-  for t in os.listdir(os.path.join('test', 'ctor-eval')):
+  test_dir = os.path.join(options.binaryen_test, 'ctor-eval')
+  for t in os.listdir(test_dir):
     if t.endswith(('.wast', '.wasm')):
       print '..', t
-      t = os.path.join('test', 'ctor-eval', t)
+      t = os.path.join(test_dir, t)
       ctors = open(t + '.ctors').read().strip()
       cmd = WASM_CTOR_EVAL + [t, '-o', 'a.wast', '-S', '--ctors', ctors]
       stdout = run_command(cmd)
       actual = open('a.wast').read()
       out = t + '.out'
-      with open(out) as f:
-        fail_if_not_identical(f.read(), actual)
+      fail_if_not_identical_to_file(actual, out)
 
 def run_wasm_metadce_tests():
   print '\n[ checking wasm-metadce ]\n'
 
-  for t in os.listdir(os.path.join('test', 'metadce')):
+  test_dir = os.path.join(options.binaryen_test, 'metadce')
+  for t in os.listdir(test_dir):
     if t.endswith(('.wast', '.wasm')):
       print '..', t
-      t = os.path.join('test', 'metadce', t)
+      t = os.path.join(test_dir, t)
       graph = t + '.graph.txt'
       cmd = WASM_METADCE + [t, '--graph-file=' + graph, '-o', 'a.wast', '-S']
       stdout = run_command(cmd)
       expected = t + '.dced'
       with open('a.wast') as seen:
-        with open(expected) as correct:
-          fail_if_not_identical(seen.read(), correct.read())
-      with open(expected + '.stdout') as correct:
-        fail_if_not_identical(stdout, correct.read())
+        fail_if_not_identical_to_file(seen.read(), expected)
+      fail_if_not_identical_to_file(stdout, expected + '.stdout')
 
 def run_wasm_reduce_tests():
-  print '\n[ checking wasm-reduce ]\n'
+  print '\n[ checking wasm-reduce testcases]\n'
 
-  for t in os.listdir(os.path.join('test', 'reduce')):
+  # fixed testcases
+  test_dir = os.path.join(options.binaryen_test, 'reduce')
+  for t in os.listdir(test_dir):
     if t.endswith('.wast'):
       print '..', t
-      t = os.path.join('test', 'reduce', t)
+      t = os.path.join(test_dir, t)
       # convert to wasm
       run_command(WASM_AS + [t, '-o', 'a.wasm'])
-      print run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm'])
+      run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm'])
       expected = t + '.txt'
       run_command(WASM_DIS + ['c.wasm', '-o', 'a.wast'])
       with open('a.wast') as seen:
-        with open(expected) as correct:
-          fail_if_not_identical(seen.read(), correct.read())
+        fail_if_not_identical_to_file(seen.read(), expected)
+
+  # run on a nontrivial fuzz testcase, for general coverage
+  # this is very slow in ThreadSanitizer, so avoid it there
+  if 'fsanitize=thread' not in str(os.environ):
+    print '\n[ checking wasm-reduce fuzz testcase ]\n'
+
+    run_command(WASM_OPT + [os.path.join(options.binaryen_test, 'unreachable-import_wasm-only.asm.js'), '-ttf', '-Os', '-o', 'a.wasm'])
+    before = os.stat('a.wasm').st_size
+    run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm'])
+    after = os.stat('c.wasm').st_size
+    assert after < 0.333 * before, [before, after]
 
 def run_spec_tests():
   print '\n[ checking wasm-shell spec testcases... ]\n'
@@ -359,7 +381,11 @@ def run_binaryen_js_tests():
     if not s.endswith('.js'): continue
     print s
     f = open('a.js', 'w')
-    binaryen_js = open(os.path.join(options.binaryen_bin, 'binaryen.js')).read()
+    # avoid stdout/stderr ordering issues in some js shells - use just stdout
+    f.write('''
+      console.warn = function(x) { console.log(x) };
+    ''')
+    binaryen_js = open(os.path.join(options.binaryen_root, 'bin', 'binaryen.js')).read()
     f.write(binaryen_js)
     if NODEJS:
       f.write(node_test_glue())
@@ -486,7 +512,7 @@ def run_gcc_torture_tests():
   else:
     for t in sorted(os.listdir(os.path.join(options.binaryen_test, 'example'))):
       output_file = os.path.join(options.binaryen_bin, 'example')
-      cmd = ['-I' + os.path.join(options.binaryen_root, 'src'), '-g', '-lasmjs', '-lsupport', '-L' + os.path.join(options.binaryen_bin, '..', 'lib'), '-pthread', '-o', output_file]
+      cmd = ['-I' + os.path.join(options.binaryen_root, 'src'), '-g', '-pthread', '-o', output_file]
       if t.endswith('.txt'):
         # check if there is a trace in the file, if so, we should build it
         out = subprocess.Popen([os.path.join('scripts', 'clean_c_api_trace.py'), os.path.join(options.binaryen_test, 'example', t)], stdout=subprocess.PIPE).communicate()[0]
@@ -504,10 +530,12 @@ def run_gcc_torture_tests():
         # build the C file separately
         extra = [NATIVECC, src, '-c', '-o', 'example.o',
                  '-I' + os.path.join(options.binaryen_root, 'src'), '-g', '-L' + os.path.join(options.binaryen_bin, '..', 'lib'), '-pthread']
+        if src.endswith('.cpp'):
+          extra += ['-std=c++11']
         print 'build: ', ' '.join(extra)
         subprocess.check_call(extra)
         # Link against the binaryen C library DSO, using an executable-relative rpath
-        cmd = ['example.o', '-lbinaryen'] + cmd + ['-Wl,-rpath=$ORIGIN/../lib']
+        cmd = ['example.o', '-L' + os.path.join(options.binaryen_bin, '..', 'lib'), '-lbinaryen'] + cmd + ['-Wl,-rpath=$ORIGIN/../lib']
       else:
         continue
       print '  ', t, src, expected
@@ -528,9 +556,7 @@ def run_gcc_torture_tests():
           # Also removes debug directory produced on Mac OS
           shutil.rmtree(output_file + '.dSYM')
 
-      expected = open(expected).read()
-      if actual != expected:
-        fail(actual, expected)
+      fail_if_not_identical_to_file(actual, expected)
 
 def run_emscripten_tests():
   print '\n[ checking wasm.js methods... ]\n'
@@ -604,6 +630,7 @@ def main():
   run_binaryen_js_tests()
   s2wasm.test_s2wasm()
   s2wasm.test_linker()
+  lld.test_wasm_emscripten_finalize()
   wasm2asm.test_wasm2asm()
   run_validator_tests()
   if options.torture and options.test_waterfall:
