@@ -43,9 +43,17 @@ extern cashew::IString GLOBAL,
                 CLZ32,
                 FROUND,
                 ASM2WASM,
+                MIN,
+                MAX,
                 F64_REM,
                 F64_TO_INT,
+                F64_TO_UINT,
                 F64_TO_INT64,
+                F64_TO_UINT64,
+                F32_TO_INT,
+                F32_TO_UINT,
+                F32_TO_INT64,
+                F32_TO_UINT64,
                 I32S_DIV,
                 I32U_DIV,
                 I32S_REM,
@@ -59,26 +67,32 @@ extern cashew::IString GLOBAL,
                 I32_TEMP,
                 DEBUGGER,
                 USE_ASM,
+                ALMOST_ASM,
                 BUFFER,
                 ENV,
                 INSTRUMENT,
                 MATH_IMUL,
                 MATH_ABS,
                 MATH_CEIL,
+                MATH_CLZ32,
                 MATH_FLOOR,
                 MATH_TRUNC,
                 MATH_NEAREST,
                 MATH_SQRT,
                 MATH_MIN,
                 MATH_MAX,
-                CTZ32,
-                CTZ64,
-                POPCNT32,
-                POPCNT64,
-                ROTL32,
-                ROTL64,
-                ROTR32,
-                ROTR64;
+                WASM_CTZ32,
+                WASM_CTZ64,
+                WASM_CLZ32,
+                WASM_CLZ64,
+                WASM_POPCNT32,
+                WASM_POPCNT64,
+                WASM_ROTL32,
+                WASM_ROTL64,
+                WASM_ROTR32,
+                WASM_ROTR64,
+                WASM_GROW_MEMORY,
+                WASM_CURRENT_MEMORY;
 }
 
 #endif // wasm_asmjs_shared_constants_h
