@@ -33,7 +33,6 @@
 #include "cfg/Relooper.h"
 #include "ir/utils.h"
 #include "shell-interface.h"
-#include "support/defaults.h"
 
 using namespace wasm;
 
@@ -45,10 +44,10 @@ BinaryenLiteral toBinaryenLiteral(Literal x) {
   BinaryenLiteral ret;
   ret.type = x.type;
   switch (x.type) {
-    case WasmType::i32: ret.i32 = x.geti32(); break;
-    case WasmType::i64: ret.i64 = x.geti64(); break;
-    case WasmType::f32: ret.i32 = x.reinterpreti32(); break;
-    case WasmType::f64: ret.i64 = x.reinterpreti64(); break;
+    case Type::i32: ret.i32 = x.geti32(); break;
+    case Type::i64: ret.i64 = x.geti64(); break;
+    case Type::f32: ret.i32 = x.reinterpreti32(); break;
+    case Type::f64: ret.i64 = x.reinterpreti64(); break;
     default: abort();
   }
   return ret;
@@ -56,10 +55,10 @@ BinaryenLiteral toBinaryenLiteral(Literal x) {
 
 Literal fromBinaryenLiteral(BinaryenLiteral x) {
   switch (x.type) {
-    case WasmType::i32: return Literal(x.i32);
-    case WasmType::i64: return Literal(x.i64);
-    case WasmType::f32: return Literal(x.i32).castToF32();
-    case WasmType::f64: return Literal(x.i64).castToF64();
+    case Type::i32: return Literal(x.i32);
+    case Type::i64: return Literal(x.i64);
+    case Type::f32: return Literal(x.i32).castToF32();
+    case Type::f64: return Literal(x.i64).castToF64();
     default: abort();
   }
 }
@@ -72,9 +71,7 @@ static std::mutex BinaryenFunctionMutex;
 static std::mutex BinaryenFunctionTypeMutex;
 
 // Optimization options
-static int optimizeLevel = BINARYEN_DEFAULT_OPTIMIZE_LEVEL;
-static int shrinkLevel = BINARYEN_DEFAULT_SHRINK_LEVEL;
-static bool debugInfo = BINARYEN_DEFAULT_DEBUG_INFO;
+static PassOptions globalPassOptions = PassOptions::getWithDefaultOptimizationOptions();
 
 // Tracing support
 
@@ -183,6 +180,8 @@ void BinaryenModuleDispose(BinaryenModuleRef module) {
     functionTypes.clear();
     expressions.clear();
     functions.clear();
+    imports.clear();
+    exports.clear();
     relooperBlocks.clear();
   }
 
@@ -196,9 +195,9 @@ BinaryenFunctionTypeRef BinaryenAddFunctionType(BinaryenModuleRef module, const
   auto* ret = new FunctionType;
   if (name) ret->name = name;
   else ret->name = Name::fromInt(wasm->functionTypes.size());
-  ret->result = WasmType(result);
+  ret->result = Type(result);
   for (BinaryenIndex i = 0; i < numParams; i++) {
-    ret->params.push_back(WasmType(paramTypes[i]));
+    ret->params.push_back(Type(paramTypes[i]));
   }
 
   // Lock. This can be called from multiple threads at once, and is a
@@ -227,6 +226,23 @@ BinaryenFunctionTypeRef BinaryenAddFunctionType(BinaryenModuleRef module, const
 
   return ret;
 }
+void BinaryenRemoveFunctionType(BinaryenModuleRef module, const char* name) {
+  if (tracing) {
+    std::cout << "  BinaryenRemoveFunctionType(the_module, ";
+    traceNameOrNULL(name);
+    std::cout << ");\n";
+  }
+
+  auto* wasm = (Module*)module;
+  assert(name != NULL);
+
+  // Lock. This can be called from multiple threads at once, and is a
+  // point where they all access and modify the module.
+  {
+    std::lock_guard<std::mutex> lock(BinaryenFunctionTypeMutex);
+    wasm->removeFunctionType(name);
+  }
+}
 
 BinaryenLiteral BinaryenLiteralInt32(int32_t x) { return toBinaryenLiteral(Literal(x)); }
 BinaryenLiteral BinaryenLiteralInt64(int64_t x) { return toBinaryenLiteral(Literal(x)); }
@@ -272,6 +288,11 @@ BinaryenOp BinaryenTruncUFloat64ToInt32(void) { return TruncUFloat64ToInt32; }
 BinaryenOp BinaryenTruncUFloat64ToInt64(void) { return TruncUFloat64ToInt64; }
 BinaryenOp BinaryenReinterpretFloat32(void) { return ReinterpretFloat32; }
 BinaryenOp BinaryenReinterpretFloat64(void) { return ReinterpretFloat64; }
+BinaryenOp BinaryenExtendS8Int32(void) { return ExtendS8Int32; }
+BinaryenOp BinaryenExtendS16Int32(void) { return ExtendS16Int32; }
+BinaryenOp BinaryenExtendS8Int64(void) { return ExtendS8Int64; }
+BinaryenOp BinaryenExtendS16Int64(void) { return ExtendS16Int64; }
+BinaryenOp BinaryenExtendS32Int64(void) { return ExtendS32Int64; }
 BinaryenOp BinaryenConvertSInt32ToFloat32(void) { return ConvertSInt32ToFloat32; }
 BinaryenOp BinaryenConvertSInt32ToFloat64(void) { return ConvertSInt32ToFloat64; }
 BinaryenOp BinaryenConvertUInt32ToFloat32(void) { return ConvertUInt32ToFloat32; }
@@ -377,7 +398,7 @@ BinaryenExpressionRef BinaryenBlock(BinaryenModuleRef module, const char* name,
   for (BinaryenIndex i = 0; i < numChildren; i++) {
     ret->list.push_back((Expression*)children[i]);
   }
-  if (type != BinaryenTypeAuto()) ret->finalize(WasmType(type));
+  if (type != BinaryenTypeAuto()) ret->finalize(Type(type));
   else ret->finalize();
 
   if (tracing) {
@@ -385,6 +406,7 @@ BinaryenExpressionRef BinaryenBlock(BinaryenModuleRef module, const char* name,
     std::cout << "    BinaryenExpressionRef children[] = { ";
     for (BinaryenIndex i = 0; i < numChildren; i++) {
       if (i > 0) std::cout << ", ";
+      if (i % 6 == 5) std::cout << "\n       "; // don't create hugely long lines
       std::cout << "expressions[" << expressions[children[i]] << "]";
     }
     if (numChildren == 0) std::cout << "0"; // ensure the array is not empty, otherwise a compiler error on VS
@@ -484,7 +506,7 @@ BinaryenExpressionRef BinaryenCall(BinaryenModuleRef module, const char *target,
   for (BinaryenIndex i = 0; i < numOperands; i++) {
     ret->operands.push_back((Expression*)operands[i]);
   }
-  ret->type = WasmType(returnType);
+  ret->type = Type(returnType);
   ret->finalize();
   return static_cast<Expression*>(ret);
 }
@@ -509,7 +531,7 @@ BinaryenExpressionRef BinaryenCallImport(BinaryenModuleRef module, const char *t
   for (BinaryenIndex i = 0; i < numOperands; i++) {
     ret->operands.push_back((Expression*)operands[i]);
   }
-  ret->type = WasmType(returnType);
+  ret->type = Type(returnType);
   ret->finalize();
   return static_cast<Expression*>(ret);
 }
@@ -549,7 +571,7 @@ BinaryenExpressionRef BinaryenGetLocal(BinaryenModuleRef module, BinaryenIndex i
   }
 
   ret->index = index;
-  ret->type = WasmType(type);
+  ret->type = Type(type);
   ret->finalize();
   return static_cast<Expression*>(ret);
 }
@@ -590,7 +612,7 @@ BinaryenExpressionRef BinaryenGetGlobal(BinaryenModuleRef module, const char *na
   }
 
   ret->name = name;
-  ret->type = WasmType(type);
+  ret->type = Type(type);
   ret->finalize();
   return static_cast<Expression*>(ret);
 }
@@ -619,7 +641,7 @@ BinaryenExpressionRef BinaryenLoad(BinaryenModuleRef module, uint32_t bytes, int
   ret->signed_ = !!signed_;
   ret->offset = offset;
   ret->align = align ? align : bytes;
-  ret->type = WasmType(type);
+  ret->type = Type(type);
   ret->ptr = (Expression*)ptr;
   ret->finalize();
   return static_cast<Expression*>(ret);
@@ -637,7 +659,7 @@ BinaryenExpressionRef BinaryenStore(BinaryenModuleRef module, uint32_t bytes, ui
   ret->align = align ? align : bytes;
   ret->ptr = (Expression*)ptr;
   ret->value = (Expression*)value;
-  ret->valueType = WasmType(type);
+  ret->valueType = Type(type);
   ret->finalize();
   return static_cast<Expression*>(ret);
 }
@@ -646,16 +668,16 @@ BinaryenExpressionRef BinaryenConst(BinaryenModuleRef module, BinaryenLiteral va
   if (tracing) {
     auto id = noteExpression(ret);
     switch (value.type) {
-      case WasmType::i32: std::cout << "  expressions[" << id << "] = BinaryenConst(the_module, BinaryenLiteralInt32(" << value.i32 << "));\n"; break;
-      case WasmType::i64: std::cout << "  expressions[" << id << "] = BinaryenConst(the_module, BinaryenLiteralInt64(" << value.i64 << "));\n"; break;
-      case WasmType::f32: {
+      case Type::i32: std::cout << "  expressions[" << id << "] = BinaryenConst(the_module, BinaryenLiteralInt32(" << value.i32 << "));\n"; break;
+      case Type::i64: std::cout << "  expressions[" << id << "] = BinaryenConst(the_module, BinaryenLiteralInt64(" << value.i64 << "));\n"; break;
+      case Type::f32: {
         std::cout << "  expressions[" << id << "] = BinaryenConst(the_module, BinaryenLiteralFloat32(";
         if (std::isnan(value.f32)) std::cout << "NAN";
         else std::cout << value.f32;
         std::cout << "));\n";
         break;
       }
-      case WasmType::f64: {
+      case Type::f64: {
         std::cout << "  expressions[" << id << "] = BinaryenConst(the_module, BinaryenLiteralFloat64(";
         if (std::isnan(value.f64)) std::cout << "NAN";
         else std::cout << value.f64;
@@ -769,7 +791,7 @@ BinaryenExpressionRef BinaryenUnreachable(BinaryenModuleRef module) {
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenAtomicLoad(BinaryenModuleRef module, uint32_t bytes, uint32_t offset, BinaryenType type, BinaryenExpressionRef ptr) {
-  auto* ret = Builder(*((Module*)module)).makeAtomicLoad(bytes, offset, (Expression*)ptr, WasmType(type));
+  auto* ret = Builder(*((Module*)module)).makeAtomicLoad(bytes, offset, (Expression*)ptr, Type(type));
 
   if (tracing) {
     auto id = noteExpression(ret);
@@ -779,7 +801,7 @@ BinaryenExpressionRef BinaryenAtomicLoad(BinaryenModuleRef module, uint32_t byte
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenAtomicStore(BinaryenModuleRef module, uint32_t bytes, uint32_t offset, BinaryenExpressionRef ptr, BinaryenExpressionRef value, BinaryenType type) {
-  auto* ret = Builder(*((Module*)module)).makeAtomicStore(bytes, offset, (Expression*)ptr, (Expression*)value, WasmType(type));
+  auto* ret = Builder(*((Module*)module)).makeAtomicStore(bytes, offset, (Expression*)ptr, (Expression*)value, Type(type));
 
   if (tracing) {
     auto id = noteExpression(ret);
@@ -789,7 +811,7 @@ BinaryenExpressionRef BinaryenAtomicStore(BinaryenModuleRef module, uint32_t byt
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenAtomicRMW(BinaryenModuleRef module, BinaryenOp op, BinaryenIndex bytes, BinaryenIndex offset, BinaryenExpressionRef ptr, BinaryenExpressionRef value, BinaryenType type) {
-  auto* ret = Builder(*((Module*)module)).makeAtomicRMW(AtomicRMWOp(op), bytes, offset, (Expression*)ptr, (Expression*)value, WasmType(type));
+  auto* ret = Builder(*((Module*)module)).makeAtomicRMW(AtomicRMWOp(op), bytes, offset, (Expression*)ptr, (Expression*)value, Type(type));
 
   if (tracing) {
     auto id = noteExpression(ret);
@@ -799,7 +821,7 @@ BinaryenExpressionRef BinaryenAtomicRMW(BinaryenModuleRef module, BinaryenOp op,
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenAtomicCmpxchg(BinaryenModuleRef module, BinaryenIndex bytes, BinaryenIndex offset, BinaryenExpressionRef ptr, BinaryenExpressionRef expected, BinaryenExpressionRef replacement, BinaryenType type) {
-  auto* ret = Builder(*((Module*)module)).makeAtomicCmpxchg(bytes, offset, (Expression*)ptr, (Expression*)expected, (Expression*)replacement, WasmType(type));
+  auto* ret = Builder(*((Module*)module)).makeAtomicCmpxchg(bytes, offset, (Expression*)ptr, (Expression*)expected, (Expression*)replacement, Type(type));
 
   if (tracing) {
     auto id = noteExpression(ret);
@@ -809,7 +831,7 @@ BinaryenExpressionRef BinaryenAtomicCmpxchg(BinaryenModuleRef module, BinaryenIn
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenAtomicWait(BinaryenModuleRef module, BinaryenExpressionRef ptr, BinaryenExpressionRef expected, BinaryenExpressionRef timeout, BinaryenType expectedType) {
-  auto* ret = Builder(*((Module*)module)).makeAtomicWait((Expression*)ptr, (Expression*)expected, (Expression*)timeout, WasmType(expectedType));
+  auto* ret = Builder(*((Module*)module)).makeAtomicWait((Expression*)ptr, (Expression*)expected, (Expression*)timeout, Type(expectedType), 0);
 
   if (tracing) {
     auto id = noteExpression(ret);
@@ -819,7 +841,7 @@ BinaryenExpressionRef BinaryenAtomicWait(BinaryenModuleRef module, BinaryenExpre
   return static_cast<Expression*>(ret);
 }
 BinaryenExpressionRef BinaryenAtomicWake(BinaryenModuleRef module, BinaryenExpressionRef ptr, BinaryenExpressionRef wakeCount) {
-  auto* ret = Builder(*((Module*)module)).makeAtomicWake((Expression*)ptr, (Expression*)wakeCount);
+  auto* ret = Builder(*((Module*)module)).makeAtomicWake((Expression*)ptr, (Expression*)wakeCount, 0);
 
   if (tracing) {
     auto id = noteExpression(ret);
@@ -835,7 +857,7 @@ BinaryenExpressionId BinaryenExpressionGetId(BinaryenExpressionRef expr) {
   if (tracing) {
     std::cout << "  BinaryenExpressionGetId(expressions[" << expressions[expr] << "]);\n";
   }
-  
+
   return ((Expression*)expr)->_id;
 }
 BinaryenType BinaryenExpressionGetType(BinaryenExpressionRef expr) {
@@ -1635,7 +1657,7 @@ BinaryenFunctionRef BinaryenAddFunction(BinaryenModuleRef module, const char* na
   ret->result = functionType->result;
   ret->params = functionType->params;
   for (BinaryenIndex i = 0; i < numVarTypes; i++) {
-    ret->vars.push_back(WasmType(varTypes[i]));
+    ret->vars.push_back(Type(varTypes[i]));
   }
   ret->body = (Expression*)body;
 
@@ -1673,7 +1695,7 @@ BinaryenGlobalRef BinaryenAddGlobal(BinaryenModuleRef module, const char* name,
   auto* wasm = (Module*)module;
   auto* ret = new Global();
   ret->name = name;
-  ret->type = WasmType(type);
+  ret->type = Type(type);
   ret->mutable_ = !!mutable_;
   ret->init = (Expression*)init;
   wasm->addGlobal(ret);
@@ -1756,7 +1778,7 @@ BinaryenImportRef BinaryenAddGlobalImport(BinaryenModuleRef module, const char*
   ret->name = internalName;
   ret->module = externalModuleName;
   ret->base = externalBaseName;
-  ret->globalType = WasmType(globalType);
+  ret->globalType = Type(globalType);
   ret->kind = ExternalKind::Global;
   wasm->addImport(ret);
   return ret;
@@ -2013,9 +2035,7 @@ void BinaryenModuleOptimize(BinaryenModuleRef module) {
 
   Module* wasm = (Module*)module;
   PassRunner passRunner(wasm);
-  passRunner.options.optimizeLevel = optimizeLevel;
-  passRunner.options.shrinkLevel = shrinkLevel;
-  passRunner.options.debugInfo = debugInfo;
+  passRunner.options = globalPassOptions;
   passRunner.addDefaultOptimizationPasses();
   passRunner.run();
 }
@@ -2025,7 +2045,7 @@ int BinaryenGetOptimizeLevel() {
     std::cout << "  BinaryenGetOptimizeLevel();\n";
   }
 
-  return optimizeLevel;
+  return globalPassOptions.optimizeLevel;
 }
 
 void BinaryenSetOptimizeLevel(int level) {
@@ -2033,7 +2053,7 @@ void BinaryenSetOptimizeLevel(int level) {
     std::cout << "  BinaryenSetOptimizeLevel(" << level << ");\n";
   }
 
-  optimizeLevel = level;
+  globalPassOptions.optimizeLevel = level;
 }
 
 int BinaryenGetShrinkLevel() {
@@ -2041,7 +2061,7 @@ int BinaryenGetShrinkLevel() {
     std::cout << "  BinaryenGetShrinkLevel();\n";
   }
 
-  return shrinkLevel;
+  return globalPassOptions.shrinkLevel;
 }
 
 void BinaryenSetShrinkLevel(int level) {
@@ -2049,7 +2069,7 @@ void BinaryenSetShrinkLevel(int level) {
     std::cout << "  BinaryenSetShrinkLevel(" << level << ");\n";
   }
 
-  shrinkLevel = level;
+  globalPassOptions.shrinkLevel = level;
 }
 
 int BinaryenGetDebugInfo() {
@@ -2057,7 +2077,7 @@ int BinaryenGetDebugInfo() {
     std::cout << "  BinaryenGetDebugInfo();\n";
   }
 
-  return debugInfo;
+  return globalPassOptions.debugInfo;
 }
 
 void BinaryenSetDebugInfo(int on) {
@@ -2065,7 +2085,7 @@ void BinaryenSetDebugInfo(int on) {
     std::cout << "  BinaryenSetDebugInfo(" << on << ");\n";
   }
 
-  debugInfo = bool(on);
+  globalPassOptions.debugInfo = on != 0;
 }
 
 void BinaryenModuleRunPasses(BinaryenModuleRef module, const char **passes, BinaryenIndex numPasses) {
@@ -2083,9 +2103,7 @@ void BinaryenModuleRunPasses(BinaryenModuleRef module, const char **passes, Bina
 
   Module* wasm = (Module*)module;
   PassRunner passRunner(wasm);
-  passRunner.options.optimizeLevel = optimizeLevel;
-  passRunner.options.shrinkLevel = shrinkLevel;
-  passRunner.options.debugInfo = debugInfo;
+  passRunner.options = globalPassOptions;
   for (BinaryenIndex i = 0; i < numPasses; i++) {
     passRunner.add(passes[i]);
   }
@@ -2099,24 +2117,75 @@ void BinaryenModuleAutoDrop(BinaryenModuleRef module) {
 
   Module* wasm = (Module*)module;
   PassRunner passRunner(wasm);
-  passRunner.options.debugInfo = debugInfo;
+  passRunner.options = globalPassOptions;
   passRunner.add<AutoDrop>();
   passRunner.run();
 }
 
+static BinaryenBufferSizes writeModule(BinaryenModuleRef module, char* output, size_t outputSize, const char* sourceMapUrl, char* sourceMap, size_t sourceMapSize) {
+  Module* wasm = (Module*)module;
+  BufferWithRandomAccess buffer(false);
+  WasmBinaryWriter writer(wasm, buffer, false);
+  writer.setNamesSection(globalPassOptions.debugInfo);
+  std::ostringstream os;
+  if (sourceMapUrl) {
+    writer.setSourceMap(&os, sourceMapUrl);
+  }
+  writer.write();
+  size_t bytes = std::min(buffer.size(), outputSize);
+  std::copy_n(buffer.begin(), bytes, output);
+  size_t sourceMapBytes = 0;
+  if (sourceMapUrl) {
+    auto str = os.str();
+    sourceMapBytes = std::min(str.length(), sourceMapSize);
+    std::copy_n(str.c_str(), sourceMapBytes, sourceMap);
+  }
+  return { bytes, sourceMapBytes };
+}
+
 size_t BinaryenModuleWrite(BinaryenModuleRef module, char* output, size_t outputSize) {
   if (tracing) {
     std::cout << "  // BinaryenModuleWrite\n";
   }
 
+  return writeModule((Module*)module, output, outputSize, nullptr, nullptr, 0).outputBytes;
+}
+
+BinaryenBufferSizes BinaryenModuleWriteWithSourceMap(BinaryenModuleRef module, const char* url, char* output, size_t outputSize, char* sourceMap, size_t sourceMapSize) {
+  if (tracing) {
+    std::cout << "  // BinaryenModuleWriteWithSourceMap\n";
+  }
+
+  assert(url);
+  assert(sourceMap);
+  return writeModule((Module*)module, output, outputSize, url, sourceMap, sourceMapSize);
+}
+
+BinaryenModuleAllocateAndWriteResult BinaryenModuleAllocateAndWrite(BinaryenModuleRef module, const char* sourceMapUrl) {
+  if (tracing) {
+    std::cout << " // BinaryenModuleAllocateAndWrite(the_module, ";
+    traceNameOrNULL(sourceMapUrl);
+    std::cout << ");\n";
+  }
+
   Module* wasm = (Module*)module;
   BufferWithRandomAccess buffer(false);
   WasmBinaryWriter writer(wasm, buffer, false);
-  writer.setNamesSection(debugInfo);
+  writer.setNamesSection(globalPassOptions.debugInfo);
+  std::ostringstream os;
+  if (sourceMapUrl) {
+    writer.setSourceMap(&os, sourceMapUrl);
+  }
   writer.write();
-  size_t bytes = std::min(buffer.size(), outputSize);
-  std::copy_n(buffer.begin(), bytes, output);
-  return bytes;
+  void* binary = malloc(buffer.size());
+  std::copy_n(buffer.begin(), buffer.size(), static_cast<char*>(binary));
+  char* sourceMap = nullptr;
+  if (sourceMapUrl) {
+    auto str = os.str();
+    sourceMap = (char*)malloc(str.length() + 1);
+    std::copy_n(str.c_str(), str.length() + 1, sourceMap);
+  }
+  return { binary, buffer.size(), sourceMap };
 }
 
 BinaryenModuleRef BinaryenModuleRead(char* input, size_t inputSize) {
@@ -2148,6 +2217,26 @@ void BinaryenModuleInterpret(BinaryenModuleRef module) {
   ModuleInstance instance(*wasm, &interface);
 }
 
+BinaryenIndex BinaryenModuleAddDebugInfoFileName(BinaryenModuleRef module, const char* filename) {
+  if (tracing) {
+    std::cout << "  BinaryenModuleAddDebugInfoFileName(the_module, \"" << filename << "\");\n";
+  }
+
+  Module* wasm = (Module*)module;
+  BinaryenIndex index = wasm->debugInfoFileNames.size();
+  wasm->debugInfoFileNames.push_back(filename);
+  return index;
+}
+
+const char* BinaryenModuleGetDebugInfoFileName(BinaryenModuleRef module, BinaryenIndex index) {
+  if (tracing) {
+    std::cout << "  BinaryenModuleGetDebugInfoFileName(the_module, \"" << index << "\");\n";
+  }
+
+  Module* wasm = (Module*)module;
+  return index < wasm->debugInfoFileNames.size() ? wasm->debugInfoFileNames.at(index).c_str() : nullptr;
+}
+
 //
 // ======== FunctionType Operations ========
 //
@@ -2254,9 +2343,7 @@ void BinaryenFunctionOptimize(BinaryenFunctionRef func, BinaryenModuleRef module
 
   Module* wasm = (Module*)module;
   PassRunner passRunner(wasm);
-  passRunner.options.optimizeLevel = optimizeLevel;
-  passRunner.options.shrinkLevel = shrinkLevel;
-  passRunner.options.debugInfo = debugInfo;
+  passRunner.options = globalPassOptions;
   passRunner.addDefaultOptimizationPasses();
   passRunner.runOnFunction((Function*)func);
 }
@@ -2275,14 +2362,27 @@ void BinaryenFunctionRunPasses(BinaryenFunctionRef func, BinaryenModuleRef modul
 
   Module* wasm = (Module*)module;
   PassRunner passRunner(wasm);
-  passRunner.options.optimizeLevel = optimizeLevel;
-  passRunner.options.shrinkLevel = shrinkLevel;
-  passRunner.options.debugInfo = debugInfo;
+  passRunner.options = globalPassOptions;
   for (BinaryenIndex i = 0; i < numPasses; i++) {
     passRunner.add(passes[i]);
   }
   passRunner.runOnFunction((Function*)func);
 }
+void BinaryenFunctionSetDebugLocation(BinaryenFunctionRef func, BinaryenExpressionRef expr, BinaryenIndex fileIndex, BinaryenIndex lineNumber, BinaryenIndex columnNumber) {
+  if (tracing) {
+    std::cout << "  BinaryenFunctionSetDebugLocation(functions[" << functions[func] << "], expressions[" << expressions[expr] << "], " << fileIndex << ", " << lineNumber << ", " << columnNumber << ");\n";
+  }
+
+  auto* fn = (Function*)func;
+  auto* ex = (Expression*)expr;
+
+  Function::DebugLocation loc;
+  loc.fileIndex = fileIndex;
+  loc.lineNumber = lineNumber;
+  loc.columnNumber = columnNumber;
+
+  fn->debugLocations[ex] = loc;
+}
 
 //
 // =========== Import operations ===========
@@ -2482,9 +2582,9 @@ BinaryenFunctionTypeRef BinaryenGetFunctionTypeBySignature(BinaryenModuleRef mod
 
   auto* wasm = (Module*)module;
   FunctionType test;
-  test.result = WasmType(result);
+  test.result = Type(result);
   for (BinaryenIndex i = 0; i < numParams; i++) {
-    test.params.push_back(WasmType(paramTypes[i]));
+    test.params.push_back(Type(paramTypes[i]));
   }
 
   // Lock. Guard against reading the list while types are being added.
