@@ -48,7 +48,6 @@
 #include <stdint.h>
 
 #include "compiler-support.h"
-#include "support/defaults.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -155,6 +154,8 @@ typedef void* BinaryenFunctionTypeRef;
 // Add a new function type. This is thread-safe.
 // Note: name can be NULL, in which case we auto-generate a name
 BinaryenFunctionTypeRef BinaryenAddFunctionType(BinaryenModuleRef module, const char* name, BinaryenType result, BinaryenType* paramTypes, BinaryenIndex numParams);
+// Removes a function type.
+void BinaryenRemoveFunctionType(BinaryenModuleRef module, const char* name);
 
 // Literals. These are passed by value.
 
@@ -234,6 +235,11 @@ BinaryenOp BinaryenPromoteFloat32(void);
 BinaryenOp BinaryenDemoteFloat64(void);
 BinaryenOp BinaryenReinterpretInt32(void);
 BinaryenOp BinaryenReinterpretInt64(void);
+BinaryenOp BinaryenExtendS8Int32(void);
+BinaryenOp BinaryenExtendS16Int32(void);
+BinaryenOp BinaryenExtendS8Int64(void);
+BinaryenOp BinaryenExtendS16Int64(void);
+BinaryenOp BinaryenExtendS32Int64(void);
 BinaryenOp BinaryenAddInt32(void);
 BinaryenOp BinaryenSubInt32(void);
 BinaryenOp BinaryenMulInt32(void);
@@ -698,6 +704,30 @@ void BinaryenModuleAutoDrop(BinaryenModuleRef module);
 // @return how many bytes were written. This will be less than or equal to outputSize
 size_t BinaryenModuleWrite(BinaryenModuleRef module, char* output, size_t outputSize);
 
+typedef struct BinaryenBufferSizes {
+  size_t outputBytes;
+  size_t sourceMapBytes;
+} BinaryenBufferSizes;
+
+// Serialize a module into binary form including its source map. Uses the currently set
+// global debugInfo option.
+// @returns how many bytes were written. This will be less than or equal to outputSize
+BinaryenBufferSizes BinaryenModuleWriteWithSourceMap(BinaryenModuleRef module, const char* url, char* output, size_t outputSize, char* sourceMap, size_t sourceMapSize);
+
+// Result structure of BinaryenModuleAllocateAndWrite. Contained buffers have been allocated
+// using malloc() and the user is expected to free() them manually once not needed anymore.
+typedef struct BinaryenModuleAllocateAndWriteResult {
+  void* binary;
+  size_t binaryBytes;
+  char* sourceMap;
+} BinaryenModuleAllocateAndWriteResult;
+
+// Serializes a module into binary form, optionally including its source map if
+// sourceMapUrl has been specified. Uses the currently set global debugInfo option.
+// Differs from BinaryenModuleWrite in that it implicitly allocates appropriate buffers
+// using malloc(), and expects the user to free() them manually once not needed anymore.
+BinaryenModuleAllocateAndWriteResult BinaryenModuleAllocateAndWrite(BinaryenModuleRef module, const char* sourceMapUrl);
+
 // Deserialize a module from binary form.
 BinaryenModuleRef BinaryenModuleRead(char* input, size_t inputSize);
 
@@ -706,6 +736,13 @@ BinaryenModuleRef BinaryenModuleRead(char* input, size_t inputSize);
 // and then destroying the instance.
 void BinaryenModuleInterpret(BinaryenModuleRef module);
 
+// Adds a debug info file name to the module and returns its index.
+BinaryenIndex BinaryenModuleAddDebugInfoFileName(BinaryenModuleRef module, const char* filename);
+
+// Gets the name of the debug info file at the specified index. Returns `NULL` if it
+// does not exist.
+const char* BinaryenModuleGetDebugInfoFileName(BinaryenModuleRef module, BinaryenIndex index);
+
 //
 // ======== FunctionType Operations ========
 //
@@ -748,6 +785,9 @@ void BinaryenFunctionOptimize(BinaryenFunctionRef func, BinaryenModuleRef module
 // optimize and shrink level.
 void BinaryenFunctionRunPasses(BinaryenFunctionRef func, BinaryenModuleRef module, const char **passes, BinaryenIndex numPasses);
 
+// Sets the debug location of the specified `Expression` within the specified `Function`.
+void BinaryenFunctionSetDebugLocation(BinaryenFunctionRef func, BinaryenExpressionRef expr, BinaryenIndex fileIndex, BinaryenIndex lineNumber, BinaryenIndex columnNumber);
+
 //
 // ========== Import Operations ==========
 //
