@@ -115,6 +115,11 @@ Module['PromoteFloat32'] = Module['_BinaryenPromoteFloat32']();
 Module['DemoteFloat64'] = Module['_BinaryenDemoteFloat64']();
 Module['ReinterpretInt32'] = Module['_BinaryenReinterpretInt32']();
 Module['ReinterpretInt64'] = Module['_BinaryenReinterpretInt64']();
+Module['ExtendS8Int32'] = Module['_BinaryenExtendS8Int32']();
+Module['ExtendS16Int32'] = Module['_BinaryenExtendS16Int32']();
+Module['ExtendS8Int64'] = Module['_BinaryenExtendS8Int64']();
+Module['ExtendS16Int64'] = Module['_BinaryenExtendS16Int64']();
+Module['ExtendS32Int64'] = Module['_BinaryenExtendS32Int64']();
 Module['AddInt32'] = Module['_BinaryenAddInt32']();
 Module['SubInt32'] = Module['_BinaryenSubInt32']();
 Module['MulInt32'] = Module['_BinaryenMulInt32']();
@@ -286,8 +291,8 @@ Module['Module'] = function(module) {
   // need to make their own Literals, as the C API handles them by value,
   // which means we would leak them. Instead, this is the only API that
   // accepts Literals, so fuse it with Literal creation
-  var literal = _malloc(16); // a single literal in memory. the LLVM C ABI
-                             // makes us pass pointers to this.
+  var temp = _malloc(16); // a single literal in memory. the LLVM C ABI
+                          // makes us pass pointers to this.
 
   this['i32'] = {
     'load': function(offset, align, ptr) {
@@ -315,8 +320,8 @@ Module['Module'] = function(module) {
       return Module['_BinaryenStore'](module, 2, offset, align, ptr, value, Module['i32']);
     },
     'const': function(x) {
-      Module['_BinaryenLiteralInt32'](literal, x);
-      return Module['_BinaryenConst'](module, literal);
+      Module['_BinaryenLiteralInt32'](temp, x);
+      return Module['_BinaryenConst'](module, temp);
     },
     'clz': function(value) {
       return Module['_BinaryenUnary'](module, Module['ClzInt32'], value);
@@ -349,6 +354,12 @@ Module['Module'] = function(module) {
     'reinterpret': function(value) {
       return Module['_BinaryenUnary'](module, Module['ReinterpretFloat32'], value);
     },
+    'extend8_s': function(value) {
+      return Module['_BinaryenUnary'](module, Module['ExtendS8Int32'], value);
+    },
+    'extend16_s': function(value) {
+      return Module['_BinaryenUnary'](module, Module['ExtendS16Int32'], value);
+    },
     'wrap': function(value) {
       return Module['_BinaryenUnary'](module, Module['WrapInt64'], value);
     },
@@ -556,8 +567,8 @@ Module['Module'] = function(module) {
       return Module['_BinaryenStore'](module, 4, offset, align, ptr, value, Module['i64']);
     },
     'const': function(x, y) {
-      Module['_BinaryenLiteralInt64'](literal, x, y);
-      return Module['_BinaryenConst'](module, literal);
+      Module['_BinaryenLiteralInt64'](temp, x, y);
+      return Module['_BinaryenConst'](module, temp);
     },
     'clz': function(value) {
       return Module['_BinaryenUnary'](module, Module['ClzInt64'], value);
@@ -590,6 +601,15 @@ Module['Module'] = function(module) {
     'reinterpret': function(value) {
       return Module['_BinaryenUnary'](module, Module['ReinterpretFloat64'], value);
     },
+    'extend8_s': function(value) {
+      return Module['_BinaryenUnary'](module, Module['ExtendS8Int64'], value);
+    },
+    'extend16_s': function(value) {
+      return Module['_BinaryenUnary'](module, Module['ExtendS16Int64'], value);
+    },
+    'extend32_s': function(value) {
+      return Module['_BinaryenUnary'](module, Module['ExtendS32Int64'], value);
+    },
     'extend_s': function(value) {
       return Module['_BinaryenUnary'](module, Module['ExtendSInt32'], value);
     },
@@ -802,12 +822,12 @@ Module['Module'] = function(module) {
       return Module['_BinaryenStore'](module, 4, offset, align, ptr, value, Module['f32']);
     },
     'const': function(x) {
-      Module['_BinaryenLiteralFloat32'](literal, x);
-      return Module['_BinaryenConst'](module, literal);
+      Module['_BinaryenLiteralFloat32'](temp, x);
+      return Module['_BinaryenConst'](module, temp);
     },
     'const_bits': function(x) {
-      Module['_BinaryenLiteralFloat32Bits'](literal, x);
-      return Module['_BinaryenConst'](module, literal);
+      Module['_BinaryenLiteralFloat32Bits'](temp, x);
+      return Module['_BinaryenConst'](module, temp);
     },
     'neg': function(value) {
       return Module['_BinaryenUnary'](module, Module['NegFloat32'], value);
@@ -901,12 +921,12 @@ Module['Module'] = function(module) {
       return Module['_BinaryenStore'](module, 8, offset, align, ptr, value, Module['f64']);
     },
     'const': function(x) {
-      Module['_BinaryenLiteralFloat64'](literal, x);
-      return Module['_BinaryenConst'](module, literal);
+      Module['_BinaryenLiteralFloat64'](temp, x);
+      return Module['_BinaryenConst'](module, temp);
     },
     'const_bits': function(x, y) {
-      Module['_BinaryenLiteralFloat64Bits'](literal, x, y);
-      return Module['_BinaryenConst'](module, literal);
+      Module['_BinaryenLiteralFloat64Bits'](temp, x, y);
+      return Module['_BinaryenConst'](module, temp);
     },
     'neg': function(value) {
       return Module['_BinaryenUnary'](module, Module['NegFloat64'], value);
@@ -1032,6 +1052,11 @@ Module['Module'] = function(module) {
                                                            i32sToStack(paramTypes), paramTypes.length);
     });
   };
+  this['removeFunctionType'] = function(name) {
+    return preserveStack(function () {
+      return Module['_BinaryenRemoveFunctionType'](module, strToStack(name));
+    });
+  };
   this['addFunction'] = function(name, functionType, varTypes, body) {
     return preserveStack(function() {
       return Module['_BinaryenAddFunction'](module, strToStack(name), functionType, i32sToStack(varTypes), varTypes.length, body);
@@ -1160,7 +1185,7 @@ Module['Module'] = function(module) {
     return Module['_BinaryenModuleOptimize'](module);
   };
   this['optimizeFunction'] = function(func) {
-    if (typeof func === "string") func = this['getFunction'](func);
+    if (typeof func === 'string') func = this['getFunction'](func);
     return Module['_BinaryenFunctionOptimize'](func, module);
   };
   this['runPasses'] = function(passes) {
@@ -1171,7 +1196,7 @@ Module['Module'] = function(module) {
     });
   };
   this['runPassesOnFunction'] = function(func, passes) {
-    if (typeof func === "string") func = this['getFunction'](func);
+    if (typeof func === 'string') func = this['getFunction'](func);
     return preserveStack(function() {
       return Module['_BinaryenFunctionRunPasses'](func, module, i32sToStack(
         passes.map(strToStack)
@@ -1184,17 +1209,38 @@ Module['Module'] = function(module) {
   this['dispose'] = function() {
     Module['_BinaryenModuleDispose'](module);
   };
-  var MAX = 1024*1024; // TODO: fix this hard-wired limit
-  var writeBuffer = null;
-  this['emitBinary'] = function() {
-    if (!writeBuffer) writeBuffer = _malloc(MAX);
-    var bytes = Module['_BinaryenModuleWrite'](module, writeBuffer, MAX);
-    assert(bytes < MAX, 'FIXME: hardcoded limit on module size'); // we should not use the whole buffer
-    return new Uint8Array(HEAPU8.subarray(writeBuffer, writeBuffer + bytes));
+  this['emitBinary'] = function(sourceMapUrl) {
+    return preserveStack(function() {
+      Module['_BinaryenModuleAllocateAndWrite'](temp, module, strToStack(sourceMapUrl));
+      var binaryPtr    = HEAPU32[ temp >>> 2     ];
+      var binaryBytes  = HEAPU32[(temp >>> 2) + 1];
+      var sourceMapPtr = HEAPU32[(temp >>> 2) + 2];
+      try {
+        var buffer = new Uint8Array(binaryBytes);
+        buffer.set(HEAPU8.subarray(binaryPtr, binaryPtr + binaryBytes));
+        return typeof sourceMapUrl === 'undefined'
+          ? buffer
+          : { 'binary': buffer, 'sourceMap': Pointer_stringify(sourceMapPtr) };
+      } finally {
+        _free(binaryPtr);
+        if (sourceMapPtr) _free(sourceMapPtr);
+      }
+    });
   };
   this['interpret'] = function() {
     return Module['_BinaryenModuleInterpret'](module);
   };
+  this['addDebugInfoFileName'] = function(filename) {
+    return preserveStack(function() {
+      return Module['_BinaryenModuleAddDebugInfoFileName'](module, strToStack(filename));
+    });
+  };
+  this['getDebugInfoFileName'] = function(index) {
+    return Pointer_stringify(Module['_BinaryenModuleGetDebugInfoFileName'](module, index));
+  };
+  this['setDebugLocation'] = function(func, expr, fileIndex, lineNumber, columnNumber) {
+    return Module['_BinaryenFunctionSetDebugLocation'](func, expr, fileIndex, lineNumber, columnNumber);
+  };
 };
 
 // 'Relooper' interface
@@ -1563,19 +1609,3 @@ Module['setDebugInfo'] = function(on) {
 Module['setAPITracing'] = function(on) {
   return Module['_BinaryenSetAPITracing'](on);
 };
-
-// Instantiates a new unique instance of the API with its own memory etc.
-Module['instantiate'] = instantiate;
-return Module;
-
-} // end of instantiate
-
-// Module loader code borrowed from webpack
-if (typeof exports === 'object' && typeof module === 'object')
-  module.exports = instantiate();
-else if (typeof define === 'function' && define['amd'])
-  define([], instantiate);
-else if (typeof exports === 'object')
-  exports['Binaryen'] = instantiate();
-else
-  (typeof self !== "undefined" ? self : this)['Binaryen'] = instantiate();
