@@ -103,7 +103,7 @@ void CoalesceLocals::increaseBackEdgePriorities() {
       auto* arrivingBlock = in[i];
       if (arrivingBlock->out.size() > 1) continue; // we just want unconditional branches to the loop top, true phi fragments
       for (auto& action : arrivingBlock->contents.actions) {
-        if (action.what == Action::Set) {
+        if (action.isSet()) {
           auto* set = (*action.origin)->cast<SetLocal>();
           if (auto* get = getCopy(set)) {
             // this is indeed a copy, add to the cost (default cost is 2, so this adds 50%, and can mostly break ties)
@@ -117,7 +117,7 @@ void CoalesceLocals::increaseBackEdgePriorities() {
 
 void CoalesceLocals::calculateInterferences() {
   interferences.resize(numLocals * numLocals);
-  std::fill(interferences.begin(), interferences.end(), 0);
+  std::fill(interferences.begin(), interferences.end(), false);
   for (auto& curr : basicBlocks) {
     if (liveBlocks.count(curr.get()) == 0) continue; // ignore dead blocks
     // everything coming in might interfere, as it might come from a different block
@@ -201,13 +201,13 @@ void CoalesceLocals::pickIndicesFromOrder(std::vector<Index>& order, std::vector
   }
 #endif
   // TODO: take into account distribution (99-1 is better than 50-50 with two registers, for gzip)
-  std::vector<WasmType> types;
+  std::vector<Type> types;
   std::vector<bool> newInterferences; // new index * numLocals => list of all interferences of locals merged to it
   std::vector<uint8_t> newCopies; // new index * numLocals => list of all copies of locals merged to it
   indices.resize(numLocals);
   types.resize(numLocals);
   newInterferences.resize(numLocals * numLocals);
-  std::fill(newInterferences.begin(), newInterferences.end(), 0);
+  std::fill(newInterferences.begin(), newInterferences.end(), false);
   auto numParams = getFunction()->getNumParams();
   newCopies.resize(numParams * numLocals); // start with enough room for the params
   std::fill(newCopies.begin(), newCopies.end(), 0);
