@@ -24,6 +24,7 @@
 #include <pass.h>
 #include <wasm-s-parser.h>
 #include <support/threads.h>
+#include <ir/abstract.h>
 #include <ir/utils.h>
 #include <ir/cost.h>
 #include <ir/effects.h>
@@ -116,7 +117,7 @@ struct Match {
       if (!call || call->operands.size() != 1 || call->operands[0]->type != i32 || !call->operands[0]->is<Const>()) return false;
       Index index = call->operands[0]->cast<Const>()->value.geti32();
       // handle our special functions
-      auto checkMatch = [&](WasmType type) {
+      auto checkMatch = [&](Type type) {
         if (type != none && subSeen->type != type) return false;
         while (index >= wildcards.size()) {
           wildcards.push_back(nullptr);
@@ -338,7 +339,7 @@ struct LocalScanner : PostWalker<LocalScanner> {
     return getBitsForType(get->type);
   }
 
-  Index getBitsForType(WasmType type) {
+  Index getBitsForType(Type type) {
     switch (type) {
       case i32: return 32;
       case i64: return 64;
@@ -506,8 +507,46 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
           }
         }
         // note that both left and right may be consts, but then we let precompute compute the constant result
-      } else if (binary->op == AddInt32 || binary->op == SubInt32) {
-        return optimizeAddedConstants(binary);
+      } else if (binary->op == AddInt32) {
+        // try to get rid of (0 - ..), that is, a zero only used to negate an
+        // int. an add of a subtract can be flipped in order to remove it:
+        //   (i32.add
+        //     (i32.sub
+        //       (i32.const 0)
+        //       X
+        //     )
+        //     Y
+        //   )
+        // =>
+        //   (i32.sub
+        //     Y
+        //     X
+        //   )
+        if (auto* sub = binary->left->dynCast<Binary>()) {
+          if (sub->op == SubInt32) {
+            if (auto* subZero = sub->left->dynCast<Const>()) {
+              if (subZero->value.geti32() == 0) {
+                sub->left = binary->right;
+                return sub;
+              }
+            }
+          }
+        }
+        if (auto* sub = binary->right->dynCast<Binary>()) {
+          if (sub->op == SubInt32) {
+            if (auto* subZero = sub->left->dynCast<Const>()) {
+              if (subZero->value.geti32() == 0) {
+                sub->left = binary->left;
+                return sub;
+              }
+            }
+          }
+        }
+        auto* ret = optimizeAddedConstants(binary);
+        if (ret) return ret;
+      } else if (binary->op == SubInt32) {
+        auto* ret = optimizeAddedConstants(binary);
+        if (ret) return ret;
       }
       // a bunch of operations on a constant right side can be simplified
       if (auto* right = binary->right->dynCast<Const>()) {
@@ -517,7 +556,7 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
           if (mask == -1) {
             return binary->left;
           }
-          // small loads do not need to be masted, the load itself masks
+          // small loads do not need to be masked, the load itself masks
           if (auto* load = binary->left->dynCast<Load>()) {
             if ((load->bytes == 1 && mask == 0xff) ||
                 (load->bytes == 2 && mask == 0xffff)) {
@@ -531,12 +570,9 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
             }
           }
         }
-        // some operations have no effect TODO: many more
-        if (right->value == Literal(int32_t(0))) {
-          if (binary->op == ShlInt32 || binary->op == ShrUInt32 || binary->op == ShrSInt32) {
-            return binary->left;
-          }
-        }
+        // some math operations have trivial results
+        Expression* ret = optimizeWithConstantOnRight(binary);
+        if (ret) return ret;
         // the square of some operations can be merged
         if (auto* left = binary->left->dynCast<Binary>()) {
           if (left->op == binary->op) {
@@ -561,6 +597,22 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
             }
           }
         }
+        // math operations on a constant power of 2 right side can be optimized
+        if (right->type == i32) {
+          uint32_t c = right->value.geti32();
+          if (IsPowerOf2(c)) {
+            if (binary->op == MulInt32) {
+              return optimizePowerOf2Mul(binary, c);
+            } else if (binary->op == RemUInt32) {
+              return optimizePowerOf2URem(binary, c);
+            }
+          }
+        }
+      }
+      // a bunch of operations on a constant left side can be simplified
+      if (binary->left->is<Const>()) {
+        Expression* ret = optimizeWithConstantOnLeft(binary);
+        if (ret) return ret;
       }
       // bitwise operations
       if (binary->op == AndInt32) {
@@ -589,7 +641,18 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
       }
       // for and and or, we can potentially conditionalize
       if (binary->op == AndInt32 || binary->op == OrInt32) {
-        return conditionalizeExpensiveOnBitwise(binary);
+        auto* ret = conditionalizeExpensiveOnBitwise(binary);
+        if (ret) return ret;
+      }
+      // relation/comparisons allow for math optimizations
+      if (binary->isRelational()) {
+        auto* ret = optimizeRelational(binary);
+        if (ret) return ret;
+      }
+      // finally, try more expensive operations on the binary
+      if (!EffectAnalyzer(getPassOptions(), binary->left).hasSideEffects() &&
+          ExpressionAnalyzer::equal(binary->left, binary->right)) {
+        return optimizeBinaryWithEqualEffectlessChildren(binary);
       }
     } else if (auto* unary = curr->dynCast<Unary>()) {
       // de-morgan's laws
@@ -668,7 +731,7 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
             } else {
               // the types diff. as the condition is reachable, that means the if must be
               // concrete while the arm is not
-              assert(isConcreteWasmType(iff->type) && iff->ifTrue->type == unreachable);
+              assert(isConcreteType(iff->type) && iff->ifTrue->type == unreachable);
               // emit a block with a forced type
               auto* ret = builder.makeBlock();
               if (needCondition) {
@@ -693,6 +756,26 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
           std::swap(select->ifTrue, select->ifFalse);
         }
       }
+      if (auto* c = select->condition->dynCast<Const>()) {
+        // constant condition, we can just pick the right side (barring side effects)
+        if (c->value.getInteger()) {
+          if (!EffectAnalyzer(getPassOptions(), select->ifFalse).hasSideEffects()) {
+            return select->ifTrue;
+          } else {
+            // don't bother - we would need to reverse the order using a temp local, which is bad
+          }
+        } else {
+          if (!EffectAnalyzer(getPassOptions(), select->ifTrue).hasSideEffects()) {
+            return select->ifFalse;
+          } else {
+            Builder builder(*getModule());
+            return builder.makeSequence(
+              builder.makeDrop(select->ifTrue),
+              select->ifFalse
+            );
+          }
+        }
+      }
       if (ExpressionAnalyzer::equal(select->ifTrue, select->ifFalse)) {
         // sides are identical, fold
         EffectAnalyzer value(getPassOptions(), select->ifTrue);
@@ -743,7 +826,7 @@ struct OptimizeInstructions : public WalkerPass<PostWalker<OptimizeInstructions,
         } else if (auto* ext = Properties::getSignExtValue(binary)) {
           // if sign extending the exact bit size we store, we can skip the extension
           // if extending something bigger, then we just alter bits we don't save anyhow
-          if (Properties::getSignExtBits(binary) >= store->bytes * 8) {
+          if (Properties::getSignExtBits(binary) >= Index(store->bytes) * 8) {
             store->value = ext;
           }
         }
@@ -988,13 +1071,38 @@ private:
     }
   }
 
+  // Optimize a multiply by a power of two on the right, which
+  // can be a shift.
+  // This doesn't shrink code size, and VMs likely optimize it anyhow,
+  // but it's still worth doing since
+  //  * Often shifts are more common than muls.
+  //  * The constant is smaller.
+  Expression* optimizePowerOf2Mul(Binary* binary, uint32_t c) {
+    uint32_t shifts = CountTrailingZeroes(c);
+    binary->op = ShlInt32;
+    binary->right->cast<Const>()->value = Literal(int32_t(shifts));
+    return binary;
+  }
+
+  // Optimize an unsigned divide by a power of two on the right,
+  // which can be an AND mask
+  // This doesn't shrink code size, and VMs likely optimize it anyhow,
+  // but it's still worth doing since
+  //  * Usually ands are more common than urems.
+  //  * The constant is slightly smaller.
+  Expression* optimizePowerOf2URem(Binary* binary, uint32_t c) {
+    binary->op = AndInt32;
+    binary->right->cast<Const>()->value = Literal(int32_t(c - 1));
+    return binary;
+  }
+
   Expression* makeZeroExt(Expression* curr, int32_t bits) {
     Builder builder(*getModule());
     return builder.makeBinary(AndInt32, curr, builder.makeConst(Literal(Bits::lowBitMask(bits))));
   }
 
   // given an "almost" sign extend - either a proper one, or it
-  // has too many shifts left - we remove the sig extend. If there are
+  // has too many shifts left - we remove the sign extend. If there are
   // too many shifts, we split the shifts first, so this removes the
   // two sign extend shifts and adds one (smaller one)
   Expression* removeAlmostSignExt(Binary* outer) {
@@ -1019,6 +1127,190 @@ private:
     }
     return false;
   }
+
+  // optimize trivial math operations, given that the right side of a binary
+  // is a constant
+  // TODO: templatize on type?
+  Expression* optimizeWithConstantOnRight(Binary* binary) {
+    auto type = binary->right->type;
+    auto* right = binary->right->cast<Const>();
+    if (isIntegerType(type)) {
+      // operations on zero
+      if (right->value == LiteralUtils::makeLiteralFromInt32(0, type)) {
+        if (binary->op == Abstract::getBinary(type, Abstract::Shl) ||
+            binary->op == Abstract::getBinary(type, Abstract::ShrU) ||
+            binary->op == Abstract::getBinary(type, Abstract::ShrS) ||
+            binary->op == Abstract::getBinary(type, Abstract::Or) ||
+            binary->op == Abstract::getBinary(type, Abstract::Xor)) {
+          return binary->left;
+        } else if ((binary->op == Abstract::getBinary(type, Abstract::Mul) ||
+                    binary->op == Abstract::getBinary(type, Abstract::And)) &&
+                   !EffectAnalyzer(getPassOptions(), binary->left).hasSideEffects()) {
+          return binary->right;
+        }
+      }
+      // operations on all 1s
+      // TODO: shortcut method to create an all-ones?
+      if (right->value == Literal(int32_t(-1)) ||
+          right->value == Literal(int64_t(-1))) {
+        if (binary->op == Abstract::getBinary(type, Abstract::And)) {
+          return binary->left;
+        } else if (binary->op == Abstract::getBinary(type, Abstract::Or) &&
+                   !EffectAnalyzer(getPassOptions(), binary->left).hasSideEffects()) {
+          return binary->right;
+        }
+      }
+      // wasm binary encoding uses signed LEBs, which slightly favor negative
+      // numbers: -64 is more efficient than +64 etc., as well as other powers
+      // of two 7 bits etc. higher. we therefore prefer x - -64 over x + 64.
+      // in theory we could just prefer negative numbers over positive, but
+      // that can have bad effects on gzip compression (as it would mean more
+      // subtractions than the more common additions).
+      if (binary->op == Abstract::getBinary(type, Abstract::Add) ||
+          binary->op == Abstract::getBinary(type, Abstract::Sub)) {
+        auto value = right->value.getInteger();
+        if (value == 0x40 ||
+            value == 0x2000 ||
+            value == 0x100000 ||
+            value == 0x8000000 ||
+            value == 0x400000000LL ||
+            value == 0x20000000000LL ||
+            value == 0x1000000000000LL ||
+            value == 0x80000000000000LL ||
+            value == 0x4000000000000000LL) {
+          right->value = right->value.neg();
+          if (binary->op == Abstract::getBinary(type, Abstract::Add)) {
+            binary->op = Abstract::getBinary(type, Abstract::Sub);
+          } else {
+            binary->op = Abstract::getBinary(type, Abstract::Add);
+          }
+          return binary;
+        }
+      }
+    }
+    // note that this is correct even on floats with a NaN on the left,
+    // as a NaN would skip the computation and just return the NaN,
+    // and that is precisely what we do here. but, the same with -1
+    // (change to a negation) would be incorrect for that reason.
+    if (right->value == LiteralUtils::makeLiteralFromInt32(1, type)) {
+      if (binary->op == Abstract::getBinary(type, Abstract::Mul) ||
+          binary->op == Abstract::getBinary(type, Abstract::DivS) ||
+          binary->op == Abstract::getBinary(type, Abstract::DivU)) {
+        return binary->left;
+      }
+    }
+    return nullptr;
+  }
+
+  // optimize trivial math operations, given that the left side of a binary
+  // is a constant. since we canonicalize constants to the right for symmetrical
+  // operations, we only need to handle asymmetrical ones here
+  // TODO: templatize on type?
+  Expression* optimizeWithConstantOnLeft(Binary* binary) {
+    auto type = binary->left->type;
+    auto* left = binary->left->cast<Const>();
+    if (isIntegerType(type)) {
+      // operations on zero
+      if (left->value == LiteralUtils::makeLiteralFromInt32(0, type)) {
+        if ((binary->op == Abstract::getBinary(type, Abstract::Shl) ||
+             binary->op == Abstract::getBinary(type, Abstract::ShrU) ||
+             binary->op == Abstract::getBinary(type, Abstract::ShrS)) &&
+            !EffectAnalyzer(getPassOptions(), binary->right).hasSideEffects()) {
+          return binary->left;
+        }
+      }
+    }
+    return nullptr;
+  }
+
+  // integer math, even on 2s complement, allows stuff like
+  // x + 5 == 7
+  //   =>
+  //     x == 2
+  // TODO: templatize on type?
+  Expression* optimizeRelational(Binary* binary) {
+    // TODO: inequalities can also work, if the constants do not overflow
+    auto type = binary->right->type;
+    if (binary->op ==Abstract::getBinary(type, Abstract::Eq) ||
+        binary->op ==Abstract::getBinary(type, Abstract::Ne)) {
+      if (isIntegerType(binary->left->type)) {
+        if (auto* left = binary->left->dynCast<Binary>()) {
+          if (left->op == Abstract::getBinary(type, Abstract::Add) ||
+              left->op == Abstract::getBinary(type, Abstract::Sub)) {
+            if (auto* leftConst = left->right->dynCast<Const>()) {
+              if (auto* rightConst = binary->right->dynCast<Const>()) {
+                return combineRelationalConstants(binary, left, leftConst, nullptr, rightConst);
+              } else if (auto* rightBinary = binary->right->dynCast<Binary>()) {
+                if (rightBinary->op == Abstract::getBinary(type, Abstract::Add) ||
+                    rightBinary->op == Abstract::getBinary(type, Abstract::Sub)) {
+                  if (auto* rightConst = rightBinary->right->dynCast<Const>()) {
+                    return combineRelationalConstants(binary, left, leftConst, rightBinary, rightConst);
+                  }
+                }
+              }
+            }
+          }
+        }
+      }
+    }
+    return nullptr;
+  }
+
+  // given a relational binary with a const on both sides, combine the constants
+  // left is also a binary, and has a constant; right may be just a constant, in which
+  // case right is nullptr
+  Expression* combineRelationalConstants(Binary* binary, Binary* left, Const* leftConst, Binary* right, Const* rightConst) {
+    auto type = binary->right->type;
+    // we fold constants to the right
+    Literal extra = leftConst->value;
+    if (left->op == Abstract::getBinary(type, Abstract::Sub)) {
+      extra = extra.neg();
+    }
+    if (right && right->op == Abstract::getBinary(type, Abstract::Sub)) {
+      extra = extra.neg();
+    }
+    rightConst->value = rightConst->value.sub(extra);
+    binary->left = left->left;
+    return binary;
+  }
+
+  // given a binary expression with equal children and no side effects in either,
+  // we can fold various things
+  // TODO: trinaries, things like (x & (y & x)) ?
+  Expression* optimizeBinaryWithEqualEffectlessChildren(Binary* binary) {
+    // TODO add: perhaps worth doing 2*x if x is quite large?
+    switch (binary->op) {
+      case SubInt32:
+      case XorInt32:
+      case SubInt64:
+      case XorInt64: return LiteralUtils::makeZero(binary->left->type, *getModule());
+      case NeInt64:
+      case LtSInt64:
+      case LtUInt64:
+      case GtSInt64:
+      case GtUInt64:
+      case NeInt32:
+      case LtSInt32:
+      case LtUInt32:
+      case GtSInt32:
+      case GtUInt32: return LiteralUtils::makeZero(i32, *getModule());
+      case AndInt32:
+      case OrInt32:
+      case AndInt64:
+      case OrInt64:  return binary->left;
+      case EqInt32:
+      case LeSInt32:
+      case LeUInt32:
+      case GeSInt32:
+      case GeUInt32:
+      case EqInt64:
+      case LeSInt64:
+      case LeUInt64:
+      case GeSInt64:
+      case GeUInt64: return LiteralUtils::makeFromInt32(1, i32, *getModule());
+      default: return nullptr;
+    }
+  }
 };
 
 Pass *createOptimizeInstructionsPass() {
