@@ -22,10 +22,10 @@
 #include <pass.h>
 #include <wasm-builder.h>
 #include <wasm-interpreter.h>
-#include <ast_utils.h>
-#include <ast/literal-utils.h>
-#include <ast/local-graph.h>
-#include <ast/manipulation.h>
+#include <ir/utils.h>
+#include <ir/literal-utils.h>
+#include <ir/local-graph.h>
+#include <ir/manipulation.h>
 
 namespace wasm {
 
@@ -38,8 +38,15 @@ class StandaloneExpressionRunner : public ExpressionRunner<StandaloneExpressionR
   // map gets to constant values, if they are known to be constant
   GetValues& getValues;
 
+  // Whether we are trying to precompute down to an expression (which we can do on
+  // say 5 + 6) or to a value (which we can't do on a tee_local that flows a 7
+  // through it). When we want to replace the expression, we can only do so
+  // when it has no side effects. When we don't care about replacing the expression,
+  // we just want to know if it will contain a known constant.
+  bool replaceExpression;
+
 public:
-  StandaloneExpressionRunner(GetValues& getValues) : getValues(getValues) {}
+  StandaloneExpressionRunner(GetValues& getValues, bool replaceExpression) : getValues(getValues), replaceExpression(replaceExpression) {}
 
   struct NonstandaloneException {}; // TODO: use a flow with a special name, as this is likely very slow
 
@@ -69,6 +76,14 @@ public:
     return Flow(NONSTANDALONE_FLOW);
   }
   Flow visitSetLocal(SetLocal *curr) {
+    // If we don't need to replace the whole expression, see if there
+    // is a value flowing through a tee.
+    if (!replaceExpression) {
+      if (isConcreteType(curr->type)) {
+        assert(curr->isTee());
+        return visit(curr->value);
+      }
+    }
     return Flow(NONSTANDALONE_FLOW);
   }
   Flow visitGetGlobal(GetGlobal *curr) {
@@ -89,6 +104,12 @@ public:
   Flow visitAtomicCmpxchg(AtomicCmpxchg *curr) {
     return Flow(NONSTANDALONE_FLOW);
   }
+  Flow visitAtomicWait(AtomicWait *curr) {
+    return Flow(NONSTANDALONE_FLOW);
+  }
+  Flow visitAtomicWake(AtomicWake *curr) {
+    return Flow(NONSTANDALONE_FLOW);
+  }
   Flow visitHost(Host *curr) {
     return Flow(NONSTANDALONE_FLOW);
   }
@@ -109,22 +130,31 @@ struct Precompute : public WalkerPass<PostWalker<Precompute, UnifiedExpressionVi
 
   GetValues getValues;
 
+  bool worked;
+
   void doWalkFunction(Function* func) {
-    // with extra effort, we can utilize the get-set graph to precompute
-    // things that use locals that are known to be constant. otherwise,
-    // we just look at what is immediately before us
-    if (propagate) {
-      optimizeLocals(func, getModule());
-    }
-    // do the main and final walk over everything
-    WalkerPass<PostWalker<Precompute, UnifiedExpressionVisitor<Precompute>>>::doWalkFunction(func);
+    // if propagating, we may need multiple rounds: each propagation can
+    // lead to the main walk removing code, which might open up more
+    // propagation opportunities
+    do {
+      getValues.clear();
+      // with extra effort, we can utilize the get-set graph to precompute
+      // things that use locals that are known to be constant. otherwise,
+      // we just look at what is immediately before us
+      if (propagate) {
+        optimizeLocals(func);
+      }
+      // do the main walk over everything
+      worked = false;
+      super::doWalkFunction(func);
+    } while (propagate && worked);
   }
 
   void visitExpression(Expression* curr) {
     // TODO: if get_local, only replace with a constant if we don't care about size...?
     if (curr->is<Const>() || curr->is<Nop>()) return;
     // try to evaluate this into a const
-    Flow flow = precomputeFlow(curr);
+    Flow flow = precomputeExpression(curr);
     if (flow.breaking()) {
       if (flow.breakTo == NONSTANDALONE_FLOW) return;
       if (flow.breakTo == RETURN_FLOW) {
@@ -175,8 +205,9 @@ struct Precompute : public WalkerPass<PostWalker<Precompute, UnifiedExpressionVi
       return;
     }
     // this was precomputed
-    if (isConcreteWasmType(flow.value.type)) {
+    if (isConcreteType(flow.value.type)) {
       replaceCurrent(Builder(*getModule()).makeConst(flow.value));
+      worked = true;
     } else {
       ExpressionManipulator::nop(curr);
     }
@@ -188,29 +219,41 @@ struct Precompute : public WalkerPass<PostWalker<Precompute, UnifiedExpressionVi
   }
 
 private:
-  Flow precomputeFlow(Expression* curr) {
+  // Precompute an expression, returning a flow, which may be a constant
+  // (that we can replace the expression with if replaceExpression is set).
+  Flow precomputeExpression(Expression* curr, bool replaceExpression = true) {
     try {
-      return StandaloneExpressionRunner(getValues).visit(curr);
-    } catch (StandaloneExpressionRunner::NonstandaloneException& e) {
+      return StandaloneExpressionRunner(getValues, replaceExpression).visit(curr);
+    } catch (StandaloneExpressionRunner::NonstandaloneException&) {
       return Flow(NONSTANDALONE_FLOW);
     }
   }
 
+  // Precomputes the value of an expression, as opposed to the expression
+  // itself. This differs from precomputeExpression in that we care about
+  // the value the expression will have, which we cannot necessary replace
+  // the expression with. For example,
+  //  (tee_local (i32.const 1))
+  // will have value 1 which we can optimize here, but in precomputeExpression
+  // we could not do anything.
   Literal precomputeValue(Expression* curr) {
-    Flow flow = precomputeFlow(curr);
+    // Note that we set replaceExpression to false, as we just care about
+    // the value here.
+    Flow flow = precomputeExpression(curr, false /* replaceExpression */);
     if (flow.breaking()) {
       return Literal();
     }
     return flow.value;
   }
 
-  void optimizeLocals(Function* func, Module* module) {
+  // Propagates values around. Returns whether we propagated.
+  void optimizeLocals(Function* func) {
     // using the graph of get-set interactions, do a constant-propagation type
     // operation: note which sets are assigned locals, then see if that lets us
     // compute other sets as locals (since some of the gets they read may be
     // constant).
     // compute all dependencies
-    LocalGraph localGraph(func, module);
+    LocalGraph localGraph(func);
     localGraph.computeInfluences();
     // prepare the work list. we add things here that might change to a constant
     // initially, that means everything
