@@ -137,6 +137,9 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
       self->stopValueFlow();
     } else if (curr->is<Loop>()) {
       // do nothing - it's ok for values to flow out
+    } else if (auto* sw = curr->dynCast<Switch>()) {
+      self->stopFlow();
+      self->optimizeSwitch(sw);
     } else {
       // anything else stops the flow
       self->stopFlow();
@@ -171,6 +174,94 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
     loops.push_back(curr);
   }
 
+  void optimizeSwitch(Switch* curr) {
+    // if the final element is the default, we don't need it
+    while (!curr->targets.empty() && curr->targets.back() == curr->default_) {
+      curr->targets.pop_back();
+    }
+    // if the first element is the default, we can remove it by shifting everything (which
+    // does add a subtraction of a constant, but often that is worth it as the constant can
+    // be folded away and/or we remove multiple elements here)
+    Index removable = 0;
+    while (removable < curr->targets.size() && curr->targets[removable] == curr->default_) {
+      removable++;
+    }
+    if (removable > 0) {
+      for (Index i = removable; i < curr->targets.size(); i++) {
+        curr->targets[i - removable] = curr->targets[i];
+      }
+      curr->targets.resize(curr->targets.size() - removable);
+      Builder builder(*getModule());
+      curr->condition = builder.makeBinary(SubInt32,
+        curr->condition,
+        builder.makeConst(Literal(int32_t(removable)))
+      );
+    }
+    // when there isn't a value, we can do some trivial optimizations without worrying about
+    // the value being executed before the condition
+    if (curr->value) return;
+    if (curr->targets.size() == 0) {
+      // a switch with just a default always goes there
+      Builder builder(*getModule());
+      replaceCurrent(builder.makeSequence(
+        builder.makeDrop(curr->condition),
+        builder.makeBreak(curr->default_)
+      ));
+    } else if (curr->targets.size() == 1) {
+      // a switch with two options is basically an if
+      Builder builder(*getModule());
+      replaceCurrent(builder.makeIf(
+        curr->condition,
+        builder.makeBreak(curr->default_),
+        builder.makeBreak(curr->targets.front())
+      ));
+    } else {
+      // there are also some other cases where we want to convert a switch into ifs,
+      // especially if the switch is large and we are focusing on size.
+      // an especially egregious case is a switch like this:
+      // [a b b [..] b b c] with default b
+      // (which may be arrived at after we trim away excess default values on both
+      // sides). in this case, we really have 3 values in a simple form, so it is the
+      // next logical case after handling 1 and 2 values right above here.
+      // to optimize this, we must add a local + a bunch of nodes (if*2, tee, eq,
+      // get, const, break*3), so the table must be big enough for it to make sense
+
+      // How many targets we need when shrinking. This is literally the size at which
+      // the transformation begins to be smaller.
+      const uint32_t MIN_SHRINK = 13;
+      // How many targets we need when not shrinking, in which case, 2 ifs may be slower,
+      // so we do this when the table is ridiculously large for one with just 3 values
+      // in it.
+      const uint32_t MIN_GENERAL = 128;
+
+      auto shrink = getPassRunner()->options.shrinkLevel > 0;
+      if ((curr->targets.size() >= MIN_SHRINK  &&  shrink) ||
+          (curr->targets.size() >= MIN_GENERAL && !shrink)) {
+        for (Index i = 1; i < curr->targets.size() - 1; i++) {
+          if (curr->targets[i] != curr->default_) {
+            return;
+          }
+        }
+        // great, we are in that case, optimize
+        Builder builder(*getModule());
+        auto temp = builder.addVar(getFunction(), i32);
+        Expression* z;
+        replaceCurrent(z = builder.makeIf(
+          builder.makeTeeLocal(temp, curr->condition),
+          builder.makeIf(
+            builder.makeBinary(EqInt32,
+              builder.makeGetLocal(temp, i32),
+              builder.makeConst(Literal(int32_t(curr->targets.size() - 1)))
+            ),
+            builder.makeBreak(curr->targets.back()),
+            builder.makeBreak(curr->default_)
+          ),
+          builder.makeBreak(curr->targets.front())
+        ));
+      }
+    }
+  }
+
   void visitIf(If* curr) {
     if (!curr->ifFalse) {
       // if without an else. try to reduce   if (condition) br  =>  br_if (condition)
@@ -185,6 +276,8 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
       }
     }
     // TODO: if-else can be turned into a br_if as well, if one of the sides is a dead end
+    //       we handle the case of a returned value to a set_local later down, see
+    //       visitSetLocal.
   }
 
   // override scan to add a pre and a post check task to all nodes
@@ -260,7 +353,7 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
         } else {
           // this is already an if-else. if one side is a dead end, we can append to the other, if
           // there is no returned value to concern us
-          assert(!isConcreteWasmType(iff->type)); // can't be, since in the middle of a block
+          assert(!isConcreteType(iff->type)); // can't be, since in the middle of a block
 
           // ensures the first node is a block, if it isn't already, and merges in the second,
           // either as a single element or, if a block, by appending to the first block. this
@@ -275,7 +368,7 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
             if (!block || block->name.is()) {
               block = builder.makeBlock(any);
             } else {
-              assert(!isConcreteWasmType(block->type));
+              assert(!isConcreteType(block->type));
             }
             auto* other = append->dynCast<Block>();
             if (!other) {
@@ -461,7 +554,7 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
         auto& list = curr->list;
         for (Index i = 0; i < list.size(); i++) {
           auto* iff = list[i]->dynCast<If>();
-          if (!iff || !iff->ifFalse || isConcreteWasmType(iff->type)) continue; // if it lacked an if-false, it would already be a br_if, as that's the easy case
+          if (!iff || !iff->ifFalse || isConcreteType(iff->type)) continue; // if it lacked an if-false, it would already be a br_if, as that's the easy case
           auto* ifTrueBreak = iff->ifTrue->dynCast<Break>();
           if (ifTrueBreak && !ifTrueBreak->condition && canTurnIfIntoBrIf(iff->condition, ifTrueBreak->value, passOptions)) {
             // we are an if-else where the ifTrue is a break without a condition, so we can do this
@@ -482,18 +575,21 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
           }
         }
         if (list.size() >= 2) {
-          // Join adjacent br_ifs to the same target, making one br_if with
-          // a "selectified" condition that executes both.
-          if (shrink) {
-            for (Index i = 0; i < list.size() - 1; i++) {
-              auto* br1 = list[i]->dynCast<Break>();
-              // avoid unreachable brs, as they are dead code anyhow, and after merging
-              // them the outer scope could need type changes
-              if (!br1 || !br1->condition || br1->type == unreachable) continue;
-              auto* br2 = list[i + 1]->dynCast<Break>();
-              if (!br2 || !br2->condition || br2->type == unreachable) continue;
-              if (br1->name == br2->name) {
-                assert(!br1->value && !br2->value);
+          // combine/optimize adjacent br_ifs + a br (maybe _if) right after it
+          for (Index i = 0; i < list.size() - 1; i++) {
+            auto* br1 = list[i]->dynCast<Break>();
+            // avoid unreachable brs, as they are dead code anyhow, and after merging
+            // them the outer scope could need type changes
+            if (!br1 || !br1->condition || br1->type == unreachable) continue;
+            assert(!br1->value);
+            auto* br2 = list[i + 1]->dynCast<Break>();
+            if (!br2 || br1->name != br2->name) continue;
+            assert(!br2->value); // same target as previous, which has no value
+            // a br_if and then a br[_if] with the same target right after it
+            if (br2->condition) {
+              if (shrink && br2->type != unreachable) {
+                // Join adjacent br_ifs to the same target, making one br_if with
+                // a "selectified" condition that executes both.
                 if (!EffectAnalyzer(passOptions, br2->condition).hasSideEffects()) {
                   // it's ok to execute them both, do it
                   Builder builder(*getModule());
@@ -501,6 +597,10 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
                   ExpressionManipulator::nop(br2);
                 }
               }
+            } else {
+              // merge, we go there anyhow
+              Builder builder(*getModule());
+              list[i] = builder.makeDrop(br1->condition);
             }
           }
           // combine adjacent br_ifs that test the same value into a br_table,
@@ -542,7 +642,7 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
         // we may have simplified ifs enough to turn them into selects
         // this is helpful for code size, but can be a tradeoff with performance as we run both code paths
         if (!shrink) return;
-        if (curr->ifFalse && isConcreteWasmType(curr->ifTrue->type) && isConcreteWasmType(curr->ifFalse->type)) {
+        if (curr->ifFalse && isConcreteType(curr->ifTrue->type) && isConcreteType(curr->ifFalse->type)) {
           // if with else, consider turning it into a select if there is no control flow
           // TODO: estimate cost
           EffectAnalyzer condition(passOptions, curr->condition);
@@ -563,6 +663,56 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
         }
       }
 
+      void visitSetLocal(SetLocal* curr) {
+        // Undo an if return value into a local, if one arm is a br, as we
+        // prefer a br_if:
+        //  (set_local $x
+        //    (if (result i32)
+        //      (..condition..)
+        //      (br $somewhere)
+        //      (..result)
+        //    )
+        //  )
+        // =>
+        //  (br_if $somewhere
+        //    (..condition..)
+        //  )
+        //  (set_local $x
+        //    (..result)
+        //  )
+        // TODO: handle a condition in the br? need to watch for side effects
+        auto* iff = curr->value->dynCast<If>();
+        if (!iff) return;
+        if (!isConcreteType(iff->type) || !isConcreteType(iff->condition->type)) return;
+        auto tryToOptimize = [&](Expression* one, Expression* two, bool flipCondition) {
+          if (one->type == unreachable && two->type != unreachable) {
+            if (auto* br = one->dynCast<Break>()) {
+              if (ExpressionAnalyzer::isSimple(br)) {
+                // Wonderful, do it!
+                Builder builder(*getModule());
+                br->condition = iff->condition;
+                if (flipCondition) {
+                  br->condition = builder.makeUnary(EqZInt32, br->condition);
+                }
+                br->finalize();
+                curr->value = two;
+                replaceCurrent(
+                  builder.makeSequence(
+                    br,
+                    curr
+                  )
+                );
+                return true;
+              }
+            }
+          }
+          return false;
+        };
+        if (!tryToOptimize(iff->ifTrue, iff->ifFalse, false)) {
+          tryToOptimize(iff->ifFalse, iff->ifTrue, true);
+        }
+      }
+
       // (br_if)+ => br_table
       // we look for the specific pattern of
       //  (br_if ..target1..
@@ -606,7 +756,7 @@ struct RemoveUnusedBrs : public WalkerPass<PostWalker<RemoveUnusedBrs>> {
           auto* c = binary->right->dynCast<Const>();
           if (!c) return nullptr;
           uint32_t value = c->value.geti32();
-          if (value >= std::numeric_limits<int32_t>::max()) return nullptr;
+          if (value >= uint32_t(std::numeric_limits<int32_t>::max())) return nullptr;
           return br;
         };
 
