@@ -32,7 +32,7 @@
 // can get rid of those (the operation is trivial there after it sorts by use
 // frequency).
 //
-// This pass has two main options:
+// This pass has options:
 //
 //   * Tee: allow teeing, i.e., sinking a local with more than one use,
 //          and so after sinking we have a tee for the first use.
@@ -40,14 +40,20 @@
 //                internal set_locals into one on the outside,
 //                that can itself then be sunk further.
 //
+// There is also an option to disallow nesting entirely, which disallows
+// Tee and Structure from those 2 options, and also disallows any sinking
+// operation that would create nesting. This keeps the IR flat while
+// removing redundant locals.
+//
 
 #include <wasm.h>
 #include <wasm-builder.h>
 #include <wasm-traversal.h>
 #include <pass.h>
-#include <ast/count.h>
-#include <ast/effects.h>
-#include <ast/manipulation.h>
+#include <ir/count.h>
+#include <ir/effects.h>
+#include <ir/find_all.h>
+#include <ir/manipulation.h>
 
 namespace wasm {
 
@@ -72,14 +78,11 @@ struct SetLocalRemover : public PostWalker<SetLocalRemover> {
 
 // Main class
 
-struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>> {
+template<bool allowTee = true, bool allowStructure = true, bool allowNesting = true>
+struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals<allowTee, allowStructure, allowNesting>>> {
   bool isFunctionParallel() override { return true; }
 
-  Pass* create() override { return new SimplifyLocals(allowTee, allowStructure); }
-
-  bool allowTee, allowStructure;
-
-  SimplifyLocals(bool allowTee, bool allowStructure) : allowTee(allowTee), allowStructure(allowStructure) {}
+  Pass* create() override { return new SimplifyLocals<allowTee, allowStructure, allowNesting>(); }
 
   // information for a set_local we can sink
   struct SinkableInfo {
@@ -126,7 +129,7 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
   // local => # of get_locals for it
   GetLocalCounter getCounter;
 
-  static void doNoteNonLinear(SimplifyLocals* self, Expression** currp) {
+  static void doNoteNonLinear(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     auto* curr = *currp;
     if (curr->is<Break>()) {
       auto* br = curr->cast<Break>();
@@ -151,25 +154,25 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     self->sinkables.clear();
   }
 
-  static void doNoteIfElseCondition(SimplifyLocals* self, Expression** currp) {
+  static void doNoteIfElseCondition(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     // we processed the condition of this if-else, and now control flow branches
     // into either the true or the false sides
     assert((*currp)->cast<If>()->ifFalse);
     self->sinkables.clear();
   }
 
-  static void doNoteIfElseTrue(SimplifyLocals* self, Expression** currp) {
+  static void doNoteIfElseTrue(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     // we processed the ifTrue side of this if-else, save it on the stack
     assert((*currp)->cast<If>()->ifFalse);
     self->ifStack.push_back(std::move(self->sinkables));
   }
 
-  static void doNoteIfElseFalse(SimplifyLocals* self, Expression** currp) {
+  static void doNoteIfElseFalse(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     // we processed the ifFalse side of this if-else, we can now try to
     // mere with the ifTrue side and optimize a return value, if possible
     auto* iff = (*currp)->cast<If>();
     assert(iff->ifFalse);
-    if (self->allowStructure) {
+    if (allowStructure) {
       self->optimizeIfReturn(iff, currp, self->ifStack.back());
     }
     self->ifStack.pop_back();
@@ -198,15 +201,50 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     }
   }
 
+  void visitLoop(Loop* curr) {
+    if (allowStructure) {
+      if (canUseLoopReturnValue(curr)) {
+        loops.push_back(this->getCurrentPointer());
+      }
+    }
+  }
+
   void visitGetLocal(GetLocal *curr) {
     auto found = sinkables.find(curr->index);
     if (found != sinkables.end()) {
+      auto* set = (*found->second.item)->template cast<SetLocal>(); // the set we may be sinking
+      bool oneUse = firstCycle || getCounter.num[curr->index] == 1;
+      auto* get = set->value->template dynCast<GetLocal>(); // the set's value may be a get (i.e., the set is a copy)
+      // if nesting is not allowed, and this might cause nesting, check if the sink would cause such a thing
+      if (!allowNesting) {
+        // a get is always ok to sink
+        if (!get) {
+          assert(expressionStack.size() >= 2);
+          assert(expressionStack[expressionStack.size() - 1] == curr);
+          auto* parent = expressionStack[expressionStack.size() - 2];
+          bool parentIsSet = parent->template is<SetLocal>();
+          // if the parent of this get is a set, we can sink into the set's value,
+          // it would not be nested.
+          if (!parentIsSet) {
+            return;
+          }
+        }
+      }
+      // we can optimize here
+      if (!allowNesting && get && !oneUse) {
+        // if we can't nest 's a copy with multiple uses, then we can't create
+        // a tee, and we can't nop the origin, but we can at least switch to
+        // the copied index, which may make the origin unneeded eventually.
+        curr->index = get->index;
+        anotherCycle = true;
+        return;
+      }
       // sink it, and nop the origin
-      auto* set = (*found->second.item)->cast<SetLocal>();
-      if (firstCycle || getCounter.num[curr->index] == 1) {
-        replaceCurrent(set->value);
+      if (oneUse) {
+        // with just one use, we can sink just the value
+        this->replaceCurrent(set->value);
       } else {
-        replaceCurrent(set);
+        this->replaceCurrent(set);
         assert(!set->isTee());
         set->setTee(true);
       }
@@ -224,7 +262,7 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     if (set) {
       assert(set->isTee());
       set->setTee(false);
-      replaceCurrent(set);
+      this->replaceCurrent(set);
     }
   }
 
@@ -241,9 +279,11 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     }
   }
 
+  // a full expression stack is used when !allowNesting, so that we can check if
+  // a sink would cause nesting
   std::vector<Expression*> expressionStack;
 
-  static void visitPre(SimplifyLocals* self, Expression** currp) {
+  static void visitPre(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     Expression* curr = *currp;
 
     EffectAnalyzer effects(self->getPassOptions());
@@ -251,10 +291,12 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
       self->checkInvalidations(effects);
     }
 
-    self->expressionStack.push_back(curr);
+    if (!allowNesting) {
+      self->expressionStack.push_back(curr);
+    }
   }
 
-  static void visitPost(SimplifyLocals* self, Expression** currp) {
+  static void visitPost(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     // perform main SetLocal processing here, since we may be the result of
     // replaceCurrent, i.e., the visitor was not called.
     auto* set = (*currp)->dynCast<SetLocal>();
@@ -264,7 +306,7 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
       // store is dead, leave just the value
       auto found = self->sinkables.find(set->index);
       if (found != self->sinkables.end()) {
-        auto* previous = (*found->second.item)->cast<SetLocal>();
+        auto* previous = (*found->second.item)->template cast<SetLocal>();
         assert(!previous->isTee());
         auto* previousValue = previous->value;
         Drop* drop = ExpressionManipulator::convert<SetLocal, Drop>(previous);
@@ -286,7 +328,9 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
       self->sinkables.emplace(std::make_pair(index, SinkableInfo(currp, self->getPassOptions())));
     }
 
-    self->expressionStack.pop_back();
+    if (!allowNesting) {
+      self->expressionStack.pop_back();
+    }
   }
 
   bool canSink(SetLocal* set) {
@@ -299,6 +343,7 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
 
   std::vector<Block*> blocksToEnlarge;
   std::vector<If*> ifsToEnlarge;
+  std::vector<Expression**> loops;
 
   void optimizeBlockReturn(Block* block) {
     if (!block->name.is() || unoptimizableBlocks.count(block->name) > 0) {
@@ -307,7 +352,7 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     auto breaks = std::move(blockBreaks[block->name]);
     blockBreaks.erase(block->name);
     if (breaks.size() == 0) return; // block has no branches TODO we might optimize trivial stuff here too
-    assert(!(*breaks[0].brp)->cast<Break>()->value); // block does not already have a return value (if one break has one, they all do)
+    assert(!(*breaks[0].brp)->template cast<Break>()->value); // block does not already have a return value (if one break has one, they all do)
     // look for a set_local that is present in them all
     bool found = false;
     Index sharedIndex = -1;
@@ -327,6 +372,51 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
       }
     }
     if (!found) return;
+    // If one of our brs is a br_if, then we will give it a value. since
+    // the value executes before the condition, it is dangerous if we are
+    // moving code out of the condition,
+    //  (br_if
+    //   (block
+    //    ..use $x..
+    //    (set_local $x ..)
+    //   )
+    //  )
+    // =>
+    //  (br_if
+    //   (tee_local $x ..) ;; this now affects the use!
+    //   (block
+    //    ..use $x..
+    //   )
+    //  )
+    // so we must check for that.
+    for (size_t j = 0; j < breaks.size(); j++) {
+      // move break set_local's value to the break
+      auto* breakSetLocalPointer = breaks[j].sinkables.at(sharedIndex).item;
+      auto* brp = breaks[j].brp;
+      auto* br = (*brp)->template cast<Break>();
+      auto* set = (*breakSetLocalPointer)->template cast<SetLocal>();
+      if (br->condition) {
+        // TODO: optimize
+        FindAll<SetLocal> findAll(br->condition);
+        for (auto* otherSet : findAll.list) {
+          if (otherSet == set) {
+            // the set is indeed in the condition, so we can't just move it
+            // but maybe there are no effects? see if, ignoring the set
+            // itself, there is any risk
+            Nop nop;
+            *breakSetLocalPointer = &nop;
+            EffectAnalyzer condition(this->getPassOptions(), br->condition);
+            EffectAnalyzer value(this->getPassOptions(), set);
+            *breakSetLocalPointer = set;
+            if (condition.invalidates(value)) {
+              // indeed, we can't do this, stop
+              return;
+            }
+            break; // we found set in the list, can stop now
+          }
+        }
+      }
+    }
     // Great, this local is set in them all, we can optimize!
     if (block->list.size() == 0 || !block->list.back()->is<Nop>()) {
       // We can't do this here, since we can't push to the block -
@@ -338,7 +428,7 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     }
     // move block set_local's value to the end, in return position, and nop the set
     auto* blockSetLocalPointer = sinkables.at(sharedIndex).item;
-    auto* value = (*blockSetLocalPointer)->cast<SetLocal>()->value;
+    auto* value = (*blockSetLocalPointer)->template cast<SetLocal>()->value;
     block->list[block->list.size() - 1] = value;
     block->type = value->type;
     ExpressionManipulator::nop(*blockSetLocalPointer);
@@ -346,25 +436,25 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
       // move break set_local's value to the break
       auto* breakSetLocalPointer = breaks[j].sinkables.at(sharedIndex).item;
       auto* brp = breaks[j].brp;
-      auto* br = (*brp)->cast<Break>();
+      auto* br = (*brp)->template cast<Break>();
       assert(!br->value);
       // if the break is conditional, then we must set the value here - if the break is not reached, we must still have the new value in the local
-      auto* set = (*breakSetLocalPointer)->cast<SetLocal>();
+      auto* set = (*breakSetLocalPointer)->template cast<SetLocal>();
       if (br->condition) {
         br->value = set;
         set->setTee(true);
-        *breakSetLocalPointer = getModule()->allocator.alloc<Nop>();
+        *breakSetLocalPointer = this->getModule()->allocator.template alloc<Nop>();
         // in addition, as this is a conditional br that now has a value, it now returns a value, so it must be dropped
         br->finalize();
-        *brp = Builder(*getModule()).makeDrop(br);
+        *brp = Builder(*this->getModule()).makeDrop(br);
       } else {
         br->value = set->value;
         ExpressionManipulator::nop(set);
       }
     }
     // finally, create a set_local on the block itself
-    auto* newSetLocal = Builder(*getModule()).makeSetLocal(sharedIndex, block);
-    replaceCurrent(newSetLocal);
+    auto* newSetLocal = Builder(*this->getModule()).makeSetLocal(sharedIndex, block);
+    this->replaceCurrent(newSetLocal);
     sinkables.clear();
     anotherCycle = true;
   }
@@ -375,16 +465,46 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     // if this if already has a result, or is unreachable code, we have
     // nothing to do
     if (iff->type != none) return;
-    // We now have the sinkables from both sides of the if.
+    // We now have the sinkables from both sides of the if, and can look
+    // for something to sink. That is either a shared index on both sides,
+    // *or* if one side is unreachable, we can sink anything from the other,
+    //   (if
+    //     (..)
+    //     (br $x)
+    //     (set_local $y (..))
+    //   )
+    //    =>
+    //   (set_local $y
+    //     (if (result i32)
+    //       (..)
+    //       (br $x)
+    //       (..)
+    //     )
+    //   )
     Sinkables& ifFalse = sinkables;
-    Index sharedIndex = -1;
+    Index goodIndex = -1;
     bool found = false;
-    for (auto& sinkable : ifTrue) {
-      Index index = sinkable.first;
-      if (ifFalse.count(index) > 0) {
-        sharedIndex = index;
+    if (iff->ifTrue->type == unreachable) {
+      assert(iff->ifFalse->type != unreachable); // since the if type is none
+      if (!ifFalse.empty()) {
+        goodIndex = ifFalse.begin()->first;
         found = true;
-        break;
+      }
+    } else if (iff->ifFalse->type == unreachable) {
+      assert(iff->ifTrue->type != unreachable); // since the if type is none
+      if (!ifTrue.empty()) {
+        goodIndex = ifTrue.begin()->first;
+        found = true;
+      }
+    } else {
+      // Look for a shared index.
+      for (auto& sinkable : ifTrue) {
+        Index index = sinkable.first;
+        if (ifFalse.count(index) > 0) {
+          goodIndex = index;
+          found = true;
+          break;
+        }
       }
     }
     if (!found) return;
@@ -392,47 +512,58 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     // ensure we have a place to write the return values for, if not, we
     // need another cycle
     auto* ifTrueBlock  = iff->ifTrue->dynCast<Block>();
+    if (iff->ifTrue->type != unreachable) {
+      if (!ifTrueBlock  || ifTrueBlock->list.size() == 0  || !ifTrueBlock->list.back()->is<Nop>()) {
+        ifsToEnlarge.push_back(iff);
+        return;
+      }
+    }
     auto* ifFalseBlock = iff->ifFalse->dynCast<Block>();
-    if (!ifTrueBlock  || ifTrueBlock->list.size() == 0  || !ifTrueBlock->list.back()->is<Nop>() ||
-        !ifFalseBlock || ifFalseBlock->list.size() == 0 || !ifFalseBlock->list.back()->is<Nop>()) {
-      ifsToEnlarge.push_back(iff);
-      return;
+    if (iff->ifFalse->type != unreachable) {
+      if (!ifFalseBlock  || ifFalseBlock->list.size() == 0  || !ifFalseBlock->list.back()->is<Nop>()) {
+        ifsToEnlarge.push_back(iff);
+        return;
+      }
     }
     // all set, go
-    auto *ifTrueItem = ifTrue.at(sharedIndex).item;
-    ifTrueBlock->list[ifTrueBlock->list.size() - 1] = (*ifTrueItem)->cast<SetLocal>()->value;
-    ExpressionManipulator::nop(*ifTrueItem);
-    ifTrueBlock->finalize();
-    assert(ifTrueBlock->type != none);
-    auto *ifFalseItem = ifFalse.at(sharedIndex).item;
-    ifFalseBlock->list[ifFalseBlock->list.size() - 1] = (*ifFalseItem)->cast<SetLocal>()->value;
-    ExpressionManipulator::nop(*ifFalseItem);
-    ifFalseBlock->finalize();
-    assert(ifTrueBlock->type != none);
+    if (iff->ifTrue->type != unreachable) {
+      auto *ifTrueItem = ifTrue.at(goodIndex).item;
+      ifTrueBlock->list[ifTrueBlock->list.size() - 1] = (*ifTrueItem)->template cast<SetLocal>()->value;
+      ExpressionManipulator::nop(*ifTrueItem);
+      ifTrueBlock->finalize();
+      assert(ifTrueBlock->type != none);
+    }
+    if (iff->ifFalse->type != unreachable) {
+      auto *ifFalseItem = ifFalse.at(goodIndex).item;
+      ifFalseBlock->list[ifFalseBlock->list.size() - 1] = (*ifFalseItem)->template cast<SetLocal>()->value;
+      ExpressionManipulator::nop(*ifFalseItem);
+      ifFalseBlock->finalize();
+      assert(ifFalseBlock->type != none);
+    }
     iff->finalize(); // update type
     assert(iff->type != none);
     // finally, create a set_local on the iff itself
-    auto* newSetLocal = Builder(*getModule()).makeSetLocal(sharedIndex, iff);
+    auto* newSetLocal = Builder(*this->getModule()).makeSetLocal(goodIndex, iff);
     *currp = newSetLocal;
     anotherCycle = true;
   }
 
   // override scan to add a pre and a post check task to all nodes
-  static void scan(SimplifyLocals* self, Expression** currp) {
+  static void scan(SimplifyLocals<allowTee, allowStructure, allowNesting>* self, Expression** currp) {
     self->pushTask(visitPost, currp);
 
     auto* curr = *currp;
 
     if (curr->is<If>() && curr->cast<If>()->ifFalse) {
       // handle if-elses in a special manner, using the ifStack
-      self->pushTask(SimplifyLocals::doNoteIfElseFalse, currp);
-      self->pushTask(SimplifyLocals::scan, &curr->cast<If>()->ifFalse);
-      self->pushTask(SimplifyLocals::doNoteIfElseTrue, currp);
-      self->pushTask(SimplifyLocals::scan, &curr->cast<If>()->ifTrue);
-      self->pushTask(SimplifyLocals::doNoteIfElseCondition, currp);
-      self->pushTask(SimplifyLocals::scan, &curr->cast<If>()->condition);
+      self->pushTask(SimplifyLocals<allowTee, allowStructure, allowNesting>::doNoteIfElseFalse, currp);
+      self->pushTask(SimplifyLocals<allowTee, allowStructure, allowNesting>::scan, &curr->cast<If>()->ifFalse);
+      self->pushTask(SimplifyLocals<allowTee, allowStructure, allowNesting>::doNoteIfElseTrue, currp);
+      self->pushTask(SimplifyLocals<allowTee, allowStructure, allowNesting>::scan, &curr->cast<If>()->ifTrue);
+      self->pushTask(SimplifyLocals<allowTee, allowStructure, allowNesting>::doNoteIfElseCondition, currp);
+      self->pushTask(SimplifyLocals<allowTee, allowStructure, allowNesting>::scan, &curr->cast<If>()->condition);
     } else {
-      WalkerPass<LinearExecutionWalker<SimplifyLocals>>::scan(self, currp);
+      WalkerPass<LinearExecutionWalker<SimplifyLocals<allowTee, allowStructure, allowNesting>>>::scan(self, currp);
     }
 
     self->pushTask(visitPre, currp);
@@ -454,11 +585,11 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     do {
       anotherCycle = false;
       // main operation
-      WalkerPass<LinearExecutionWalker<SimplifyLocals>>::doWalkFunction(func);
+      WalkerPass<LinearExecutionWalker<SimplifyLocals<allowTee, allowStructure, allowNesting>>>::doWalkFunction(func);
       // enlarge blocks that were marked, for the next round
       if (blocksToEnlarge.size() > 0) {
         for (auto* block : blocksToEnlarge) {
-          block->list.push_back(getModule()->allocator.alloc<Nop>());
+          block->list.push_back(this->getModule()->allocator.template alloc<Nop>());
         }
         blocksToEnlarge.clear();
         anotherCycle = true;
@@ -466,20 +597,37 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
       // enlarge ifs that were marked, for the next round
       if (ifsToEnlarge.size() > 0) {
         for (auto* iff : ifsToEnlarge) {
-          auto ifTrue = Builder(*getModule()).blockify(iff->ifTrue);
+          auto ifTrue = Builder(*this->getModule()).blockify(iff->ifTrue);
           iff->ifTrue = ifTrue;
-          if (ifTrue->list.size() == 0 || !ifTrue->list.back()->is<Nop>()) {
-            ifTrue->list.push_back(getModule()->allocator.alloc<Nop>());
+          if (ifTrue->list.size() == 0 || !ifTrue->list.back()->template is<Nop>()) {
+            ifTrue->list.push_back(this->getModule()->allocator.template alloc<Nop>());
           }
-          auto ifFalse = Builder(*getModule()).blockify(iff->ifFalse);
+          auto ifFalse = Builder(*this->getModule()).blockify(iff->ifFalse);
           iff->ifFalse = ifFalse;
-          if (ifFalse->list.size() == 0 || !ifFalse->list.back()->is<Nop>()) {
-            ifFalse->list.push_back(getModule()->allocator.alloc<Nop>());
+          if (ifFalse->list.size() == 0 || !ifFalse->list.back()->template is<Nop>()) {
+            ifFalse->list.push_back(this->getModule()->allocator.template alloc<Nop>());
           }
         }
         ifsToEnlarge.clear();
         anotherCycle = true;
       }
+      // handle loops. note that a lot happens in this pass, and we can't just modify
+      // set_locals when we see a loop - it might be tracked - and we can't also just
+      // assume our loop didn't move either (might be in a block now). So we do this
+      // when all other work is done - as loop return values are rare, that is fine.
+      if (!anotherCycle) {
+        for (auto* currp : loops) {
+          auto* curr = (*currp)->template cast<Loop>();
+          assert(canUseLoopReturnValue(curr));
+          auto* set = curr->body->template cast<SetLocal>();
+          curr->body = set->value;
+          set->value = curr;
+          curr->finalize(curr->body->type);
+          *currp = set;
+          anotherCycle = true;
+        }
+      }
+      loops.clear();
       // clean up
       sinkables.clear();
       blockBreaks.clear();
@@ -499,22 +647,37 @@ struct SimplifyLocals : public WalkerPass<LinearExecutionWalker<SimplifyLocals>>
     remover.numGetLocals = &getCounter.num;
     remover.walkFunction(func);
   }
+
+  bool canUseLoopReturnValue(Loop* curr) {
+    // Optimizing a loop return value is trivial: just see if it contains
+    // a set_local, and pull that out.
+    if (auto* set = curr->body->template dynCast<SetLocal>()) {
+      if (isConcreteType(set->value->type)) {
+        return true;
+      }
+    }
+    return false;
+  }
 };
 
 Pass *createSimplifyLocalsPass() {
-  return new SimplifyLocals(true, true);
+  return new SimplifyLocals<true, true>();
 }
 
 Pass *createSimplifyLocalsNoTeePass() {
-  return new SimplifyLocals(false, true);
+  return new SimplifyLocals<false, true>();
 }
 
 Pass *createSimplifyLocalsNoStructurePass() {
-  return new SimplifyLocals(true, false);
+  return new SimplifyLocals<true, false>();
 }
 
 Pass *createSimplifyLocalsNoTeeNoStructurePass() {
-  return new SimplifyLocals(false, false);
+  return new SimplifyLocals<false, false>();
+}
+
+Pass *createSimplifyLocalsNoNestingPass() {
+  return new SimplifyLocals<false, false, false>();
 }
 
 } // namespace wasm
