@@ -56,11 +56,11 @@ class TranslateToFuzzReader {
 public:
   TranslateToFuzzReader(Module& wasm, std::string& filename) : wasm(wasm), builder(wasm) {
     auto input(read_file<std::vector<char>>(filename, Flags::Binary, Flags::Release));
-    read_data(input);
+    readData(input);
   }
 
   TranslateToFuzzReader(Module& wasm, std::vector<char> input) : wasm(wasm), builder(wasm) {
-    read_data(input);
+    readData(input);
   }
 
   void pickPasses(OptimizationOptions& options) {
@@ -120,7 +120,8 @@ public:
     std::cout << "shrink level: " << options.passOptions.shrinkLevel << '\n';
   }
 
-  void build() {
+  void build(bool initEmitAtomics = true) {
+    emitAtomics = initEmitAtomics;
     setupMemory();
     setupTable();
     setupGlobals();
@@ -145,15 +146,18 @@ private:
   size_t pos; // the position in the input
   bool finishedInput; // whether we already cycled through all the input (if so, we should try to finish things off)
 
+  // The maximum amount of params to each function.
+  static const int MAX_PARAMS = 10;
+
+  // The maximum amount of vars in each function.
+  static const int MAX_VARS = 20;
+
   // some things require luck, try them a few times
   static const int TRIES = 10;
 
   // beyond a nesting limit, greatly decrease the chance to continue to nest
   static const int NESTING_LIMIT = 11;
 
-  // reduce the chance for a function to call itself by this factor
-  static const int RECURSION_FACTOR = 10;
-
   // the maximum size of a block
   static const int BLOCK_FACTOR = 5;
 
@@ -172,14 +176,17 @@ private:
   static const bool DE_NAN = true;
 
   // Whether to emit atomics
-  static const bool ATOMICS = true;
+  bool emitAtomics = true;
+
+  // Whether to emit atomic waits (which in single-threaded mode, may hang...)
+  static const bool ATOMIC_WAITS = false;
 
   // after we finish the input, we start going through it again, but xoring
   // so it's not identical
   int xorFactor = 0;
 
 
-  void read_data(std::vector<char> input) {
+  void readData(std::vector<char> input) {
     bytes.swap(input);
     pos = 0;
     finishedInput = false;
@@ -240,7 +247,7 @@ private:
     wasm.table.segments.emplace_back(builder.makeConst(Literal(int32_t(0))));
   }
 
-  std::map<WasmType, std::vector<Name>> globalsByType;
+  std::map<Type, std::vector<Name>> globalsByType;
 
   void setupGlobals() {
     size_t index = 0;
@@ -311,7 +318,7 @@ private:
   }
 
   void addDeNanSupport() {
-    auto add = [&](Name name, WasmType type, Literal literal, BinaryOp op) {
+    auto add = [&](Name name, Type type, Literal literal, BinaryOp op) {
       auto* func = new Function;
       func->name = name;
       func->params.push_back(type);
@@ -351,7 +358,7 @@ private:
   // which we try to minimize the risk of
   std::vector<Expression*> hangStack;
 
-  std::map<WasmType, std::vector<Index>> typeLocals; // type => list of locals with that type
+  std::map<Type, std::vector<Index>> typeLocals; // type => list of locals with that type
 
   Function* addFunction() {
     Index num = wasm.functions.size();
@@ -359,13 +366,13 @@ private:
     func->name = std::string("func_") + std::to_string(num);
     func->result = getReachableType();
     assert(typeLocals.empty());
-    Index numParams = upToSquared(5);
+    Index numParams = upToSquared(MAX_PARAMS);
     for (Index i = 0; i < numParams; i++) {
       auto type = getConcreteType();
       typeLocals[type].push_back(func->params.size());
       func->params.push_back(type);
     }
-    Index numVars = upToSquared(10);
+    Index numVars = upToSquared(MAX_VARS);
     for (Index i = 0; i < numVars; i++) {
       auto type = getConcreteType();
       typeLocals[type].push_back(func->params.size() + func->vars.size());
@@ -424,7 +431,7 @@ private:
         args.push_back(makeConst(type));
       }
       Expression* invoke = builder.makeCall(func->name, args, func->result);
-      if (isConcreteWasmType(func->result)) {
+      if (isConcreteType(func->result)) {
         invoke = builder.makeDrop(invoke);
       }
       invocations.push_back(invoke);
@@ -451,12 +458,12 @@ private:
 
   int nesting = 0;
 
-  Expression* make(WasmType type) {
+  Expression* make(Type type) {
     // when we should stop, emit something small (but not necessarily trivial)
     if (finishedInput ||
         nesting >= 5 * NESTING_LIMIT || // hard limit
         (nesting >= NESTING_LIMIT && !oneIn(3))) {
-      if (isConcreteWasmType(type)) {
+      if (isConcreteType(type)) {
         if (oneIn(2)) {
           return makeConst(type);
         } else {
@@ -488,7 +495,7 @@ private:
     return ret;
   }
 
-  Expression* _makeConcrete(WasmType type) {
+  Expression* _makeConcrete(Type type) {
     auto choice = upTo(100);
     if (choice < 10) return makeConst(type);
     if (choice < 30) return makeSetLocal(type);
@@ -562,8 +569,8 @@ private:
   }
 
   // make something with no chance of infinite recursion
-  Expression* makeTrivial(WasmType type) {
-    if (isConcreteWasmType(type)) {
+  Expression* makeTrivial(Type type) {
+    if (isConcreteType(type)) {
       if (oneIn(2)) {
         return makeGetLocal(type);
       } else {
@@ -574,7 +581,7 @@ private:
     }
     assert(type == unreachable);
     Expression* ret = nullptr;
-    if (isConcreteWasmType(func->result)) {
+    if (isConcreteType(func->result)) {
       ret = makeTrivial(func->result);
     }
     return builder.makeReturn(ret);
@@ -582,7 +589,7 @@ private:
 
   // specific expression creators
 
-  Expression* makeBlock(WasmType type) {
+  Expression* makeBlock(Type type) {
     auto* ret = builder.makeBlock();
     ret->type = type; // so we have it during child creation
     ret->name = makeLabel();
@@ -606,13 +613,13 @@ private:
     }
     // give a chance to make the final element an unreachable break, instead
     // of concrete - a common pattern (branch to the top of a loop etc.)
-    if (!finishedInput && isConcreteWasmType(type) && oneIn(2)) {
+    if (!finishedInput && isConcreteType(type) && oneIn(2)) {
       ret->list.push_back(makeBreak(unreachable));
     } else {
       ret->list.push_back(make(type));
     }
     breakableStack.pop_back();
-    if (isConcreteWasmType(type)) {
+    if (isConcreteType(type)) {
       ret->finalize(type);
     } else {
       ret->finalize();
@@ -625,7 +632,7 @@ private:
     return ret;
   }
 
-  Expression* makeLoop(WasmType type) {
+  Expression* makeLoop(Type type) {
     auto* ret = wasm.allocator.alloc<Loop>();
     ret->type = type; // so we have it during child creation
     ret->name = makeLabel();
@@ -666,7 +673,7 @@ private:
   }
 
   // make something, with a good chance of it being a block
-  Expression* makeMaybeBlock(WasmType type) {
+  Expression* makeMaybeBlock(Type type) {
     // if past the limit, prefer not to emit blocks
     if (nesting >= NESTING_LIMIT || oneIn(3)) {
       return make(type);
@@ -675,7 +682,7 @@ private:
     }
   }
 
-  Expression* makeIf(WasmType type) {
+  Expression* makeIf(Type type) {
     auto* condition = makeCondition();
     hangStack.push_back(nullptr);
     auto* ret = makeIf({ condition, makeMaybeBlock(type), makeMaybeBlock(type) });
@@ -687,7 +694,7 @@ private:
     return builder.makeIf(args.a, args.b, args.c);
   }
 
-  Expression* makeBreak(WasmType type) {
+  Expression* makeBreak(Type type) {
     if (breakableStack.empty()) return makeTrivial(type);
     Expression* condition = nullptr;
     if (type != unreachable) {
@@ -700,7 +707,7 @@ private:
       auto* target = vectorPick(breakableStack);
       auto name = getTargetName(target);
       auto valueType = getTargetType(target);
-      if (isConcreteWasmType(type)) {
+      if (isConcreteType(type)) {
         // we are flowing out a value
         if (valueType != type) {
           // we need to break to a proper place
@@ -762,7 +769,7 @@ private:
     return makeTrivial(type);
   }
 
-  Expression* makeCall(WasmType type) {
+  Expression* makeCall(Type type) {
     // seems ok, go on
     int tries = TRIES;
     while (tries-- > 0) {
@@ -771,8 +778,6 @@ private:
         target = vectorPick(wasm.functions).get();
       }
       if (target->result != type) continue;
-      // reduce the odds of recursion dramatically, to limit infinite loops
-      if (target == func && !oneIn(RECURSION_FACTOR * TRIES)) continue;
       // we found one!
       std::vector<Expression*> args;
       for (auto argType : target->params) {
@@ -784,7 +789,7 @@ private:
     return make(type);
   }
 
-  Expression* makeCallIndirect(WasmType type) {
+  Expression* makeCallIndirect(Type type) {
     auto& data = wasm.table.segments[0].data;
     if (data.empty()) return make(type);
     // look for a call target with the right type
@@ -822,15 +827,15 @@ private:
     );
   }
 
-  Expression* makeGetLocal(WasmType type) {
+  Expression* makeGetLocal(Type type) {
     auto& locals = typeLocals[type];
     if (locals.empty()) return makeConst(type);
     return builder.makeGetLocal(vectorPick(locals), type);
   }
 
-  Expression* makeSetLocal(WasmType type) {
+  Expression* makeSetLocal(Type type) {
     bool tee = type != none;
-    WasmType valueType;
+    Type valueType;
     if (tee) {
       valueType = type;
     } else {
@@ -846,13 +851,13 @@ private:
     }
   }
 
-  Expression* makeGetGlobal(WasmType type) {
+  Expression* makeGetGlobal(Type type) {
     auto& globals = globalsByType[type];
     if (globals.empty()) return makeConst(type);
     return builder.makeGetGlobal(vectorPick(globals), type);
   }
 
-  Expression* makeSetGlobal(WasmType type) {
+  Expression* makeSetGlobal(Type type) {
     assert(type == none);
     type = getConcreteType();
     auto& globals = globalsByType[type];
@@ -875,7 +880,7 @@ private:
     return ret;
   }
 
-  Load* makeNonAtomicLoad(WasmType type) {
+  Load* makeNonAtomicLoad(Type type) {
     auto offset = logify(get());
     auto ptr = makePointer();
     switch (type) {
@@ -908,10 +913,10 @@ private:
     }
   }
 
-  Expression* makeLoad(WasmType type) {
+  Expression* makeLoad(Type type) {
     auto* ret = makeNonAtomicLoad(type);
     if (type != i32 && type != i64) return ret;
-    if (!ATOMICS || oneIn(2)) return ret;
+    if (!emitAtomics || oneIn(2)) return ret;
     // make it atomic
     wasm.memory.shared = true;
     ret->isAtomic = true;
@@ -920,7 +925,7 @@ private:
     return ret;
   }
 
-  Store* makeNonAtomicStore(WasmType type) {
+  Store* makeNonAtomicStore(Type type) {
     if (type == unreachable) {
       // make a normal store, then make it unreachable
       auto* ret = makeNonAtomicStore(getConcreteType());
@@ -968,10 +973,10 @@ private:
     }
   }
 
-  Store* makeStore(WasmType type) {
+  Store* makeStore(Type type) {
     auto* ret = makeNonAtomicStore(type);
     if (ret->value->type != i32 && ret->value->type != i64) return ret;
-    if (!ATOMICS || oneIn(2)) return ret;
+    if (!emitAtomics || oneIn(2)) return ret;
     // make it atomic
     wasm.memory.shared = true;
     ret->isAtomic = true;
@@ -979,7 +984,7 @@ private:
     return ret;
   }
 
-  Expression* makeConst(WasmType type) {
+  Expression* makeConst(Type type) {
     Literal value;
     switch (upTo(4)) {
       case 0: {
@@ -1082,7 +1087,7 @@ private:
     return builder.makeUnary(args.a, args.b);
   }
 
-  Expression* makeUnary(WasmType type) {
+  Expression* makeUnary(Type type) {
     if (type == unreachable) {
       if (auto* unary = makeUnary(getConcreteType())->dynCast<Unary>()) {
         return makeDeNanOp(builder.makeUnary(unary->op, make(unreachable)));
@@ -1094,7 +1099,7 @@ private:
       case i32: {
         switch (upTo(4)) {
           case 0: {
-            if (ATOMICS) {
+            if (emitAtomics) {
               return makeUnary({ pick(EqZInt32, ClzInt32, CtzInt32, PopcntInt32, ExtendS8Int32, ExtendS16Int32), make(i32) });
             } else {
               return makeUnary({ pick(EqZInt32, ClzInt32, CtzInt32, PopcntInt32), make(i32) });
@@ -1110,7 +1115,7 @@ private:
       case i64: {
         switch (upTo(4)) {
           case 0: {
-            if (ATOMICS) {
+            if (emitAtomics) {
               return makeUnary({ pick(ClzInt64, CtzInt64, PopcntInt64, ExtendS8Int64, ExtendS16Int64, ExtendS32Int64), make(i64) });
             } else {
               return makeUnary({ pick(ClzInt64, CtzInt64, PopcntInt64), make(i64) });
@@ -1150,7 +1155,7 @@ private:
     return builder.makeBinary(args.a, args.b, args.c);
   }
 
-  Expression* makeBinary(WasmType type) {
+  Expression* makeBinary(Type type) {
     if (type == unreachable) {
       if (auto* binary = makeBinary(getConcreteType())->dynCast<Binary>()) {
         return makeDeNanOp(makeBinary({ binary->op, make(unreachable), make(unreachable) }));
@@ -1186,17 +1191,17 @@ private:
     return builder.makeSelect(args.a, args.b, args.c);
   }
 
-  Expression* makeSelect(WasmType type) {
+  Expression* makeSelect(Type type) {
     return makeDeNanOp(makeSelect({ make(i32), make(type), make(type) }));
   }
 
-  Expression* makeSwitch(WasmType type) {
+  Expression* makeSwitch(Type type) {
     assert(type == unreachable);
     if (breakableStack.empty()) return make(type);
     // we need to find proper targets to break to; try a bunch
     int tries = TRIES;
     std::vector<Name> names;
-    WasmType valueType = unreachable;
+    Type valueType = unreachable;
     while (tries-- > 0) {
       auto* target = vectorPick(breakableStack);
       auto name = getTargetName(target);
@@ -1216,42 +1221,42 @@ private:
     }
     auto default_ = names.back();
     names.pop_back();
-    auto temp1 = make(i32), temp2 = isConcreteWasmType(valueType) ? make(valueType) : nullptr;
+    auto temp1 = make(i32), temp2 = isConcreteType(valueType) ? make(valueType) : nullptr;
     return builder.makeSwitch(names, default_, temp1, temp2);
   }
 
-  Expression* makeDrop(WasmType type) {
+  Expression* makeDrop(Type type) {
     return builder.makeDrop(make(type == unreachable ? type : getConcreteType()));
   }
 
-  Expression* makeReturn(WasmType type) {
-    return builder.makeReturn(isConcreteWasmType(func->result) ? make(func->result) : nullptr);
+  Expression* makeReturn(Type type) {
+    return builder.makeReturn(isConcreteType(func->result) ? make(func->result) : nullptr);
   }
 
-  Expression* makeNop(WasmType type) {
+  Expression* makeNop(Type type) {
     assert(type == none);
     return builder.makeNop();
   }
 
-  Expression* makeUnreachable(WasmType type) {
+  Expression* makeUnreachable(Type type) {
     assert(type == unreachable);
     return builder.makeUnreachable();
   }
 
-  Expression* makeAtomic(WasmType type) {
-    if (!ATOMICS || (type != i32 && type != i64)) return makeTrivial(type);
+  Expression* makeAtomic(Type type) {
+    if (!emitAtomics || (type != i32 && type != i64)) return makeTrivial(type);
     wasm.memory.shared = true;
     if (type == i32 && oneIn(2)) {
-      if (oneIn(2)) {
+      if (ATOMIC_WAITS && oneIn(2)) {
         auto* ptr = makePointer();
         auto expectedType = pick(i32, i64);
         auto* expected = make(expectedType);
         auto* timeout = make(i64);
-        return builder.makeAtomicWait(ptr, expected, timeout, expectedType);
+        return builder.makeAtomicWait(ptr, expected, timeout, expectedType, logify(get()));
       } else {
         auto* ptr = makePointer();
         auto* count = make(i32);
-        return builder.makeAtomicWake(ptr, count);
+        return builder.makeAtomicWake(ptr, count, logify(get()));
       }
     }
     Index bytes;
@@ -1292,7 +1297,7 @@ private:
 
   // special getters
 
-  WasmType getType() {
+  Type getType() {
     switch (upTo(6)) {
       case 0: return i32;
       case 1: return i64;
@@ -1304,7 +1309,7 @@ private:
     WASM_UNREACHABLE();
   }
 
-  WasmType getReachableType() {
+  Type getReachableType() {
     switch (upTo(5)) {
       case 0: return i32;
       case 1: return i64;
@@ -1315,7 +1320,7 @@ private:
     WASM_UNREACHABLE();
   }
 
-  WasmType getConcreteType() {
+  Type getConcreteType() {
     switch (upTo(4)) {
       case 0: return i32;
       case 1: return i64;
@@ -1420,7 +1425,7 @@ private:
     WASM_UNREACHABLE();
   }
 
-  WasmType getTargetType(Expression* target) {
+  Type getTargetType(Expression* target) {
     if (auto* block = target->dynCast<Block>()) {
       return block->type;
     } else if (target->is<Loop>()) {
