@@ -15,7 +15,6 @@
  */
 
 #include "support/command-line.h"
-#include "support/defaults.h"
 
 //
 // Shared optimization options for commandline tools
@@ -30,12 +29,11 @@ struct OptimizationOptions : public Options {
   PassOptions passOptions;
   FeatureSet features = Feature::Atomics;
 
-  OptimizationOptions(const std::string &command, const std::string &description) : Options(command, description) {
+  OptimizationOptions(const std::string& command, const std::string& description) : Options(command, description) {
     (*this).add("", "-O", "execute default optimization passes",
                 Options::Arguments::Zero,
                 [this](Options*, const std::string&) {
-                  passOptions.optimizeLevel = BINARYEN_DEFAULT_OPTIMIZE_LEVEL;
-                  passOptions.shrinkLevel = BINARYEN_DEFAULT_SHRINK_LEVEL;
+                  passOptions.setDefaultOptimizationOptions();
                   passes.push_back(DEFAULT_OPT_PASSES);
                 })
            .add("", "-O0", "execute no optimization passes",
@@ -89,6 +87,11 @@ struct OptimizationOptions : public Options {
                 [this](Options* o, const std::string& argument) {
                   passOptions.shrinkLevel = atoi(argument.c_str());
                 })
+           .add("--no-validation", "-n", "Disables validation, assumes inputs are correct",
+                Options::Arguments::Zero,
+                [this](Options* o, const std::string& argument) {
+                  passOptions.validate = false;
+                })
            .add("--ignore-implicit-traps", "-iit", "Optimize under the helpful assumption that no surprising traps occur (from load, div/mod, etc.)",
                 Options::Arguments::Zero,
                 [this](Options*, const std::string&) {
