@@ -64,6 +64,7 @@ int main(int argc, const char* argv[]) {
   Name entry;
   bool emitBinary = true;
   bool debugInfo = false;
+  bool converge = false;
   bool fuzzExec = false;
   bool fuzzBinary = false;
   std::string extraFuzzCommand;
@@ -82,31 +83,34 @@ int main(int argc, const char* argv[]) {
            })
       .add("--emit-text", "-S", "Emit text instead of binary for the output file",
            Options::Arguments::Zero,
-           [&](Options *o, const std::string &argument) { emitBinary = false; })
+           [&](Options *o, const std::string& argument) { emitBinary = false; })
       .add("--debuginfo", "-g", "Emit names section and debug info",
            Options::Arguments::Zero,
-           [&](Options *o, const std::string &arguments) { debugInfo = true; })
+           [&](Options *o, const std::string& arguments) { debugInfo = true; })
+      .add("--converge", "-c", "Run passes to convergence, continuing while binary size decreases",
+           Options::Arguments::Zero,
+           [&](Options *o, const std::string& arguments) { converge = true; })
       .add("--fuzz-exec", "-fe", "Execute functions before and after optimization, helping fuzzing find bugs",
            Options::Arguments::Zero,
-           [&](Options *o, const std::string &arguments) { fuzzExec = true; })
+           [&](Options *o, const std::string& arguments) { fuzzExec = true; })
       .add("--fuzz-binary", "-fb", "Convert to binary and back after optimizations and before fuzz-exec, helping fuzzing find binary format bugs",
            Options::Arguments::Zero,
-           [&](Options *o, const std::string &arguments) { fuzzBinary = true; })
+           [&](Options *o, const std::string& arguments) { fuzzBinary = true; })
       .add("--extra-fuzz-command", "-efc", "An extra command to run on the output before and after optimizing. The output is compared between the two, and an error occurs if they are not equal",
            Options::Arguments::One,
-           [&](Options *o, const std::string &arguments) { extraFuzzCommand = arguments; })
+           [&](Options *o, const std::string& arguments) { extraFuzzCommand = arguments; })
       .add("--translate-to-fuzz", "-ttf", "Translate the input into a valid wasm module *somehow*, useful for fuzzing",
            Options::Arguments::Zero,
-           [&](Options *o, const std::string &arguments) { translateToFuzz = true; })
+           [&](Options *o, const std::string& arguments) { translateToFuzz = true; })
       .add("--fuzz-passes", "-fp", "Pick a random set of passes to run, useful for fuzzing. this depends on translate-to-fuzz (it picks the passes from the input)",
            Options::Arguments::Zero,
-           [&](Options *o, const std::string &arguments) { fuzzPasses = true; })
+           [&](Options *o, const std::string& arguments) { fuzzPasses = true; })
       .add("--emit-js-wrapper", "-ejw", "Emit a JavaScript wrapper file that can run the wasm with some test values, useful for fuzzing",
            Options::Arguments::One,
-           [&](Options *o, const std::string &arguments) { emitJSWrapper = arguments; })
+           [&](Options *o, const std::string& arguments) { emitJSWrapper = arguments; })
       .add("--emit-spec-wrapper", "-esw", "Emit a wasm spec interpreter wrapper file that can run the wasm with some test values, useful for fuzzing",
            Options::Arguments::One,
-           [&](Options *o, const std::string &arguments) { emitSpecWrapper = arguments; })
+           [&](Options *o, const std::string& arguments) { emitSpecWrapper = arguments; })
       .add_positional("INFILE", Options::Arguments::One,
                       [](Options* o, const std::string& argument) {
                         o->extra["infile"] = argument;
@@ -127,14 +131,17 @@ int main(int argc, const char* argv[]) {
       reader.read(options.extra["infile"], wasm);
     } catch (ParseException& p) {
       p.dump(std::cerr);
+      std::cerr << '\n';
       Fatal() << "error in parsing input";
     } catch (std::bad_alloc& b) {
       Fatal() << "error in building module, std::bad_alloc (possibly invalid request for silly amounts of memory)";
     }
 
-    if (!WasmValidator().validate(wasm, features)) {
-      WasmPrinter::printModule(&wasm);
-      Fatal() << "error in validating input";
+    if (options.passOptions.validate) {
+      if (!WasmValidator().validate(wasm, features)) {
+        WasmPrinter::printModule(&wasm);
+        Fatal() << "error in validating input";
+      }
     }
   } else {
     // translate-to-fuzz
@@ -143,10 +150,12 @@ int main(int argc, const char* argv[]) {
       reader.pickPasses(options);
     }
     reader.build();
-    if (!WasmValidator().validate(wasm, features)) {
-      WasmPrinter::printModule(&wasm);
-      std::cerr << "translate-to-fuzz must always generate a valid module";
-      abort();
+    if (options.passOptions.validate) {
+      if (!WasmValidator().validate(wasm, features)) {
+        WasmPrinter::printModule(&wasm);
+        std::cerr << "translate-to-fuzz must always generate a valid module";
+        abort();
+      }
     }
   }
 
@@ -194,22 +203,47 @@ int main(int argc, const char* argv[]) {
     auto input = buffer.getAsChars();
     WasmBinaryBuilder parser(other, input, false);
     parser.read();
-    bool valid = WasmValidator().validate(other, features);
-    if (!valid) {
-      WasmPrinter::printModule(&other);
+    if (options.passOptions.validate) {
+      bool valid = WasmValidator().validate(other, features);
+      if (!valid) {
+        WasmPrinter::printModule(&other);
+      }
+      assert(valid);
     }
-    assert(valid);
     curr = &other;
   }
 
   if (options.runningPasses()) {
     if (options.debug) std::cerr << "running passes...\n";
-    options.runPasses(*curr);
-    bool valid = WasmValidator().validate(*curr, features);
-    if (!valid) {
-      WasmPrinter::printModule(&*curr);
+    auto runPasses = [&]() {
+      options.runPasses(*curr);
+      if (options.passOptions.validate) {
+        bool valid = WasmValidator().validate(*curr, features);
+        if (!valid) {
+          WasmPrinter::printModule(&*curr);
+        }
+        assert(valid);
+      }
+    };
+    runPasses();
+    if (converge) {
+      // Keep on running passes to convergence, defined as binary
+      // size no longer decreasing.
+      auto getSize = [&]() {
+        BufferWithRandomAccess buffer;
+        WasmBinaryWriter writer(curr, buffer);
+        writer.write();
+        return buffer.size();
+      };
+      auto lastSize = getSize();
+      while (1) {
+        if (options.debug) std::cerr << "running iteration for convergence (" << lastSize << ")...\n";
+        runPasses();
+        auto currSize = getSize();
+        if (currSize >= lastSize) break;
+        lastSize = currSize;
+      }
     }
-    assert(valid);
   }
 
   if (fuzzExec) {
