@@ -31,11 +31,43 @@
 #include "support/command-line.h"
 #include "support/file.h"
 #include "support/path.h"
+#include "support/timing.h"
 #include "wasm-io.h"
 #include "wasm-builder.h"
+#include "ir/branch-utils.h"
+#include "ir/iteration.h"
 #include "ir/literal-utils.h"
+#include "ir/properties.h"
 #include "wasm-validator.h"
-
+#ifdef _WIN32
+#ifndef NOMINMAX
+#define NOMINMAX
+#endif
+#include <Windows.h>
+// Create a string with last error message
+std::string GetLastErrorStdStr() {
+  DWORD error = GetLastError();
+  if (error) {
+    LPVOID lpMsgBuf;
+    DWORD bufLen = FormatMessage(
+        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
+        FORMAT_MESSAGE_FROM_SYSTEM |
+        FORMAT_MESSAGE_IGNORE_INSERTS,
+        NULL,
+        error,
+        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
+        (LPTSTR) &lpMsgBuf,
+        0, NULL );
+    if (bufLen) {
+      LPCSTR lpMsgStr = (LPCSTR)lpMsgBuf;
+      std::string result(lpMsgStr, lpMsgStr+bufLen);
+      LocalFree(lpMsgBuf);
+      return result;
+    }
+  }
+  return std::string();
+}
+#endif
 using namespace wasm;
 
 // a timeout on every execution of the command
@@ -44,15 +76,101 @@ size_t timeout = 2;
 struct ProgramResult {
   int code;
   std::string output;
+  double time;
 
   ProgramResult() {}
   ProgramResult(std::string command) {
     getFromExecution(command);
   }
 
+#ifdef _WIN32
+  void getFromExecution(std::string command) {
+    Timer timer;
+    timer.start();
+    SECURITY_ATTRIBUTES saAttr;
+    saAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
+    saAttr.bInheritHandle = TRUE;
+    saAttr.lpSecurityDescriptor = NULL;
+
+    HANDLE hChildStd_OUT_Rd;
+    HANDLE hChildStd_OUT_Wr;
+
+    if (
+      // Create a pipe for the child process's STDOUT.
+      !CreatePipe(&hChildStd_OUT_Rd, &hChildStd_OUT_Wr, &saAttr, 0) ||
+      // Ensure the read handle to the pipe for STDOUT is not inherited.
+      !SetHandleInformation(hChildStd_OUT_Rd, HANDLE_FLAG_INHERIT, 0)
+    ) {
+      Fatal() << "CreatePipe \"" << command << "\" failed: " << GetLastErrorStdStr() << ".\n";
+    }
+
+    STARTUPINFO si;
+    PROCESS_INFORMATION pi;
+
+    ZeroMemory(&si, sizeof(si));
+    si.cb = sizeof(si);
+    si.hStdError = hChildStd_OUT_Wr;
+    si.hStdOutput = hChildStd_OUT_Wr;
+    si.dwFlags |= STARTF_USESTDHANDLES;
+    ZeroMemory(&pi, sizeof(pi));
+
+    // Start the child process.
+    if (!CreateProcess(NULL,   // No module name (use command line)
+      (LPSTR)command.c_str(),// Command line
+      NULL,           // Process handle not inheritable
+      NULL,           // Thread handle not inheritable
+      TRUE,           // Set handle inheritance to TRUE
+      0,              // No creation flags
+      NULL,           // Use parent's environment block
+      NULL,           // Use parent's starting directory
+      &si,            // Pointer to STARTUPINFO structure
+      &pi )           // Pointer to PROCESS_INFORMATION structure
+    ) {
+      Fatal() << "CreateProcess \"" << command << "\" failed: " << GetLastErrorStdStr() << ".\n";
+    }
+
+    // Wait until child process exits.
+    DWORD retVal = WaitForSingleObject(pi.hProcess, timeout * 1000);
+    if (retVal == WAIT_TIMEOUT) {
+      printf("Command timeout: %s", command.c_str());
+      TerminateProcess(pi.hProcess, -1);
+    }
+    DWORD dwordExitCode;
+    if (!GetExitCodeProcess(pi.hProcess, &dwordExitCode)) {
+      Fatal() << "GetExitCodeProcess failed: " << GetLastErrorStdStr() << ".\n";
+    }
+    code = (int)dwordExitCode;
+
+    // Close process and thread handles.
+    CloseHandle(pi.hProcess);
+    CloseHandle(pi.hThread);
+
+    // Read output from the child process's pipe for STDOUT
+    // Stop when there is no more data.
+    {
+      const int BUFSIZE = 4096;
+      DWORD dwRead, dwTotal, dwTotalRead = 0;
+      CHAR chBuf[BUFSIZE];
+      BOOL bSuccess = FALSE;
+
+      PeekNamedPipe(hChildStd_OUT_Rd, NULL, 0, NULL, &dwTotal, NULL);
+      while (dwTotalRead < dwTotal) {
+        bSuccess = ReadFile(hChildStd_OUT_Rd, chBuf, BUFSIZE - 1, &dwRead, NULL);
+        if (!bSuccess || dwRead == 0) break;
+        chBuf[dwRead] = 0;
+        dwTotalRead += dwRead;
+        output.append(chBuf);
+      }
+    }
+    timer.stop();
+    time = timer.getTotal();
+  }
+#else // POSIX
   // runs the command and notes the output
   // TODO: also stderr, not just stdout?
   void getFromExecution(std::string command) {
+    Timer timer;
+    timer.start();
     // do this using just core stdio.h and stdlib.h, for portability
     // sadly this requires two invokes
     code = system(("timeout " + std::to_string(timeout) + "s " + command + " > /dev/null 2> /dev/null").c_str());
@@ -63,7 +181,10 @@ struct ProgramResult {
       output.append(buffer);
     }
     pclose(stream);
+    timer.stop();
+    time = timer.getTotal() / 2;
   }
+#endif // _WIN32
 
   bool operator==(ProgramResult& other) {
     return code == other.code && output == other.output;
@@ -77,7 +198,7 @@ struct ProgramResult {
   }
 
   void dump(std::ostream& o) {
-    o << "[ProgramResult] code: " << code << " stdout: \n" << output << "[/ProgramResult]\n";
+    o << "[ProgramResult] code: " << code << " stdout: \n" << output << "[====]\nin " << time << " seconds\n[/ProgramResult]\n";
   }
 };
 
@@ -92,6 +213,11 @@ inline std::ostream& operator<<(std::ostream& o, ProgramResult& result) {
 
 ProgramResult expected;
 
+// Removing functions is extremely beneficial and efficient. We aggressively
+// try to remove functions, unless we've seen they can't be removed, in which
+// case we may try again but much later.
+static std::unordered_set<Name> functionsWeTriedToRemove;
+
 struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<Reducer>>> {
   std::string command, test, working;
   bool verbose, debugInfo;
@@ -125,6 +251,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       "--remove-memory",
       "--remove-unused-names --remove-unused-brs",
       "--remove-unused-module-elements",
+      "--remove-unused-nonfunction-module-elements",
       "--reorder-functions",
       "--reorder-locals",
       "--simplify-locals --vacuum",
@@ -171,28 +298,35 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
   //               from the start
   size_t reduceDestructively(int factor_) {
     factor = factor_;
-    Module wasm;
-    ModuleReader reader;
-    reader.read(working, wasm);
     // prepare
+    loadWorking();
     reduced = 0;
-    builder = make_unique<Builder>(wasm);
     funcsSeen = 0;
     // before we do any changes, it should be valid to write out the module:
     // size should be as expected, and output should be as expected
-    setModule(&wasm);
-    if (!writeAndTestReduction()) {
-      std::cerr << "\n|! WARNING: writing before destructive reduction fails, very unlikely reduction can work\n\n";
+    ProgramResult result;
+    if (!writeAndTestReduction(result)) {
+      std::cerr << "\n|! WARNING: writing before destructive reduction fails, very unlikely reduction can work\n" << result << '\n';
     }
     // destroy!
-    walkModule(&wasm);
+    walkModule(getModule());
     return reduced;
   }
 
+  void loadWorking() {
+    module = make_unique<Module>();
+    Module wasm;
+    ModuleReader reader;
+    reader.read(working, *module);
+    builder = make_unique<Builder>(*module);
+    setModule(module.get());
+  }
+
   // destructive reduction state
 
   size_t reduced;
   Expression* beforeReduction;
+  std::unique_ptr<Module> module;
   std::unique_ptr<Builder> builder;
   Index funcsSeen;
   int factor;
@@ -240,8 +374,8 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     return true;
   }
 
-  void noteReduction() {
-    reduced++;
+  void noteReduction(size_t amount = 1) {
+    reduced += amount;
     copy_file(test, working);
   }
 
@@ -271,9 +405,10 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
   // don't need to duplicate work that they do
 
   void visitExpression(Expression* curr) {
+    // type-based reductions
     if (curr->type == none) {
-      if (tryToReduceCurrentToNone()) return;
-    } else if (isConcreteWasmType(curr->type)) {
+      if (tryToReduceCurrentToNop()) return;
+    } else if (isConcreteType(curr->type)) {
       if (tryToReduceCurrentToConst()) return;
     } else {
       assert(curr->type == unreachable);
@@ -294,26 +429,97 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
       handleCondition(select->condition);
     } else if (auto* sw = curr->dynCast<Switch>()) {
       handleCondition(sw->condition);
-    } else if (auto* set = curr->dynCast<SetLocal>()) {
-      if (set->isTee()) {
-        // maybe we don't need the set
-        tryToReplaceCurrent(set->value);
+    } else if (auto* block = curr->dynCast<Block>()) {
+      if (!shouldTryToReduce()) return;
+      // replace a singleton
+      auto& list = block->list;
+      if (list.size() == 1 && !BranchUtils::BranchSeeker::hasNamed(block, block->name)) {
+        if (tryToReplaceCurrent(block->list[0])) return;
       }
-    } else if (auto* unary = curr->dynCast<Unary>()) {
-      // maybe we can pass through
-      tryToReplaceCurrent(unary->value);
-    } else if (auto* binary = curr->dynCast<Binary>()) {
-      // maybe we can pass through
-      if (!tryToReplaceCurrent(binary->left)) {
-        tryToReplaceCurrent(binary->right);
+      // try to get rid of nops
+      Index i = 0;
+      while (list.size() > 1 && i < list.size()) {
+        auto* curr = list[i];
+        if (curr->is<Nop>() && shouldTryToReduce()) {
+          // try to remove it
+          for (Index j = i; j < list.size() - 1; j++) {
+            list[j] = list[j + 1];
+          }
+          list.resize(list.size() - 1);
+          if (writeAndTestReduction()) {
+            std::cerr << "|      block-nop removed\n";
+            noteReduction();
+            return;
+          }
+          list.resize(list.size() + 1);
+          // we failed; undo
+          for (Index j = list.size() - 1; j > i; j--) {
+            list[j] = list[j - 1];
+          }
+          list[i] = curr;
+        }
+        i++;
+      }
+      return; // nothing more to do
+    } else if (auto* loop = curr->dynCast<Loop>()) {
+      if (shouldTryToReduce() && !BranchUtils::BranchSeeker::hasNamed(loop, loop->name)) {
+        tryToReplaceCurrent(loop->body);
+      }
+      return; // nothing more to do
+    }
+    // Finally, try to replace with a child.
+    for (auto* child : ChildIterator(curr)) {
+      if (tryToReplaceCurrent(child)) return;
+    }
+    // If that didn't work, try to replace with a child + a unary conversion
+    if (isConcreteType(curr->type) &&
+        !curr->is<Unary>()) { // but not if it's already unary
+      for (auto* child : ChildIterator(curr)) {
+        if (child->type == curr->type) continue; // already tried
+        if (!isConcreteType(child->type)) continue; // no conversion
+        Expression* fixed;
+        switch (curr->type) {
+          case i32: {
+            switch (child->type) {
+              case i64: fixed = builder->makeUnary(WrapInt64, child); break;
+              case f32: fixed = builder->makeUnary(TruncSFloat32ToInt32, child); break;
+              case f64: fixed = builder->makeUnary(TruncSFloat64ToInt32, child); break;
+              default: WASM_UNREACHABLE();
+            }
+            break;
+          }
+          case i64: {
+            switch (child->type) {
+              case i32: fixed = builder->makeUnary(ExtendSInt32, child); break;
+              case f32: fixed = builder->makeUnary(TruncSFloat32ToInt64, child); break;
+              case f64: fixed = builder->makeUnary(TruncSFloat64ToInt64, child); break;
+              default: WASM_UNREACHABLE();
+            }
+            break;
+          }
+          case f32: {
+            switch (child->type) {
+              case i32: fixed = builder->makeUnary(ConvertSInt32ToFloat32, child); break;
+              case i64: fixed = builder->makeUnary(ConvertSInt64ToFloat32, child); break;
+              case f64: fixed = builder->makeUnary(DemoteFloat64, child); break;
+              default: WASM_UNREACHABLE();
+            }
+            break;
+          }
+          case f64: {
+            switch (child->type) {
+              case i32: fixed = builder->makeUnary(ConvertSInt32ToFloat64, child); break;
+              case i64: fixed = builder->makeUnary(ConvertSInt64ToFloat64, child); break;
+              case f32: fixed = builder->makeUnary(PromoteFloat32, child); break;
+              default: WASM_UNREACHABLE();
+            }
+            break;
+          }
+          default: WASM_UNREACHABLE();
+        }
+        assert(fixed->type == curr->type);
+        if (tryToReplaceCurrent(fixed)) return;
       }
-    } else if (auto* call = curr->dynCast<Call>()) {
-      handleCall(call);
-    } else if (auto* call = curr->dynCast<CallImport>()) {
-      handleCall(call);
-    } else if (auto* call = curr->dynCast<CallIndirect>()) {
-      if (tryToReplaceCurrent(call->target)) return;
-      handleCall(call);
     }
   }
 
@@ -353,19 +559,23 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
 
   template<typename T, typename U>
   void visitSegmented(T* curr, U zero, size_t bonus) {
-    // try to reduce to first function
-    // shrink segment elements
+    // try to reduce to first function. first, shrink segment elements.
+    // while we are shrinking successfully, keep going exponentially.
+    bool justShrank = false;
+    bool shrank = false;
     for (auto& segment : curr->segments) {
       auto& data = segment.data;
       size_t skip = 1; // when we succeed, try to shrink by more and more, similar to bisection
       for (size_t i = 0; i < data.size() && !data.empty(); i++) {
-        if (!shouldTryToReduce(bonus)) continue;
+        if (!justShrank && !shouldTryToReduce(bonus)) continue;
         auto save = data;
         for (size_t j = 0; j < skip; j++) {
           if (!data.empty()) data.pop_back();
         }
-        if (writeAndTestReduction()) {
+        auto justShrank = writeAndTestReduction();
+        if (justShrank) {
           std::cerr << "|      shrank segment (skip: " << skip << ")\n";
+          shrank = true;
           noteReduction();
           skip = std::min(size_t(factor), 2 * skip);
         } else {
@@ -388,43 +598,175 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
         } else {
           item = save;
         }
+        if (shrank) {
+          // zeroing is fairly inefficient. if we are managing to shrink
+          // (which we do exponentially), just zero one per segment at most
+          break;
+        }
       }
     }
   }
 
   void visitModule(Module* curr) {
-    // try to remove exports
-    std::cerr << "|    try to remove exports\n";
-    std::vector<Export> exports;
-    for (auto& exp : curr->exports) {
-      if (!shouldTryToReduce(10000)) continue;
-      exports.push_back(*exp);
-    }
-    for (auto& exp : exports) {
-      curr->removeExport(exp.name);
-      if (!writeAndTestReduction()) {
-        curr->addExport(new Export(exp));
-      } else {
-        std::cerr << "|      removed export " << exp.name << '\n';
-        noteReduction();
-      }
-    }
+    assert(curr == module.get());
     // try to remove functions
     std::cerr << "|    try to remove functions\n";
-    std::vector<Function> functions;
-    for (auto& func : curr->functions) {
-      if (!shouldTryToReduce(10000)) continue;
-      functions.push_back(*func);
+    std::vector<Name> functionNames;
+    for (auto& func : module->functions) {
+      functionNames.push_back(func->name);
     }
-    for (auto& func : functions) {
-      curr->removeFunction(func.name);
-      if (WasmValidator().validate(*curr, Feature::MVP, WasmValidator::Globally | WasmValidator::Quiet) &&
-          writeAndTestReduction()) {
-        std::cerr << "|      removed function " << func.name << '\n';
-        noteReduction();
-      } else {
-        curr->addFunction(new Function(func));
+    size_t skip = 1;
+    // If we just removed some functions in the previous iteration, keep trying to remove more
+    // as this is one of the most efficient ways to reduce.
+    bool justRemoved = false;
+    for (size_t i = 0; i < functionNames.size(); i++) {
+      if (!justRemoved &&
+          functionsWeTriedToRemove.count(functionNames[i]) == 1 &&
+         !shouldTryToReduce(std::max((factor / 100) + 1, 1000))) continue;
+      std::vector<Name> names;
+      for (size_t j = 0; names.size() < skip && i + j < functionNames.size(); j++) {
+        auto name = functionNames[i + j];
+        if (module->getFunctionOrNull(name)) {
+          names.push_back(name);
+          functionsWeTriedToRemove.insert(name);
+        }
       }
+      if (names.size() == 0) continue;
+      std::cout << "|    try to remove " << names.size() << " functions (skip: " << skip << ")\n";
+      justRemoved = tryToRemoveFunctions(names);
+      if (justRemoved) {
+        noteReduction(names.size());
+        i += skip;
+        skip = std::min(size_t(factor), 2 * skip);
+      } else {
+        skip = std::max(skip / 2, size_t(1)); // or 1?
+      }
+    }
+    // try to remove exports
+    std::cerr << "|    try to remove exports (with factor " << factor << ")\n";
+    std::vector<Export> exports;
+    for (auto& exp : module->exports) {
+      exports.push_back(*exp);
+    }
+    skip = 1;
+    for (size_t i = 0; i < exports.size(); i++) {
+      if (!shouldTryToReduce(std::max((factor / 100) + 1, 1000))) continue;
+      std::vector<Export> currExports;
+      for (size_t j = 0; currExports.size() < skip && i + j < exports.size(); j++) {
+        auto exp = exports[i + j];
+        if (module->getExportOrNull(exp.name)) {
+          currExports.push_back(exp);
+          module->removeExport(exp.name);
+        }
+      }
+      ProgramResult result;
+      if (!writeAndTestReduction(result)) {
+        for (auto exp : currExports) {
+          module->addExport(new Export(exp));
+        }
+        skip = std::max(skip / 2, size_t(1)); // or 1?
+      } else {
+        std::cerr << "|      removed " << currExports.size() << " exports\n";
+        noteReduction(currExports.size());
+        i += skip;
+        skip = std::min(size_t(factor), 2 * skip);
+      }
+    }
+    // If we are left with a single function that is not exported or used in
+    // a table, that is useful as then we can change the return type.
+    if (module->functions.size() == 1 && module->exports.empty() && module->table.segments.empty()) {
+      auto* func = module->functions[0].get();
+      // We can't remove something that might have breaks to it.
+      if (!Properties::isNamedControlFlow(func->body)) {
+        auto funcType = func->type;
+        auto funcResult = func->result;
+        auto* funcBody = func->body;
+        for (auto* child : ChildIterator(func->body)) {
+          if (!(isConcreteType(child->type) || child->type == none)) {
+            continue; // not something a function can return
+          }
+          // Try to replace the body with the child, fixing up the function
+          // to accept it.
+          func->type = Name();
+          func->result = child->type;
+          func->body = child;
+          if (writeAndTestReduction()) {
+            // great, we succeeded!
+            std::cerr << "|    altered function result type\n";
+            noteReduction(1);
+            break;
+          }
+          // Undo.
+          func->type = funcType;
+          func->result = funcResult;
+          func->body = funcBody;
+        }
+      }
+    }
+  }
+
+  bool tryToRemoveFunctions(std::vector<Name> names) {
+    for (auto name : names) {
+      module->removeFunction(name);
+    }
+
+    // remove all references to them
+    struct FunctionReferenceRemover : public PostWalker<FunctionReferenceRemover> {
+      std::unordered_set<Name> names;
+      std::vector<Name> exportsToRemove;
+
+      FunctionReferenceRemover(std::vector<Name>& vec) {
+        for (auto name : vec) {
+          names.insert(name);
+        }
+      }
+      void visitCall(Call* curr) {
+        if (names.count(curr->target)) {
+          replaceCurrent(Builder(*getModule()).replaceWithIdenticalType(curr));
+        }
+      }
+      void visitExport(Export* curr) {
+        if (names.count(curr->value)) {
+          exportsToRemove.push_back(curr->name);
+        }
+      }
+      void visitTable(Table* curr) {
+        Name other;
+        for (auto& segment : curr->segments) {
+          for (auto name : segment.data) {
+            if (!names.count(name)) {
+              other = name;
+              break;
+            }
+          }
+          if (!other.isNull()) break;
+        }
+        if (other.isNull()) return; // we failed to find a replacement
+        for (auto& segment : curr->segments) {
+          for (auto& name : segment.data) {
+            if (names.count(name)) {
+              name = other;
+            }
+          }
+        }
+      }
+      void doWalkModule(Module* module) {
+        PostWalker<FunctionReferenceRemover>::doWalkModule(module);
+        for (auto name : exportsToRemove) {
+          module->removeExport(name);
+        }
+      }
+    };
+    FunctionReferenceRemover referenceRemover(names);
+    referenceRemover.walkModule(module.get());
+
+    if (WasmValidator().validate(*module, Feature::All, WasmValidator::Globally | WasmValidator::Quiet) &&
+        writeAndTestReduction()) {
+      std::cerr << "|      removed " << names.size() << " functions\n";
+      return true;
+    } else {
+      loadWorking(); // restore it from orbit
+      return false;
     }
   }
 
@@ -441,14 +783,7 @@ struct Reducer : public WalkerPass<PostWalker<Reducer, UnifiedExpressionVisitor<
     }
   }
 
-  template<typename T>
-  void handleCall(T* call) {
-    for (auto* op : call->operands) {
-      if (tryToReplaceCurrent(op)) return;
-    }
-  }
-
-  bool tryToReduceCurrentToNone() {
+  bool tryToReduceCurrentToNop() {
     auto* curr = getCurrent();
     if (curr->is<Nop>()) return false;
     // try to replace with a trivial value
@@ -515,6 +850,12 @@ int main(int argc, const char* argv[]) {
            [&](Options* o, const std::string& argument) {
              working = argument;
            })
+      .add("--binaries", "-b", "binaryen binaries location (bin/ directory)",
+           Options::Arguments::One,
+           [&](Options* o, const std::string& argument) {
+             // Add separator just in case
+             Path::setBinaryenBinDir(argument + Path::getPathSeparator());
+           })
       .add("--verbose", "-v", "Verbose output mode",
            Options::Arguments::Zero,
            [&](Options* o, const std::string& argument) {
@@ -545,6 +886,7 @@ int main(int argc, const char* argv[]) {
   if (test.size() == 0) Fatal() << "test file not provided\n";
   if (working.size() == 0) Fatal() << "working file not provided\n";
 
+  std::cerr << "|wasm-reduce\n";
   std::cerr << "|input: " << input << '\n';
   std::cerr << "|test: " << test << '\n';
   std::cerr << "|working: " << working << '\n';
@@ -554,6 +896,7 @@ int main(int argc, const char* argv[]) {
   expected.getFromExecution(command);
 
   std::cerr << "|expected result:\n" << expected << '\n';
+  std::cerr << "|!! Make sure the above is what you expect! !!\n\n";
 
   auto stopIfNotForced = [&](std::string message, ProgramResult& result) {
     std::cerr << "|! " << message << '\n' << result << '\n';
@@ -562,7 +905,11 @@ int main(int argc, const char* argv[]) {
     }
   };
 
-  std::cerr << "|checking that command has different behavior on invalid binary\n";
+  if (expected.time + 1 >= timeout) {
+    stopIfNotForced("execution time is dangerously close to the timeout - you should probably increase the timeout", expected);
+  }
+
+  std::cerr << "|checking that command has different behavior on invalid binary (this verifies that the test file is used by the command)\n";
   {
     {
       std::ofstream dst(test, std::ios::binary);
@@ -589,11 +936,12 @@ int main(int argc, const char* argv[]) {
   }
 
   copy_file(input, working);
-  std::cerr << "|input size: " << file_size(working) << "\n";
+  auto workingSize = file_size(working);
+  std::cerr << "|input size: " << workingSize << "\n";
 
   std::cerr << "|starting reduction!\n";
 
-  int factor = 4096;
+  int factor = workingSize * 2;
   size_t lastDestructiveReductions = 0;
   size_t lastPostPassesSize = 0;
 
