@@ -104,7 +104,7 @@ static void run_asserts(Name moduleName, size_t* i, bool* checked, Module* wasm,
         std::cerr << "Unknown entry " << entry << std::endl;
       } else {
         LiteralList arguments;
-        for (WasmType param : function->params) {
+        for (Type param : function->params) {
           arguments.push_back(Literal(param));
         }
         try {
@@ -287,7 +287,11 @@ int main(int argc, const char* argv[]) {
         builders[moduleName].swap(builder);
         modules[moduleName].swap(module);
         i++;
-        assert(WasmValidator().validate(*modules[moduleName]));
+        bool valid = WasmValidator().validate(*modules[moduleName]);
+        if (!valid) {
+          WasmPrinter::printModule(modules[moduleName].get());
+        }
+        assert(valid);
         run_asserts(moduleName, &i, &checked, modules[moduleName].get(), &root, builders[moduleName].get(), entry);
       } else {
         run_asserts(Name(), &i, &checked, nullptr, &root, nullptr, entry);
