@@ -58,7 +58,7 @@ struct LEB {
 
   bool hasMore(T temp, MiniT byte) {
     // for signed, we must ensure the last bit has the right sign, as it will zero extend
-    return std::is_signed<T>::value ? (temp != 0 && temp != -1) || (value >= 0 && (byte & 64)) || (value < 0 && !(byte & 64)) : (temp != 0);
+    return std::is_signed<T>::value ? (temp != 0 && temp != T(-1)) || (value >= 0 && (byte & 64)) || (value < 0 && !(byte & 64)) : (temp != 0);
   }
 
   void write(std::vector<uint8_t>* out) {
@@ -638,7 +638,7 @@ enum MemoryFlags {
 } // namespace BinaryConsts
 
 
-inline S32LEB binaryWasmType(WasmType type) {
+inline S32LEB binaryType(Type type) {
   int ret;
   switch (type) {
     // None only used for block signatures. TODO: Separate out?
@@ -704,7 +704,7 @@ public:
   void writeImports();
 
   std::map<Index, size_t> mappedLocals; // local index => index in compact form of [all int32s][all int64s]etc
-  std::map<WasmType, size_t> numLocalsByType; // type => number of locals of that type in the compact form
+  std::map<Type, size_t> numLocalsByType; // type => number of locals of that type in the compact form
 
   void mapLocals(Function* function);
   void writeFunctionSignatures();
@@ -724,6 +724,7 @@ public:
   void writeNames();
   void writeSourceMapUrl();
   void writeSymbolMap();
+  void writeUserSections();
 
   void writeSourceMapProlog();
   void writeSourceMapEpilog();
@@ -802,7 +803,7 @@ public:
 class WasmBinaryBuilder {
   Module& wasm;
   MixedArena& allocator;
-  std::vector<char>& input;
+  const std::vector<char>& input;
   bool debug;
   std::istream* sourceMap;
   std::pair<uint32_t, Function::DebugLocation> nextDebugLocation;
@@ -814,7 +815,14 @@ class WasmBinaryBuilder {
   std::set<BinaryConsts::Section> seenSections;
 
 public:
-  WasmBinaryBuilder(Module& wasm, std::vector<char>& input, bool debug) : wasm(wasm), allocator(wasm.allocator), input(input), debug(debug), sourceMap(nullptr), nextDebugLocation(0, { 0, 0, 0 }), useDebugLocation(false) {}
+  WasmBinaryBuilder(Module& wasm, const std::vector<char>& input, bool debug)
+    : wasm(wasm),
+      allocator(wasm.allocator),
+      input(input),
+      debug(debug),
+      sourceMap(nullptr),
+      nextDebugLocation(0, { 0, 0, 0 }),
+      useDebugLocation(false) {}
 
   void read();
   void readUserSection(size_t payloadLen);
@@ -831,7 +839,8 @@ public:
   uint64_t getU64LEB();
   int32_t getS32LEB();
   int64_t getS64LEB();
-  WasmType getWasmType();
+  Type getType();
+  Type getConcreteType();
   Name getString();
   Name getInlineString();
   void verifyInt8(int8_t x);
@@ -854,9 +863,7 @@ public:
   void readFunctionSignatures();
   size_t nextLabel;
 
-  Name getNextLabel() {
-    return cashew::IString(("label$" + std::to_string(nextLabel++)).c_str(), false);
-  }
+  Name getNextLabel();
 
   // We read functions before we know their names, so we need to backpatch the names later
   std::vector<Function*> functions; // we store functions here before wasm.addFunction after we know their names
@@ -866,6 +873,9 @@ public:
   Function* currFunction = nullptr;
   Index endOfFunction = -1; // before we see a function (like global init expressions), there is no end of function to check
 
+  // Throws a parsing error if we are not in a function context
+  void requireFunctionContext(const char* error);
+
   void readFunctions();
 
   std::map<Export*, Index> exportIndexes;
@@ -881,12 +891,27 @@ public:
     BreakTarget(Name name, int arity) : name(name), arity(arity) {}
   };
   std::vector<BreakTarget> breakStack;
+  // the names that breaks target. this lets us know if a block has breaks to it or not.
   std::unordered_set<Name> breakTargetNames;
 
   std::vector<Expression*> expressionStack;
 
-  bool definitelyUnreachable; // set when we know code is definitely unreachable. this helps parse
-                              // stacky wasm code, which can be unsuitable for our IR when unreachable
+  // set when we know code is unreachable in the sense of the wasm spec: we are in a block
+  // and after an unreachable element.
+  // this helps parse stacky wasm code, which can be unsuitable for our IR when unreachable.
+  bool unreachableInTheWasmSense;
+
+  // set when the current code being processed will not be emitted in the output, which is the
+  // case when it is literally unreachable, for example,
+  // (block $a
+  //   (unreachable)
+  //   (block $b
+  //     ;; code here is reachable in the wasm sense, even though $b as a whole is not
+  //     (unreachable)
+  //     ;; code here is unreachable in the wasm sense
+  //   )
+  // )
+  bool willBeIgnored;
 
   BinaryConsts::ASTNodes lastSeparator = BinaryConsts::End;
 
@@ -926,7 +951,7 @@ public:
   void visitBlock(Block *curr);
 
   // Gets a block of expressions. If it's just one, return that singleton.
-  Expression* getBlockOrSingleton(WasmType type);
+  Expression* getBlockOrSingleton(Type type);
 
   void visitIf(If *curr);
   void visitLoop(Loop *curr);
@@ -951,7 +976,7 @@ public:
   void visitSetLocal(SetLocal *curr, uint8_t code);
   void visitGetGlobal(GetGlobal *curr);
   void visitSetGlobal(SetGlobal *curr);
-  void readMemoryAccess(Address& alignment, size_t bytes, Address& offset);
+  void readMemoryAccess(Address& alignment, Address& offset);
   bool maybeVisitLoad(Expression*& out, uint8_t code, bool isAtomic);
   bool maybeVisitStore(Expression*& out, uint8_t code, bool isAtomic);
   bool maybeVisitAtomicRMW(Expression*& out, uint8_t code);
@@ -967,6 +992,8 @@ public:
   void visitNop(Nop *curr);
   void visitUnreachable(Unreachable *curr);
   void visitDrop(Drop *curr);
+
+  void throwError(std::string text);
 };
 
 } // namespace wasm
