@@ -109,6 +109,8 @@ enum UnaryOp {
   // Extend signed subword-sized integer. This differs from e.g. ExtendSInt32
   // because the input integer is in an i64 value insetad of an i32 value.
   ExtendS8Int32, ExtendS16Int32, ExtendS8Int64, ExtendS16Int64, ExtendS32Int64,
+
+  InvalidUnary
 };
 
 enum BinaryOp {
@@ -135,6 +137,8 @@ enum BinaryOp {
   // relational ops
   EqFloat64, NeFloat64, // int or float
   LtFloat64, LeFloat64, GtFloat64, GeFloat64, // float
+
+  InvalidBinary
 };
 
 enum HostOp {
@@ -192,15 +196,15 @@ public:
     HostId,
     NopId,
     UnreachableId,
-    AtomicCmpxchgId,
     AtomicRMWId,
+    AtomicCmpxchgId,
     AtomicWaitId,
     AtomicWakeId,
     NumExpressionIds
   };
   Id _id;
 
-  WasmType type; // the type of the expression: its *output*, not necessarily its input(s)
+  Type type; // the type of the expression: its *output*, not necessarily its input(s)
 
   Expression(Id id) : _id(id), type(none) {}
 
@@ -250,13 +254,20 @@ public:
   Name name;
   ExpressionList list;
 
+  // set the type purely based on its contents. this scans the block, so it is not fast.
+  void finalize();
+
   // set the type given you know its type, which is the case when parsing
   // s-expression or binary, as explicit types are given. the only additional work
-  // this does is to set the type to unreachable in the cases that is needed.
-  void finalize(WasmType type_);
+  // this does is to set the type to unreachable in the cases that is needed
+  // (which may require scanning the block)
+  void finalize(Type type_);
 
-  // set the type purely based on its contents. this scans the block, so it is not fast
-  void finalize();
+  // set the type given you know its type, and you know if there is a break to this
+  // block. this avoids the need to scan the contents of the block in the case that
+  // it might be unreachable, so it is recommended if you already know the type
+  // and breakability anyhow.
+  void finalize(Type type_, bool hasBreak);
 };
 
 class If : public SpecificExpression<Expression::IfId> {
@@ -271,7 +282,7 @@ public:
   // set the type given you know its type, which is the case when parsing
   // s-expression or binary, as explicit types are given. the only additional work
   // this does is to set the type to unreachable in the cases that is needed.
-  void finalize(WasmType type_);
+  void finalize(Type type_);
 
   // set the type purely based on its contents.
   void finalize();
@@ -288,7 +299,7 @@ public:
   // set the type given you know its type, which is the case when parsing
   // s-expression or binary, as explicit types are given. the only additional work
   // this does is to set the type to unreachable in the cases that is needed.
-  void finalize(WasmType type_);
+  void finalize(Type type_);
 
   // set the type purely based on its contents.
   void finalize();
@@ -345,8 +356,8 @@ public:
 class FunctionType {
 public:
   Name name;
-  WasmType result;
-  std::vector<WasmType> params;
+  Type result;
+  std::vector<Type> params;
 
   FunctionType() : result(none) {}
 
@@ -436,7 +447,7 @@ public:
   bool isAtomic;
   Expression* ptr;
   Expression* value;
-  WasmType valueType; // the store never returns a value
+  Type valueType; // the store never returns a value
 
   void finalize();
 };
@@ -474,10 +485,11 @@ class AtomicWait : public SpecificExpression<Expression::AtomicWaitId> {
   AtomicWait() = default;
   AtomicWait(MixedArena& allocator) : AtomicWait() {}
 
+  Address offset;
   Expression* ptr;
   Expression* expected;
   Expression* timeout;
-  WasmType expectedType;
+  Type expectedType;
 
   void finalize();
 };
@@ -487,6 +499,7 @@ class AtomicWake : public SpecificExpression<Expression::AtomicWakeId> {
   AtomicWake() = default;
   AtomicWake(MixedArena& allocator) : AtomicWake() {}
 
+  Address offset;
   Expression* ptr;
   Expression* wakeCount;
 
@@ -591,9 +604,9 @@ public:
 class Function {
 public:
   Name name;
-  WasmType result;
-  std::vector<WasmType> params; // function locals are
-  std::vector<WasmType> vars;   // params plus vars
+  Type result;
+  std::vector<Type> params; // function locals are
+  std::vector<Type> vars;   // params plus vars
   Name type; // if null, it is implicit in params and result
   Expression* body;
 
@@ -620,7 +633,7 @@ public:
   Name getLocalName(Index index);
   Index getLocalIndex(Name name);
   Index getVarIndexBase();
-  WasmType getLocalType(Index index);
+  Type getLocalType(Index index);
 
   Name getLocalNameOrDefault(Index index);
   Name getLocalNameOrGeneric(Index index);
@@ -642,7 +655,7 @@ public:
   Name name, module, base; // name = module.base
   ExternalKind kind;
   Name functionType; // for Function imports
-  WasmType globalType; // for Global imports
+  Type globalType; // for Global imports
 };
 
 class Export {
@@ -719,7 +732,7 @@ public:
 class Global {
 public:
   Name name;
-  WasmType type;
+  Type type;
   Expression* init;
   bool mutable_;
 };
