@@ -153,7 +153,7 @@ struct ValidationInfo {
     return true;
   }
 
-  void shouldBeIntOrUnreachable(WasmType ty, Expression* curr, const char* text, Function* func = nullptr) {
+  void shouldBeIntOrUnreachable(Type ty, Expression* curr, const char* text, Function* func = nullptr) {
     switch (ty) {
       case i32:
       case i64:
@@ -176,16 +176,16 @@ struct FunctionValidator : public WalkerPass<PostWalker<FunctionValidator>> {
   FunctionValidator(ValidationInfo* info) : info(*info) {}
 
   struct BreakInfo {
-    WasmType type;
+    Type type;
     Index arity;
     BreakInfo() {}
-    BreakInfo(WasmType type, Index arity) : type(type), arity(arity) {}
+    BreakInfo(Type type, Index arity) : type(type), arity(arity) {}
   };
 
   std::map<Name, Expression*> breakTargets;
   std::map<Expression*, BreakInfo> breakInfos;
 
-  WasmType returnType = unreachable; // type used in returns
+  Type returnType = unreachable; // type used in returns
 
   std::set<Name> labelNames; // Binaryen IR requires that label names must be unique - IR generators must ensure that
 
@@ -201,15 +201,15 @@ public:
     if (curr->name.is()) self->breakTargets[curr->name] = curr;
   }
 
-  void visitBlock(Block *curr);
+  void visitBlock(Block* curr);
 
   static void visitPreLoop(FunctionValidator* self, Expression** currp) {
     auto* curr = (*currp)->cast<Loop>();
     if (curr->name.is()) self->breakTargets[curr->name] = curr;
   }
 
-  void visitLoop(Loop *curr);
-  void visitIf(If *curr);
+  void visitLoop(Loop* curr);
+  void visitIf(If* curr);
 
   // override scan to add a pre and a post check task to all nodes
   static void scan(FunctionValidator* self, Expression** currp) {
@@ -221,28 +221,28 @@ public:
   }
 
   void noteBreak(Name name, Expression* value, Expression* curr);
-  void visitBreak(Break *curr);
-  void visitSwitch(Switch *curr);
-  void visitCall(Call *curr);
-  void visitCallImport(CallImport *curr);
-  void visitCallIndirect(CallIndirect *curr);
+  void visitBreak(Break* curr);
+  void visitSwitch(Switch* curr);
+  void visitCall(Call* curr);
+  void visitCallImport(CallImport* curr);
+  void visitCallIndirect(CallIndirect* curr);
   void visitGetLocal(GetLocal* curr);
-  void visitSetLocal(SetLocal *curr);
+  void visitSetLocal(SetLocal* curr);
   void visitGetGlobal(GetGlobal* curr);
-  void visitSetGlobal(SetGlobal *curr);
-  void visitLoad(Load *curr);
-  void visitStore(Store *curr);
-  void visitAtomicRMW(AtomicRMW *curr);
-  void visitAtomicCmpxchg(AtomicCmpxchg *curr);
-  void visitAtomicWait(AtomicWait *curr);
-  void visitAtomicWake(AtomicWake *curr);
-  void visitBinary(Binary *curr);
-  void visitUnary(Unary *curr);
+  void visitSetGlobal(SetGlobal* curr);
+  void visitLoad(Load* curr);
+  void visitStore(Store* curr);
+  void visitAtomicRMW(AtomicRMW* curr);
+  void visitAtomicCmpxchg(AtomicCmpxchg* curr);
+  void visitAtomicWait(AtomicWait* curr);
+  void visitAtomicWake(AtomicWake* curr);
+  void visitBinary(Binary* curr);
+  void visitUnary(Unary* curr);
   void visitSelect(Select* curr);
   void visitDrop(Drop* curr);
   void visitReturn(Return* curr);
   void visitHost(Host* curr);
-  void visitFunction(Function *curr);
+  void visitFunction(Function* curr);
 
   // helpers
 private:
@@ -274,13 +274,13 @@ private:
     return info.shouldBeUnequal(left, right, curr, text, getFunction());
   }
 
-  void shouldBeIntOrUnreachable(WasmType ty, Expression* curr, const char* text) {
+  void shouldBeIntOrUnreachable(Type ty, Expression* curr, const char* text) {
     return info.shouldBeIntOrUnreachable(ty, curr, text, getFunction());
   }
 
-  void validateAlignment(size_t align, WasmType type, Index bytes, bool isAtomic,
+  void validateAlignment(size_t align, Type type, Index bytes, bool isAtomic,
                          Expression* curr);
-  void validateMemBytes(uint8_t bytes, WasmType type, Expression* curr);
+  void validateMemBytes(uint8_t bytes, Type type, Expression* curr);
 };
 
 void FunctionValidator::noteLabelName(Name name) {
@@ -289,28 +289,28 @@ void FunctionValidator::noteLabelName(Name name) {
   labelNames.insert(name);
 }
 
-void FunctionValidator::visitBlock(Block *curr) {
+void FunctionValidator::visitBlock(Block* curr) {
   // if we are break'ed to, then the value must be right for us
   if (curr->name.is()) {
     noteLabelName(curr->name);
     if (breakInfos.count(curr) > 0) {
       auto& info = breakInfos[curr];
-      if (isConcreteWasmType(curr->type)) {
+      if (isConcreteType(curr->type)) {
         shouldBeTrue(info.arity != 0, curr, "break arities must be > 0 if block has a value");
       } else {
         shouldBeTrue(info.arity == 0, curr, "break arities must be 0 if block has no value");
       }
       // none or unreachable means a poison value that we should ignore - if consumed, it will error
-      if (isConcreteWasmType(info.type) && isConcreteWasmType(curr->type)) {
+      if (isConcreteType(info.type) && isConcreteType(curr->type)) {
         shouldBeEqual(curr->type, info.type, curr, "block+breaks must have right type if breaks return a value");
       }
-      if (isConcreteWasmType(curr->type) && info.arity && info.type != unreachable) {
+      if (isConcreteType(curr->type) && info.arity && info.type != unreachable) {
         shouldBeEqual(curr->type, info.type, curr, "block+breaks must have right type if breaks have arity");
       }
       shouldBeTrue(info.arity != Index(-1), curr, "break arities must match");
       if (curr->list.size() > 0) {
         auto last = curr->list.back()->type;
-        if (isConcreteWasmType(last) && info.type != unreachable) {
+        if (isConcreteType(last) && info.type != unreachable) {
           shouldBeEqual(last, info.type, curr, "block+breaks must have right type if block ends with a reachable value");
         }
         if (last == none) {
@@ -322,29 +322,29 @@ void FunctionValidator::visitBlock(Block *curr) {
   }
   if (curr->list.size() > 1) {
     for (Index i = 0; i < curr->list.size() - 1; i++) {
-      if (!shouldBeTrue(!isConcreteWasmType(curr->list[i]->type), curr, "non-final block elements returning a value must be drop()ed (binaryen's autodrop option might help you)") && !info.quiet) {
+      if (!shouldBeTrue(!isConcreteType(curr->list[i]->type), curr, "non-final block elements returning a value must be drop()ed (binaryen's autodrop option might help you)") && !info.quiet) {
         getStream() << "(on index " << i << ":\n" << curr->list[i] << "\n), type: " << curr->list[i]->type << "\n";
       }
     }
   }
   if (curr->list.size() > 0) {
     auto backType = curr->list.back()->type;
-    if (!isConcreteWasmType(curr->type)) {
-      shouldBeFalse(isConcreteWasmType(backType), curr, "if block is not returning a value, final element should not flow out a value");
+    if (!isConcreteType(curr->type)) {
+      shouldBeFalse(isConcreteType(backType), curr, "if block is not returning a value, final element should not flow out a value");
     } else {
-      if (isConcreteWasmType(backType)) {
+      if (isConcreteType(backType)) {
         shouldBeEqual(curr->type, backType, curr, "block with value and last element with value must match types");
       } else {
         shouldBeUnequal(backType, none, curr, "block with value must not have last element that is none");
       }
     }
   }
-  if (isConcreteWasmType(curr->type)) {
+  if (isConcreteType(curr->type)) {
     shouldBeTrue(curr->list.size() > 0, curr, "block with a value must not be empty");
   }
 }
 
-void FunctionValidator::visitLoop(Loop *curr) {
+void FunctionValidator::visitLoop(Loop* curr) {
   if (curr->name.is()) {
     noteLabelName(curr->name);
     breakTargets.erase(curr->name);
@@ -354,14 +354,14 @@ void FunctionValidator::visitLoop(Loop *curr) {
     }
   }
   if (curr->type == none) {
-    shouldBeFalse(isConcreteWasmType(curr->body->type), curr, "bad body for a loop that has no value");
+    shouldBeFalse(isConcreteType(curr->body->type), curr, "bad body for a loop that has no value");
   }
 }
 
-void FunctionValidator::visitIf(If *curr) {
+void FunctionValidator::visitIf(If* curr) {
   shouldBeTrue(curr->condition->type == unreachable || curr->condition->type == i32, curr, "if condition must be valid");
   if (!curr->ifFalse) {
-    shouldBeFalse(isConcreteWasmType(curr->ifTrue->type), curr, "if without else must not return a value in body");
+    shouldBeFalse(isConcreteType(curr->ifTrue->type), curr, "if without else must not return a value in body");
     if (curr->condition->type != unreachable) {
       shouldBeEqual(curr->type, none, curr, "if without else and reachable condition must be none");
     }
@@ -375,11 +375,11 @@ void FunctionValidator::visitIf(If *curr) {
         shouldBeEqual(curr->ifFalse->type, unreachable, curr, "unreachable if-else must have unreachable false");
       }
     }
-    if (isConcreteWasmType(curr->ifTrue->type)) {
+    if (isConcreteType(curr->ifTrue->type)) {
       shouldBeEqual(curr->type, curr->ifTrue->type, curr, "if type must match concrete ifTrue");
       shouldBeEqualOrFirstIsUnreachable(curr->ifFalse->type, curr->ifTrue->type, curr, "other arm must match concrete ifTrue");
     }
-    if (isConcreteWasmType(curr->ifFalse->type)) {
+    if (isConcreteType(curr->ifFalse->type)) {
       shouldBeEqual(curr->type, curr->ifFalse->type, curr, "if type must match concrete ifFalse");
       shouldBeEqualOrFirstIsUnreachable(curr->ifTrue->type, curr->ifFalse->type, curr, "other arm must match concrete ifFalse");
     }
@@ -387,7 +387,7 @@ void FunctionValidator::visitIf(If *curr) {
 }
 
 void FunctionValidator::noteBreak(Name name, Expression* value, Expression* curr) {
-  WasmType valueType = none;
+  Type valueType = none;
   Index arity = 0;
   if (value) {
     valueType = value->type;
@@ -412,14 +412,14 @@ void FunctionValidator::noteBreak(Name name, Expression* value, Expression* curr
     }
   }
 }
-void FunctionValidator::visitBreak(Break *curr) {
+void FunctionValidator::visitBreak(Break* curr) {
   noteBreak(curr->name, curr->value, curr);
   if (curr->condition) {
     shouldBeTrue(curr->condition->type == unreachable || curr->condition->type == i32, curr, "break condition must be i32");
   }
 }
 
-void FunctionValidator::visitSwitch(Switch *curr) {
+void FunctionValidator::visitSwitch(Switch* curr) {
   for (auto& target : curr->targets) {
     noteBreak(target, curr->value, curr);
   }
@@ -427,7 +427,7 @@ void FunctionValidator::visitSwitch(Switch *curr) {
   shouldBeTrue(curr->condition->type == unreachable || curr->condition->type == i32, curr, "br_table condition must be i32");
 }
 
-void FunctionValidator::visitCall(Call *curr) {
+void FunctionValidator::visitCall(Call* curr) {
   if (!info.validateGlobally) return;
   auto* target = getModule()->getFunctionOrNull(curr->target);
   if (!shouldBeTrue(!!target, curr, "call target must exist")) {
@@ -444,7 +444,7 @@ void FunctionValidator::visitCall(Call *curr) {
   }
 }
 
-void FunctionValidator::visitCallImport(CallImport *curr) {
+void FunctionValidator::visitCallImport(CallImport* curr) {
   if (!info.validateGlobally) return;
   auto* import = getModule()->getImportOrNull(curr->target);
   if (!shouldBeTrue(!!import, curr, "call_import target must exist")) return;
@@ -458,7 +458,7 @@ void FunctionValidator::visitCallImport(CallImport *curr) {
   }
 }
 
-void FunctionValidator::visitCallIndirect(CallIndirect *curr) {
+void FunctionValidator::visitCallIndirect(CallIndirect* curr) {
   if (!info.validateGlobally) return;
   auto* type = getModule()->getFunctionTypeOrNull(curr->fullType);
   if (!shouldBeTrue(!!type, curr, "call_indirect type must exist")) return;
@@ -473,10 +473,10 @@ void FunctionValidator::visitCallIndirect(CallIndirect *curr) {
 
 void FunctionValidator::visitGetLocal(GetLocal* curr) {
   shouldBeTrue(curr->index < getFunction()->getNumLocals(), curr, "get_local index must be small enough");
-  shouldBeTrue(isConcreteWasmType(curr->type), curr, "get_local must have a valid type - check what you provided when you constructed the node");
+  shouldBeTrue(isConcreteType(curr->type), curr, "get_local must have a valid type - check what you provided when you constructed the node");
 }
 
-void FunctionValidator::visitSetLocal(SetLocal *curr) {
+void FunctionValidator::visitSetLocal(SetLocal* curr) {
   shouldBeTrue(curr->index < getFunction()->getNumLocals(), curr, "set_local index must be small enough");
   if (curr->value->type != unreachable) {
     if (curr->type != none) { // tee is ok anyhow
@@ -491,15 +491,16 @@ void FunctionValidator::visitGetGlobal(GetGlobal* curr) {
   shouldBeTrue(getModule()->getGlobalOrNull(curr->name) || getModule()->getImportOrNull(curr->name), curr, "get_global name must be valid");
 }
 
-void FunctionValidator::visitSetGlobal(SetGlobal *curr) {
+void FunctionValidator::visitSetGlobal(SetGlobal* curr) {
   if (!info.validateGlobally) return;
   auto* global = getModule()->getGlobalOrNull(curr->name);
-  shouldBeTrue(global, curr, "set_global name must be valid (and not an import; imports can't be modified)");
-  shouldBeTrue(global->mutable_, curr, "set_global global must be mutable");
-  shouldBeEqualOrFirstIsUnreachable(curr->value->type, global->type, curr, "set_global value must have right type");
+  if (shouldBeTrue(global != NULL, curr, "set_global name must be valid (and not an import; imports can't be modified)")) {
+    shouldBeTrue(global->mutable_, curr, "set_global global must be mutable");
+    shouldBeEqualOrFirstIsUnreachable(curr->value->type, global->type, curr, "set_global value must have right type");
+  }
 }
 
-void FunctionValidator::visitLoad(Load *curr) {
+void FunctionValidator::visitLoad(Load* curr) {
   if (curr->isAtomic) shouldBeTrue(info.features & Feature::Atomics, curr, "Atomic operation (atomics are disabled)");
   shouldBeFalse(curr->isAtomic && !getModule()->memory.shared, curr, "Atomic operation with non-shared memory");
   validateMemBytes(curr->bytes, curr->type, curr);
@@ -508,7 +509,7 @@ void FunctionValidator::visitLoad(Load *curr) {
   if (curr->isAtomic) shouldBeFalse(curr->signed_, curr, "atomic loads must be unsigned");
 }
 
-void FunctionValidator::visitStore(Store *curr) {
+void FunctionValidator::visitStore(Store* curr) {
   if (curr->isAtomic) shouldBeTrue(info.features & Feature::Atomics, curr, "Atomic operation (atomics are disabled)");
   shouldBeFalse(curr->isAtomic && !getModule()->memory.shared, curr, "Atomic operation with non-shared memory");
   validateMemBytes(curr->bytes, curr->valueType, curr);
@@ -558,7 +559,7 @@ void FunctionValidator::visitAtomicWake(AtomicWake* curr) {
   shouldBeEqualOrFirstIsUnreachable(curr->wakeCount->type, i32, curr, "AtomicWake wakeCount type must be i32");
 }
 
-void FunctionValidator::validateMemBytes(uint8_t bytes, WasmType type, Expression* curr) {
+void FunctionValidator::validateMemBytes(uint8_t bytes, Type type, Expression* curr) {
   switch (bytes) {
     case 1:
     case 2:
@@ -567,7 +568,7 @@ void FunctionValidator::validateMemBytes(uint8_t bytes, WasmType type, Expressio
       // if we have a concrete type for the load, then we know the size of the mem operation and
       // can validate it
       if (type != unreachable) {
-        shouldBeEqual(getWasmTypeSize(type), 8U, curr, "8-byte mem operations are only allowed with 8-byte wasm types");
+        shouldBeEqual(getTypeSize(type), 8U, curr, "8-byte mem operations are only allowed with 8-byte wasm types");
       }
       break;
     }
@@ -575,7 +576,7 @@ void FunctionValidator::validateMemBytes(uint8_t bytes, WasmType type, Expressio
   }
 }
 
-void FunctionValidator::visitBinary(Binary *curr) {
+void FunctionValidator::visitBinary(Binary* curr) {
   if (curr->left->type != unreachable && curr->right->type != unreachable) {
     shouldBeEqual(curr->left->type, curr->right->type, curr, "binary child types must be equal");
   }
@@ -672,7 +673,7 @@ void FunctionValidator::visitBinary(Binary *curr) {
   }
 }
 
-void FunctionValidator::visitUnary(Unary *curr) {
+void FunctionValidator::visitUnary(Unary* curr) {
   shouldBeUnequal(curr->value->type, none, curr, "unaries must not receive a none as their input");
   if (curr->value->type == unreachable) return; // nothing to check
   switch (curr->op) {
@@ -764,7 +765,7 @@ void FunctionValidator::visitSelect(Select* curr) {
 }
 
 void FunctionValidator::visitDrop(Drop* curr) {
-  shouldBeTrue(isConcreteWasmType(curr->value->type) || curr->value->type == unreachable, curr, "can only drop a valid value");
+  shouldBeTrue(isConcreteType(curr->value->type) || curr->value->type == unreachable, curr, "can only drop a valid value");
 }
 
 void FunctionValidator::visitReturn(Return* curr) {
@@ -793,7 +794,13 @@ void FunctionValidator::visitHost(Host* curr) {
   }
 }
 
-void FunctionValidator::visitFunction(Function *curr) {
+void FunctionValidator::visitFunction(Function* curr) {
+  for (auto type : curr->params) {
+    shouldBeTrue(isConcreteType(type), curr, "params must be concretely typed");
+  }
+  for (auto type : curr->vars) {
+    shouldBeTrue(isConcreteType(type), curr, "vars must be concretely typed");
+  }
   // if function has no result, it is ignored
   // if body is unreachable, it might be e.g. a return
   if (curr->body->type != unreachable) {
@@ -805,6 +812,12 @@ void FunctionValidator::visitFunction(Function *curr) {
   shouldBeTrue(breakTargets.empty(), curr->body, "all named break targets must exist");
   returnType = unreachable;
   labelNames.clear();
+  // if function has a named type, it must match up with the function's params and result
+  if (info.validateGlobally && curr->type.is()) {
+    auto* ft = getModule()->getFunctionType(curr->type);
+    shouldBeTrue(ft->params == curr->params, curr->name, "function params must match its declared type");
+    shouldBeTrue(ft->result == curr->result, curr->name, "function result must match its declared type");
+  }
   // expressions must not be seen more than once
   struct Walker : public PostWalker<Walker, UnifiedExpressionVisitor<Walker>> {
     std::unordered_set<Expression*>& seen;
@@ -840,7 +853,7 @@ static bool checkOffset(Expression* curr, Address add, Address max) {
   return offset + add <= max;
 }
 
-void FunctionValidator::validateAlignment(size_t align, WasmType type, Index bytes,
+void FunctionValidator::validateAlignment(size_t align, Type type, Index bytes,
                                       bool isAtomic, Expression* curr) {
   if (isAtomic) {
     shouldBeEqual(align, (size_t)bytes, curr, "atomic accesses must have natural alignment");
@@ -891,9 +904,9 @@ static void validateBinaryenIR(Module& wasm, ValidationInfo& info) {
         //
         // The block has an added type, not derived from the ast itself, so it is
         // ok for it to be either i32 or unreachable.
-        if (!(isConcreteWasmType(oldType) && newType == unreachable)) {
+        if (!(isConcreteType(oldType) && newType == unreachable)) {
           std::ostringstream ss;
-          ss << "stale type found in " << (getFunction() ? getFunction()->name : Name("(global scope)")) << " on " << curr << "\n(marked as " << printWasmType(oldType) << ", should be " << printWasmType(newType) << ")\n";
+          ss << "stale type found in " << (getFunction() ? getFunction()->name : Name("(global scope)")) << " on " << curr << "\n(marked as " << printType(oldType) << ", should be " << printType(newType) << ")\n";
           info.fail(ss.str(), curr, getFunction());
         }
         curr->type = oldType;
@@ -912,7 +925,7 @@ static void validateImports(Module& module, ValidationInfo& info) {
       if (info.validateWeb) {
         auto* functionType = module.getFunctionType(curr->functionType);
         info.shouldBeUnequal(functionType->result, i64, curr->name, "Imported function must not have i64 return type");
-        for (WasmType param : functionType->params) {
+        for (Type param : functionType->params) {
           info.shouldBeUnequal(param, i64, curr->name, "Imported function must not have i64 parameters");
         }
       }
@@ -1018,10 +1031,10 @@ static void validateModule(Module& module, ValidationInfo& info) {
 // perhaps by moving some of the pass infrastructure into libsupport.
 bool WasmValidator::validate(Module& module, FeatureSet features, Flags flags) {
   ValidationInfo info;
-  info.validateWeb = flags & Web;
-  info.validateGlobally = flags & Globally;
+  info.validateWeb = (flags & Web) != 0;
+  info.validateGlobally = (flags & Globally) != 0;
   info.features = features;
-  info.quiet = flags & Quiet;
+  info.quiet = (flags & Quiet) != 0;
   // parallel wasm logic validation
   PassRunner runner(&module);
   runner.add<FunctionValidator>(&info);
