@@ -26,12 +26,13 @@
 #include <numeric>
 
 #include "asmjs/shared-constants.h"
+#include "asmjs/asmangle.h"
 #include "wasm.h"
 #include "wasm-builder.h"
 #include "emscripten-optimizer/optimizer.h"
 #include "mixed_arena.h"
 #include "asm_v_wasm.h"
-#include "ast_utils.h"
+#include "ir/utils.h"
 #include "passes/passes.h"
 
 namespace wasm {
@@ -135,14 +136,14 @@ public:
   Ref processAsserts(Element& e, SExpressionWasmBuilder& sexpBuilder);
 
   // Get a temp var.
-  IString getTemp(WasmType type, Function* func) {
+  IString getTemp(Type type, Function* func) {
     IString ret;
     if (frees[type].size() > 0) {
       ret = frees[type].back();
       frees[type].pop_back();
     } else {
       size_t index = temps[type]++;
-      ret = IString((std::string("wasm2asm_") + printWasmType(type) + "$" +
+      ret = IString((std::string("wasm2asm_") + printType(type) + "$" +
                      std::to_string(index)).c_str(), false);
     }
     if (func->localIndices.find(ret) == func->localIndices.end()) {
@@ -152,31 +153,20 @@ public:
   }
 
   // Free a temp var.
-  void freeTemp(WasmType type, IString temp) {
+  void freeTemp(Type type, IString temp) {
     frees[type].push_back(temp);
   }
 
-  static IString fromName(Name name) {
-    // TODO: more clever name fixing, including checking we do not collide
-    const char* str = name.str;
-    // check the various issues, and recurse so we check the others
-    if (strchr(str, '-')) {
-      char* mod = strdup(str);
-      str = mod;
-      while (*mod) {
-        if (*mod == '-') *mod = '_';
-        mod++;
-      }
-      IString result = fromName(IString(str, false));
-      free((void*)str);
-      return result;
+  IString fromName(Name name) {
+    // TODO: checking names do not collide after mangling
+    auto it = mangledNames.find(name.c_str());
+    if (it != mangledNames.end()) {
+      return it->second;
     }
-    if (isdigit(str[0]) || strcmp(str, "if") == 0) {
-      std::string prefixed = "$$";
-      prefixed += name.str;
-      return fromName(IString(prefixed.c_str(), false));
-    }
-    return name;
+    auto mangled = asmangle(std::string(name.c_str()));
+    IString ret(mangled.c_str(), false);
+    mangledNames[name.c_str()] = ret;
+    return ret;
   }
 
   void setStatement(Expression* curr) {
@@ -202,14 +192,23 @@ private:
   // Expressions that will be a statement.
   std::set<Expression*> willBeStatement;
 
+  // Mangled names cache by interned names.
+  // Utilizes the usually reused underlying cstring's pointer as the key.
+  std::unordered_map<const char*, IString> mangledNames;
+
   // All our function tables have the same size TODO: optimize?
   size_t tableSize;
 
+  bool almostASM = false;
+
   void addBasics(Ref ast);
   void addImport(Ref ast, Import* import);
   void addTables(Ref ast, Module* wasm);
   void addExports(Ref ast, Module* wasm);
+  void addGlobal(Ref ast, Global* global);
   void addWasmCompatibilityFuncs(Module* wasm);
+  void setNeedsAlmostASM(const char *reason);
+  void addMemoryGrowthFuncs(Ref ast);
   bool isAssertHandled(Element& e);
   Ref makeAssertReturnFunc(SExpressionWasmBuilder& sexpBuilder,
                            Builder& wasmBuilder,
@@ -227,12 +226,12 @@ static Function* makeCtzFunc(MixedArena& allocator, UnaryOp op) {
   Builder b(allocator);
   // if eqz(x) then 32 else (32 - clz(x ^ (x - 1)))
   bool is32Bit = (op == CtzInt32);
-  Name funcName    = is32Bit ? Name(CTZ32) : Name(CTZ64);
+  Name funcName    = is32Bit ? Name(WASM_CTZ32) : Name(WASM_CTZ64);
   BinaryOp subOp   = is32Bit ? SubInt32    : SubInt64;
   BinaryOp xorOp   = is32Bit ? XorInt32    : XorInt64;
   UnaryOp  clzOp   = is32Bit ? ClzInt32    : ClzInt64;
   UnaryOp  eqzOp   = is32Bit ? EqZInt32    : EqZInt64;
-  WasmType argType = is32Bit ? i32         : i64;
+  Type argType = is32Bit ? i32         : i64;
   Binary* xorExp = b.makeBinary(
     xorOp,
     b.makeGetLocal(0, i32),
@@ -270,12 +269,12 @@ static Function* makePopcntFunc(MixedArena& allocator, UnaryOp op) {
   // popcnt implemented as:
   // int c; for (c = 0; x != 0; c++) { x = x & (x - 1) }; return c
   bool is32Bit = (op == PopcntInt32);
-  Name funcName    = is32Bit ? Name(POPCNT32) : Name(POPCNT64);
+  Name funcName    = is32Bit ? Name(WASM_POPCNT32) : Name(WASM_POPCNT64);
   BinaryOp addOp   = is32Bit ? AddInt32       : AddInt64;
   BinaryOp subOp   = is32Bit ? SubInt32       : SubInt64;
   BinaryOp andOp   = is32Bit ? AndInt32       : AndInt64;
   UnaryOp  eqzOp   = is32Bit ? EqZInt32       : EqZInt64;
-  WasmType argType = is32Bit ? i32            : i64;
+  Type argType = is32Bit ? i32            : i64;
   Name loopName("l");
   Name blockName("b");
   Break* brIf = b.makeBreak(
@@ -328,8 +327,8 @@ Function* makeRotFunc(MixedArena& allocator, BinaryOp op) {
   // where k is shift modulo w. reverse shifts for right rotate
   bool is32Bit = (op == RotLInt32 || op == RotRInt32);
   bool isLRot  = (op == RotLInt32 || op == RotLInt64);
-  static Name names[2][2] = {{Name(ROTR64), Name(ROTR32)},
-                             {Name(ROTL64), Name(ROTL32)}};
+  static Name names[2][2] = {{Name(WASM_ROTR64), Name(WASM_ROTR32)},
+                             {Name(WASM_ROTL64), Name(WASM_ROTL32)}};
   static BinaryOp shifters[2][2] = {{ShrUInt64, ShrUInt32},
                                     {ShlInt64, ShlInt32}};
   Name funcName = names[isLRot][is32Bit];
@@ -338,7 +337,7 @@ Function* makeRotFunc(MixedArena& allocator, BinaryOp op) {
   BinaryOp orOp    = is32Bit ? OrInt32  : OrInt64;
   BinaryOp andOp   = is32Bit ? AndInt32 : AndInt64;
   BinaryOp subOp   = is32Bit ? SubInt32 : SubInt64;
-  WasmType argType = is32Bit ? i32      : i64;
+  Type argType = is32Bit ? i32      : i64;
   Literal widthMask =
       is32Bit ? Literal(int32_t(32 - 1)) : Literal(int64_t(64 - 1));
   Literal width =
@@ -386,7 +385,9 @@ Ref Wasm2AsmBuilder::processWasm(Module* wasm) {
   PassRunner runner(wasm);
   runner.add<AutoDrop>();
   runner.add("i64-to-i32-lowering");
-  runner.add("flatten-control-flow");
+  runner.add("flatten");
+  runner.add("simplify-locals-notee-nostructure");
+  runner.add("reorder-locals");
   runner.add("vacuum");
   runner.setDebug(flags.debug);
   runner.run();
@@ -413,6 +414,10 @@ Ref Wasm2AsmBuilder::processWasm(Module* wasm) {
     pow2ed <<= 1;
   }
   tableSize = pow2ed;
+  // globals
+  for (auto& global : wasm->globals) {
+    addGlobal(asmFunc[3], global.get());
+  }
   // functions
   for (auto& func : wasm->functions) {
     asmFunc[3]->push_back(processFunction(func.get()));
@@ -468,6 +473,8 @@ void Wasm2AsmBuilder::addBasics(Ref ast) {
   addMath(MATH_FROUND, FROUND);
   addMath(MATH_ABS, ABS);
   addMath(MATH_CLZ32, CLZ32);
+  addMath(MATH_MIN, MIN);
+  addMath(MATH_MAX, MAX);
 }
 
 void Wasm2AsmBuilder::addImport(Ref ast, Import* import) {
@@ -521,15 +528,91 @@ void Wasm2AsmBuilder::addTables(Ref ast, Module* wasm) {
 void Wasm2AsmBuilder::addExports(Ref ast, Module* wasm) {
   Ref exports = ValueBuilder::makeObject();
   for (auto& export_ : wasm->exports) {
-    ValueBuilder::appendToObject(
-      exports,
-      fromName(export_->name),
-      ValueBuilder::makeName(fromName(export_->value))
-    );
+    if (export_->kind == ExternalKind::Function) {
+      ValueBuilder::appendToObject(
+        exports,
+        fromName(export_->name),
+        ValueBuilder::makeName(fromName(export_->value))
+      );
+    }
+    if (export_->kind == ExternalKind::Memory) {
+      setNeedsAlmostASM("memory export");
+      Ref descs = ValueBuilder::makeObject();
+      Ref growDesc = ValueBuilder::makeObject();
+      ValueBuilder::appendToObject(
+        descs,
+        IString("grow"),
+        growDesc);
+      ValueBuilder::appendToObject(
+        growDesc,
+        IString("value"),
+        ValueBuilder::makeName(WASM_GROW_MEMORY));
+      Ref bufferDesc = ValueBuilder::makeObject();
+      Ref bufferGetter = ValueBuilder::makeFunction(IString(""));
+      bufferGetter[3]->push_back(ValueBuilder::makeReturn(
+        ValueBuilder::makeName(BUFFER)
+      ));
+      ValueBuilder::appendToObject(
+        bufferDesc,
+        IString("get"),
+        bufferGetter);
+      ValueBuilder::appendToObject(
+        descs,
+        IString("buffer"),
+        bufferDesc);
+      Ref memory = ValueBuilder::makeCall(
+        ValueBuilder::makeDot(ValueBuilder::makeName(IString("Object")), IString("create")),
+        ValueBuilder::makeDot(ValueBuilder::makeName(IString("Object")), IString("prototype")));
+      ValueBuilder::appendToCall(
+        memory,
+        descs);
+      ValueBuilder::appendToObject(
+        exports,
+        fromName(export_->name),
+        memory);
+    }
+  }
+  if (almostASM) {
+    // replace "use asm"
+    ast[0] = ValueBuilder::makeStatement(ValueBuilder::makeString(ALMOST_ASM));
+    addMemoryGrowthFuncs(ast);
   }
   ast->push_back(ValueBuilder::makeStatement(ValueBuilder::makeReturn(exports)));
 }
 
+void Wasm2AsmBuilder::addGlobal(Ref ast, Global* global) {
+  if (auto* const_ = global->init->dynCast<Const>()) {
+    Ref theValue;
+    switch (const_->type) {
+      case Type::i32: {
+        theValue = ValueBuilder::makeInt(const_->value.geti32());
+        break;
+      }
+      case Type::f32: {
+        theValue = ValueBuilder::makeCall(MATH_FROUND,
+          makeAsmCoercion(ValueBuilder::makeDouble(const_->value.getf32()), ASM_DOUBLE)
+        );
+        break;
+      }
+      case Type::f64: {
+        theValue = makeAsmCoercion(ValueBuilder::makeDouble(const_->value.getf64()), ASM_DOUBLE);
+        break;
+      }
+      default: {
+        assert(false && "Global const type not supported");
+      }
+    }
+    Ref theVar = ValueBuilder::makeVar();
+    ast->push_back(theVar);
+    ValueBuilder::appendToVar(theVar,
+      fromName(global->name),
+      theValue
+    );
+  } else {
+    assert(false && "Global init type not supported");
+  }
+}
+
 Ref Wasm2AsmBuilder::processFunction(Function* func) {
   if (flags.debug) {
     static int fns = 0;
@@ -571,11 +654,12 @@ Ref Wasm2AsmBuilder::processFunction(Function* func) {
     );
   };
   scanFunctionBody(func->body);
-  bool isBodyBlock = (func->body->_id == Expression::BlockId);
+  bool isBodyBlock = func->body->is<Block>();
   ExpressionList* stats = isBodyBlock ?
       &static_cast<Block*>(func->body)->list : nullptr;
   bool endsInReturn =
-      (isBodyBlock && ((*stats)[stats->size()-1]->_id == Expression::ReturnId));
+      (isBodyBlock && ((*stats)[stats->size()-1]->is<Return>())) ||
+    func->body->is<Return>();
   if (endsInReturn) {
     // return already taken care of
     flattenAppend(ret, processFunctionBody(func, NO_RESULT));
@@ -722,13 +806,13 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
     // A scoped temporary variable.
     struct ScopedTemp {
       Wasm2AsmBuilder* parent;
-      WasmType type;
+      Type type;
       IString temp;
       bool needFree;
       // @param possible if provided, this is a variable we can use as our temp. it has already been
       //                 allocated in a higher scope, and we can just assign to it as our result is
       //                 going there anyhow.
-      ScopedTemp(WasmType type, Wasm2AsmBuilder* parent, Function* func,
+      ScopedTemp(Type type, Wasm2AsmBuilder* parent, Function* func,
                  IString possible = NO_RESULT) : parent(parent), type(type) {
         assert(possible != EXPRESSION_RESULT);
         if (possible == NO_RESULT) {
@@ -766,7 +850,7 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
     }
 
     // this result is for an asm expression slot, but it might be a statement
-    Ref visitForExpression(Expression* curr, WasmType type, IString& tempName) {
+    Ref visitForExpression(Expression* curr, Type type, IString& tempName) {
       if (isStatement(curr)) {
         ScopedTemp temp(type, parent, func);
         tempName = temp.temp;
@@ -914,7 +998,7 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
       return ret;
     }
 
-    Ref makeStatementizedCall(ExpressionList& operands, Ref ret, Ref theCall, IString result, WasmType type) {
+    Ref makeStatementizedCall(ExpressionList& operands, Ref ret, Ref theCall, IString result, Type type) {
       std::vector<ScopedTemp*> temps; // TODO: utility class, with destructor?
       for (auto& operand : operands) {
         temps.push_back(new ScopedTemp(operand->type, parent, func));
@@ -1267,15 +1351,26 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
         case i32: {
           switch (curr->op) {
             case ClzInt32:
-              return ValueBuilder::makeCall(MATH_CLZ32,
-                                            visit(curr->value,
-                                                  EXPRESSION_RESULT));
+              return ValueBuilder::makeCall(
+                MATH_CLZ32,
+                visit(curr->value, EXPRESSION_RESULT)
+              );
             case CtzInt32:
-              return ValueBuilder::makeCall(CTZ32, visit(curr->value,
-                                                         EXPRESSION_RESULT));
+              return makeSigning(
+                ValueBuilder::makeCall(
+                  WASM_CTZ32,
+                  visit(curr->value, EXPRESSION_RESULT)
+                ),
+                ASM_SIGNED
+              );
             case PopcntInt32:
-              return ValueBuilder::makeCall(POPCNT32, visit(curr->value,
-                                                            EXPRESSION_RESULT));
+              return makeSigning(
+                ValueBuilder::makeCall(
+                  WASM_POPCNT32,
+                  visit(curr->value, EXPRESSION_RESULT)
+                ),
+                ASM_SIGNED
+              );
             case EqZInt32:
               return ValueBuilder::makeBinary(
                   makeAsmCoercion(visit(curr->value,
@@ -1341,6 +1436,12 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
                 visit(curr->value, EXPRESSION_RESULT)
               );
               break;
+            case PromoteFloat32:
+              return makeAsmCoercion(visit(curr->value, EXPRESSION_RESULT),
+                                     ASM_DOUBLE);
+            case DemoteFloat64:
+              return makeAsmCoercion(visit(curr->value, EXPRESSION_RESULT),
+                                     ASM_FLOAT);
             // TODO: more complex unary conversions
             default:
               std::cerr << "Unhandled unary float operator: " << curr
@@ -1379,110 +1480,169 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
       Ref left = visit(curr->left, EXPRESSION_RESULT);
       Ref right = visit(curr->right, EXPRESSION_RESULT);
       Ref ret;
-      switch (curr->op) {
-        case AddInt32:
-          ret = ValueBuilder::makeBinary(left, PLUS, right);
-          break;
-        case SubInt32:
-          ret = ValueBuilder::makeBinary(left, MINUS, right);
-          break;
-        case MulInt32: {
-          if (curr->type == i32) {
-            // TODO: when one operand is a small int, emit a multiply
-            return ValueBuilder::makeCall(MATH_IMUL, left, right);
-          } else {
-            return ValueBuilder::makeBinary(left, MUL, right);
+      switch (curr->type) {
+        case i32: {
+          switch (curr->op) {
+            case AddInt32:
+              ret = ValueBuilder::makeBinary(left, PLUS, right);
+              break;
+            case SubInt32:
+              ret = ValueBuilder::makeBinary(left, MINUS, right);
+              break;
+            case MulInt32: {
+              if (curr->type == i32) {
+                // TODO: when one operand is a small int, emit a multiply
+                return ValueBuilder::makeCall(MATH_IMUL, left, right);
+              } else {
+                return ValueBuilder::makeBinary(left, MUL, right);
+              }
+            }
+            case DivSInt32:
+              ret = ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), DIV,
+                                             makeSigning(right, ASM_SIGNED));
+              break;
+            case DivUInt32:
+              ret = ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), DIV,
+                                             makeSigning(right, ASM_UNSIGNED));
+              break;
+            case RemSInt32:
+              ret = ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), MOD,
+                                             makeSigning(right, ASM_SIGNED));
+              break;
+            case RemUInt32:
+              ret = ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), MOD,
+                                             makeSigning(right, ASM_UNSIGNED));
+              break;
+            case AndInt32:
+              ret = ValueBuilder::makeBinary(left, AND, right);
+              break;
+            case OrInt32:
+              ret = ValueBuilder::makeBinary(left, OR, right);
+              break;
+            case XorInt32:
+              ret = ValueBuilder::makeBinary(left, XOR, right);
+              break;
+            case ShlInt32:
+              ret = ValueBuilder::makeBinary(left, LSHIFT, right);
+              break;
+            case ShrUInt32:
+              ret = ValueBuilder::makeBinary(left, TRSHIFT, right);
+              break;
+            case ShrSInt32:
+              ret = ValueBuilder::makeBinary(left, RSHIFT, right);
+              break;
+            case EqInt32: {
+              // TODO: check if this condition is still valid/necessary
+              if (curr->left->type == i32) {
+                return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), EQ,
+                                                makeSigning(right, ASM_SIGNED));
+              } else {
+                return ValueBuilder::makeBinary(left, EQ, right);
+              }
+            }
+            case NeInt32: {
+              if (curr->left->type == i32) {
+                return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), NE,
+                                                makeSigning(right, ASM_SIGNED));
+              } else {
+                return ValueBuilder::makeBinary(left, NE, right);
+              }
+            }
+            case LtSInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), LT,
+                                              makeSigning(right, ASM_SIGNED));
+            case LtUInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), LT,
+                                              makeSigning(right, ASM_UNSIGNED));
+            case LeSInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), LE,
+                                              makeSigning(right, ASM_SIGNED));
+            case LeUInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), LE,
+                                              makeSigning(right, ASM_UNSIGNED));
+            case GtSInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), GT,
+                                              makeSigning(right, ASM_SIGNED));
+            case GtUInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), GT,
+                                              makeSigning(right, ASM_UNSIGNED));
+            case GeSInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), GE,
+                                              makeSigning(right, ASM_SIGNED));
+            case GeUInt32:
+              return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), GE,
+                                              makeSigning(right, ASM_UNSIGNED));
+            case RotLInt32:
+              return makeSigning(ValueBuilder::makeCall(WASM_ROTL32, left, right),
+                                 ASM_SIGNED);
+            case RotRInt32:
+              return makeSigning(ValueBuilder::makeCall(WASM_ROTR32, left, right),
+                                 ASM_SIGNED);
+            case EqFloat32:
+            case EqFloat64:
+              return ValueBuilder::makeBinary(left, EQ, right);
+            case NeFloat32:
+            case NeFloat64:
+              return ValueBuilder::makeBinary(left, NE, right);
+            case GeFloat32:
+            case GeFloat64:
+              return ValueBuilder::makeBinary(left, GE, right);
+            case GtFloat32:
+            case GtFloat64:
+              return ValueBuilder::makeBinary(left, GT, right);
+            case LeFloat32:
+            case LeFloat64:
+              return ValueBuilder::makeBinary(left, LE, right);
+            case LtFloat32:
+            case LtFloat64:
+              return ValueBuilder::makeBinary(left, LT, right);
+            default: {
+              std::cerr << "Unhandled i32 binary operator: " << curr << std::endl;
+              abort();
+            }
           }
+          break;
         }
-        case DivSInt32:
-          ret = ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), DIV,
-                                         makeSigning(right, ASM_SIGNED));
-          break;
-        case DivUInt32:
-          ret = ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), DIV,
-                                         makeSigning(right, ASM_UNSIGNED));
-          break;
-        case RemSInt32:
-          ret = ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), MOD,
-                                         makeSigning(right, ASM_SIGNED));
-          break;
-        case RemUInt32:
-          ret = ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), MOD,
-                                         makeSigning(right, ASM_UNSIGNED));
-          break;
-        case AndInt32:
-          ret = ValueBuilder::makeBinary(left, AND, right);
-          break;
-        case OrInt32:
-          ret = ValueBuilder::makeBinary(left, OR, right);
-          break;
-        case XorInt32:
-          ret = ValueBuilder::makeBinary(left, XOR, right);
-          break;
-        case ShlInt32:
-          ret = ValueBuilder::makeBinary(left, LSHIFT, right);
-          break;
-        case ShrUInt32:
-          ret = ValueBuilder::makeBinary(left, TRSHIFT, right);
-          break;
-        case ShrSInt32:
-          ret = ValueBuilder::makeBinary(left, RSHIFT, right);
-          break;
-        case MinFloat32:
-          ret = ValueBuilder::makeCall(MATH_MIN, left, right);
-          break;
-        case MaxFloat32:
-          ret = ValueBuilder::makeCall(MATH_MAX, left, right);
-          break;
-        case EqInt32: {
-          // TODO: check if this condition is still valid/necessary
-          if (curr->left->type == i32) {
-            return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), EQ,
-                                            makeSigning(right, ASM_SIGNED));
-          } else {
-            return ValueBuilder::makeBinary(left, EQ, right);
+        case f32:
+        case f64:
+	  switch (curr->op) {
+	    case AddFloat32:
+	    case AddFloat64:
+	      ret = ValueBuilder::makeBinary(left, PLUS, right);
+              break;
+            case SubFloat32:
+            case SubFloat64:
+              ret = ValueBuilder::makeBinary(left, MINUS, right);
+              break;
+            case MulFloat32:
+            case MulFloat64:
+              ret = ValueBuilder::makeBinary(left, MUL, right);
+              break;
+            case DivFloat32:
+            case DivFloat64:
+              ret = ValueBuilder::makeBinary(left, DIV, right);
+              break;
+            case MinFloat32:
+            case MinFloat64:
+              ret = ValueBuilder::makeCall(MATH_MIN, left, right);
+              break;
+            case MaxFloat32:
+            case MaxFloat64:
+              ret = ValueBuilder::makeCall(MATH_MAX, left, right);
+              break;
+            case CopySignFloat32:
+            case CopySignFloat64:
+            default:
+              std::cerr << "Unhandled binary float operator: " << curr << std::endl;
+              abort();
           }
-        }
-        case NeInt32: {
-          if (curr->left->type == i32) {
-            return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), NE,
-                                            makeSigning(right, ASM_SIGNED));
-          } else {
-            return ValueBuilder::makeBinary(left, NE, right);
+          if (curr->type == f32) {
+            return makeAsmCoercion(ret, ASM_FLOAT);
           }
-        }
-        case LtSInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), LT,
-                                          makeSigning(right, ASM_SIGNED));
-        case LtUInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), LT,
-                                          makeSigning(right, ASM_UNSIGNED));
-        case LeSInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), LE,
-                                          makeSigning(right, ASM_SIGNED));
-        case LeUInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), LE,
-                                          makeSigning(right, ASM_UNSIGNED));
-        case GtSInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), GT,
-                                          makeSigning(right, ASM_SIGNED));
-        case GtUInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), GT,
-                                          makeSigning(right, ASM_UNSIGNED));
-        case GeSInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_SIGNED), GE,
-                                          makeSigning(right, ASM_SIGNED));
-        case GeUInt32:
-          return ValueBuilder::makeBinary(makeSigning(left, ASM_UNSIGNED), GE,
-                                          makeSigning(right, ASM_UNSIGNED));
-        case RotLInt32:
-          return ValueBuilder::makeCall(ROTL32, left, right);
-        case RotRInt32:
-          return ValueBuilder::makeCall(ROTR32, left, right);
-        default: {
-          std::cerr << "Unhandled binary operator: " << curr << std::endl;
+          return ret;
+        default:
+          std::cerr << "Unhandled type in binary: " << curr << std::endl;
           abort();
-        }
       }
       return makeAsmCoercion(ret, wasmToAsmType(curr->type));
     }
@@ -1543,7 +1703,18 @@ Ref Wasm2AsmBuilder::processFunctionBody(Function* func, IString result) {
     }
 
     Ref visitHost(Host* curr) {
-      abort();
+      if (curr->op == HostOp::GrowMemory) {
+        parent->setNeedsAlmostASM("grow_memory op");
+        return ValueBuilder::makeCall(WASM_GROW_MEMORY,
+          makeAsmCoercion(
+            visit(curr->operands[0], EXPRESSION_RESULT),
+            wasmToAsmType(curr->operands[0]->type)));
+      }
+      if (curr->op == HostOp::CurrentMemory) {
+        parent->setNeedsAlmostASM("current_memory op");
+        return ValueBuilder::makeCall(WASM_CURRENT_MEMORY);
+      }
+      return ValueBuilder::makeCall(ABORT_FUNC);
     }
 
     Ref visitNop(Nop* curr) {
@@ -1611,7 +1782,7 @@ Ref Wasm2AsmBuilder::makeAssertReturnFunc(SExpressionWasmBuilder& sexpBuilder,
     }
   } else if (e.size() == 3) {
     Expression* expected = sexpBuilder.parseExpression(e[2]);
-    WasmType resType = expected->type;
+    Type resType = expected->type;
     actual->type = resType;
     BinaryOp eqOp;
     switch (resType) {
@@ -1682,6 +1853,164 @@ Ref Wasm2AsmBuilder::makeAssertTrapFunc(SExpressionWasmBuilder& sexpBuilder,
   return outerFunc;
 }
 
+void Wasm2AsmBuilder::setNeedsAlmostASM(const char *reason) {
+  if (!almostASM) {
+    almostASM = true;
+    std::cerr << "Switching to \"almost asm\" mode, reason: " << reason << std::endl;
+  }
+}
+
+void Wasm2AsmBuilder::addMemoryGrowthFuncs(Ref ast) {
+  Ref growMemoryFunc = ValueBuilder::makeFunction(WASM_GROW_MEMORY);
+  ValueBuilder::appendArgumentToFunction(growMemoryFunc, IString("pagesToAdd"));
+
+  growMemoryFunc[3]->push_back(
+    ValueBuilder::makeStatement(
+      ValueBuilder::makeBinary(
+        ValueBuilder::makeName(IString("pagesToAdd")), SET,
+        makeAsmCoercion(
+          ValueBuilder::makeName(IString("pagesToAdd")),
+          AsmType::ASM_INT
+        )
+      )
+    )
+  );
+
+  Ref oldPages = ValueBuilder::makeVar();
+  growMemoryFunc[3]->push_back(oldPages);
+  ValueBuilder::appendToVar(
+    oldPages,
+    IString("oldPages"),
+    makeAsmCoercion(ValueBuilder::makeCall(WASM_CURRENT_MEMORY), AsmType::ASM_INT));
+
+  Ref newPages = ValueBuilder::makeVar();
+  growMemoryFunc[3]->push_back(newPages);
+  ValueBuilder::appendToVar(
+    newPages,
+    IString("newPages"),
+    makeAsmCoercion(ValueBuilder::makeBinary(
+      ValueBuilder::makeName(IString("oldPages")),
+      PLUS,
+      ValueBuilder::makeName(IString("pagesToAdd"))
+    ), AsmType::ASM_INT));
+
+  Ref block = ValueBuilder::makeBlock();
+  growMemoryFunc[3]->push_back(ValueBuilder::makeIf(
+    ValueBuilder::makeBinary(
+      ValueBuilder::makeBinary(
+        ValueBuilder::makeName(IString("oldPages")),
+        LT,
+        ValueBuilder::makeName(IString("newPages"))
+      ),
+      IString("&&"),
+      ValueBuilder::makeBinary(
+        ValueBuilder::makeName(IString("newPages")),
+        LT,
+        ValueBuilder::makeInt(Memory::kMaxSize)
+      )
+    ), block, NULL));
+
+  Ref newBuffer = ValueBuilder::makeVar();
+  ValueBuilder::appendToBlock(block, newBuffer);
+  ValueBuilder::appendToVar(
+    newBuffer,
+    IString("newBuffer"),
+    ValueBuilder::makeNew(ValueBuilder::makeCall(
+      ARRAY_BUFFER,
+      ValueBuilder::makeCall(
+        MATH_IMUL,
+        ValueBuilder::makeName(IString("newPages")),
+        ValueBuilder::makeInt(Memory::kPageSize)))));
+
+  Ref newHEAP8 = ValueBuilder::makeVar();
+  ValueBuilder::appendToBlock(block, newHEAP8);
+  ValueBuilder::appendToVar(
+    newHEAP8,
+    IString("newHEAP8"),
+    ValueBuilder::makeNew(
+      ValueBuilder::makeCall(
+        ValueBuilder::makeDot(
+          ValueBuilder::makeName(GLOBAL),
+          INT8ARRAY
+        ),
+        ValueBuilder::makeName(IString("newBuffer"))
+      )
+    ));
+
+  ValueBuilder::appendToBlock(block,
+    ValueBuilder::makeCall(
+      ValueBuilder::makeDot(
+        ValueBuilder::makeName(IString("newHEAP8")),
+        IString("set")
+      ),
+      ValueBuilder::makeName(HEAP8)
+    )
+  );
+
+  ValueBuilder::appendToBlock(block,
+    ValueBuilder::makeBinary(
+      ValueBuilder::makeName(HEAP8),
+      SET,
+      ValueBuilder::makeName(IString("newHEAP8"))
+    )
+  );
+
+  auto setHeap = [&](IString name, IString view) {
+    ValueBuilder::appendToBlock(block,
+      ValueBuilder::makeBinary(
+        ValueBuilder::makeName(name),
+        SET,
+        ValueBuilder::makeNew(
+          ValueBuilder::makeCall(
+            ValueBuilder::makeDot(
+              ValueBuilder::makeName(GLOBAL),
+              view
+            ),
+            ValueBuilder::makeName(IString("newBuffer"))
+          )
+        )
+      )
+    );
+  };
+
+  setHeap(HEAP16, INT16ARRAY);
+  setHeap(HEAP32, INT32ARRAY);
+  setHeap(HEAPU8,  UINT8ARRAY);
+  setHeap(HEAPU16, UINT16ARRAY);
+  setHeap(HEAPU32, UINT32ARRAY);
+  setHeap(HEAPF32, FLOAT32ARRAY);
+  setHeap(HEAPF64, FLOAT64ARRAY);
+
+  ValueBuilder::appendToBlock(block,
+    ValueBuilder::makeBinary(
+      ValueBuilder::makeName(BUFFER),
+      SET,
+      ValueBuilder::makeName(IString("newBuffer"))
+    )
+  );
+
+  growMemoryFunc[3]->push_back(
+    ValueBuilder::makeReturn(
+      ValueBuilder::makeName(IString("oldPages"))));
+
+  Ref currentMemoryFunc = ValueBuilder::makeFunction(WASM_CURRENT_MEMORY);
+  currentMemoryFunc[3]->push_back(ValueBuilder::makeReturn(
+    makeAsmCoercion(
+      ValueBuilder::makeBinary(
+        ValueBuilder::makeDot(
+          ValueBuilder::makeName(BUFFER),
+          IString("byteLength")
+        ),
+        DIV,
+        ValueBuilder::makeInt(Memory::kPageSize)
+      ),
+      AsmType::ASM_INT
+    )
+  ));
+  ast->push_back(growMemoryFunc);
+  ast->push_back(currentMemoryFunc);
+}
+
 bool Wasm2AsmBuilder::isAssertHandled(Element& e) {
   return e.isList() && e.size() >= 2 && e[0]->isStr()
       && (e[0]->str() == Name("assert_return") ||
